/*{{{  'lex.c'*/
/*{{{  *** mods for IDA ****/
/*------------------------------------------------------------
Mods for IDA
------------
-- suppress o/p of comments by:
	-- commenting out 'putchar ('\n') in 'process_comment()'
-- suppress o/p of blank lines by commenting out 'putchar' in
	line handling fns.
------------------------------------------------------------*/
/*}}}*/
/*----------------------------------------------------------------------
Some mods to 'yyparse()' rules are necessary
-- in access stmt, generic spec can be OPERATOR() and ASSIGNMENT(=) -- but
	it looks like these are returned as TK_ID in this context
	(cf. in INTERFACE stmt!)
-- Also, in USE stmt, is OPERATOR & ASSIGNMENT returned as a token??

*** GET RID OF THE DEFN OF KEYWORD "HPF" (TK_HPF) LATER -- JUST USED FOR
INITIAL TESTING NOW!! ***
----------------------------------------------------------------------*/
/*{{{  #include's*/
#include <stdio.h>
#define STDIO_H
#include <ctype.h>
#include <signal.h>

#include "adapt.h"
#include "idafile.h"
#include "label.h"
#include "lex.h"
#include "map.h"	/* for 'default_proc_sp []' */
#include "output.h"
#include "strmacs.h"
#include "symbol.h"
#include "tokens.h"
#include "type.h"
#include "typedef.h"
#include "unit.h"
#include "unixstuff.h"

/*}}}*/
/*{{{  decls*/
/*{{{  globals in "lex.h"*/

char  textbuf [TEXTBUF_LEN];

char  next_label [8];	/* To hold the next F77-type label and     */
int   start_col;	/* col no.  Used in 'lex.c' and 'filer.c'. */

char  *end_tk,  *start_tk,  *stmt_startp;

bool  free_source_form,  give_warnings,  in_HPF_dir,  need_label;

char  *src_filename;
FILE  *ifp;
int    lineno,  nparens = 0;

/*--------------------------------------------------------------*/
/*	'textbuf' holds a complete stmt at a time, which may	*/
/* have been spread over several lines of input.  Therefore,	*/
/* the following pairs of arrays record line numbers for the	*/
/* text in a stmt:  'stmt_endp [i]' gives the posn in 'textbuf'	*/
/* of the *final* character on line 'stmt_lineno [i]'.		*/
/* (Note that, when a token has been scanned, 'start_tk' points	*/
/* to its start in 'textbuf', so the test:			*/
/*	stmt_endp [i-1] < start_tk <= stmt_endp [i]		*/
/* establishes that the token began on line 'stmt_lineno [i]'.)	*/
/*	This info is used, e.g., in 'write_err_lineno()'.	*/
/*--------------------------------------------------------------*/
char  *stmt_endp [MAX_CONTNS+1];
int    stmt_lineno [MAX_CONTNS+1];
/*}}}*/
/*{{{  globals in 'adapt.h'*/
bool   check_stack;
int    stack_len;
/*}}}*/

/*{{{  macros*/
#define  isidchar(c)	(isupper(c) || isdigit(c) || (c) == '_')

#define  get_stmt()	(free_source_form ? get_free_form_stmt () :	\
			                    get_fixed_form_stmt ())

#define  get_nonwhite_after_parens(c,ptr)				\
		(c = *(ptr = GET_NONWHITE_AFTER_PARENS (ptr)))

/*------------------------------------------------------*/
/* read or write while incrementing a column counter	*/
/*------------------------------------------------------*/
#define  readchar(col)		(col++, getc (ifp))

#ifdef  V_ADAPT
#define  writechar(col,c)	(col++, putchar(c))
#else
#define  writechar(col,c)	/* empty */
#endif
/*}}}*/

/*------------------------------------------------------*/
/*  Keywords						*/
/*  (the constants TK_XXX are defined in "y.tab.h")	*/
/*------------------------------------------------------*/

/*{{{  decls of 'keyword_fn's*/
int	attr_keyword (),	local_scope (),		type_keyword (),
	uncond_keyword (),	CASE_keyword (),	DO_keyword (),
	FUNCTION_keyword (),	IMPLICIT_keyword (),	INTERFACE_keyword (),
	MODULE_keyword (),	USE_keyword (),		ALIGN_keyword (),
	DISTRIBUTE_keyword (),	ONTO_keyword (),	WITH_keyword ();

/*}}}*/

/*{{{  typedef 'Keyword'*/
typedef struct Keyword {
	char *name;
	int  token;
	bool  maybe_split;
	int  (*keyword_fn)();
	struct Keyword  *next;
} Keyword;

/*}}}*/

/*{{{  def & init of 'keywords []'*/
Keyword  keywords [] = {
/*{{{  .NAME.    (searched for by 'lookup_dotted_keyword()')*/
/*					maybe_			     */
/*  name	token			split	keyword_fn	next */
/*  ----	-----			-----	----------	---- */

".AND.",	TK_AND_OP,		0,	0,		0,
".EQ.",		TK_REL_OP,		0,	0,		0,
".EQV.",	TK_EQV_OP,		0,	0,		0,
".FALSE.",	TK_LOGICAL_LIT_CONST,	0,	0,		0,
".GE.",		TK_REL_OP,		0,	0,		0,
".GT.",		TK_REL_OP,		0,	0,		0,
".LE.",		TK_REL_OP,		0,	0,		0,
".LT.",		TK_REL_OP,		0,	0,		0,
".NE.",		TK_REL_OP,		0,	0,		0,
".NEQV.",	TK_NEQV_OP,		0,	0,		0,
".NOT.",	TK_NOT_OP,		0,	0,		0,
".OR.",		TK_OR_OP,		0,	0,		0,
".TRUE.",	TK_LOGICAL_LIT_CONST,	0,	0,		0,

/*}}}*/
/*{{{  A - D     (searched for by 'lookup_keyword ()')*/
/*					maybe_			     */
/*  name	token			split	keyword_fn	next */
/*  ----	-----			-----	----------	---- */

"ALLOCATABLE",	TK_ALLOCATABLE,		0,	attr_keyword,	0,
"ALLOCATE",	TK_ALLOCATE,		0,	0,		0,
"ASSIGN",	TK_ASSIGN,		0,	uncond_keyword,	0,
"ASSIGNMENT",	TK_ASSIGNMENT,		0,	0,		0,

"BACKSPACE",	TK_BACKSPACE,		0,	0,		0,
"BLOCK",	0,			TRUE,	0,		0,
"BLOCKDATA",	TK_BLOCK_DATA,		0,	0,		0,

"CALL",		TK_CALL,		0,	0,		0,
"CASE",		TK_CASE,		0,	CASE_keyword,	0,
"CHARACTER",	TK_CHARACTER,		0,	type_keyword,	0,
"CLOSE",	TK_CLOSE,		0,	0,		0,
"COMMON",	TK_COMMON,		0,	local_scope,	0,
"COMPLEX",  	TK_COMPLEX,		0,	type_keyword,	0,
"CONTAINS",  	TK_CONTAINS,		0,	0,		0,
"CONTINUE",  	TK_CONTINUE,		0,	0,		0,
"CYCLE",  	TK_CYCLE,		0,	0,		0,

"DATA",		TK_DATA,		0,	0,		0,
"DEALLOCATE",	TK_DEALLOCATE,		0,	0,		0,
"DEFAULT",	TK_DEFAULT,		0,	0,		0,
"DIMENSION",	TK_DIMENSION,		0,	attr_keyword,	0,
"DO",		TK_DO,			0,	DO_keyword,	0,
"DOUBLE",	0,			TRUE,	0,		0,
"DOUBLEPRECISION",
		TK_DOUBLE_PRECISION,	0,	type_keyword,	0,
/*}}}*/
/*{{{  E - G                     "*/
/*					maybe_			     */
/*  name	token			split	keyword_fn	next */
/*  ----	-----			-----	----------	---- */

"ELSE",		TK_ELSE,		TRUE,	0,		0,
"ELSEIF",	TK_ELSE_IF,		0,	uncond_keyword,	0,
"ELSEWHERE",	TK_ELSE_WHERE,		0,	0,		0,
"END",		TK_END,			TRUE,	0,		0,
"ENDBLOCK",	0,			TRUE,	0,		0,
"ENDBLOCKDATA",	TK_END_BLOCK_DATA,	0,	0,		0,
"ENDDO",	TK_END_DO,		0,	0,		0,
"ENDFILE",	TK_END_FILE,		0,	0,		0,
"ENDFUNCTION",	TK_END_FUNCTION,	0,	0,		0,
"ENDIF",	TK_END_IF,		0,	0,		0,
"ENDINTERFACE",	TK_END_INTERFACE,	0,	0,		0,
"ENDMODULE",	TK_END_MODULE,		0,	0,		0,
"ENDPROGRAM",	TK_END_PROGRAM,		0,	0,		0,
"ENDSELECT",	TK_END_SELECT,		0,	0,		0,
"ENDSUBROUTINE",TK_END_SUBROUTINE,	0,	0,		0,
"ENDTYPE",	TK_END_TYPE,		0,	0,		0,
"ENDWHERE",	TK_END_WHERE,		0,	0,		0,
"ENTRY",	TK_ENTRY,		0,	local_scope,	0,
"EQUIVALENCE",	TK_EQUIVALENCE,		0,	local_scope,	0,
"EXIT",		TK_EXIT,		0,	0,		0,
"EXTERNAL",	TK_EXTERNAL,		0,	attr_keyword,	0,

"FORMAT",	TK_FORMAT,		0,	0,		0,
"FUNCTION",	TK_FUNCTION,		0,	FUNCTION_keyword,  0,

"GO",		0,			TRUE,	0,		0,
"GOTO",		TK_GO_TO,		0,	0,		0,

/*}}}*/
/*{{{  I - N                     "*/
/*					maybe_			     */
/*  name	token			split	keyword_fn	next */
/*  ----	-----			-----	----------	---- */

"IF",		TK_IF,			0,	0,		0,
"IMPLICIT",	TK_IMPLICIT,		0,	IMPLICIT_keyword,  0,
"INCLUDE",	TK_INCLUDE,		0,	0,		0,
"INQUIRE",	TK_INQUIRE,		0,	0,		0,
"INTEGER",  	TK_INTEGER,		0,	type_keyword,	0,
"INTENT",	TK_INTENT,		0,	attr_keyword,	0,
"INTERFACE",	TK_INTERFACE,		0,	INTERFACE_keyword,  0,
"INTRINSIC",	TK_INTRINSIC,		0,	attr_keyword,	0,

"LOGICAL",	TK_LOGICAL,		0,	type_keyword,	0,

"MODULE",	TK_MODULE,		0,	MODULE_keyword,	0,

"NAMELIST",	TK_NAMELIST,		0,	0,		0,
"NONE",		TK_NONE,		0,	0,		0,
"NULLIFY",	TK_NULLIFY,		0,	0,		0,

/*}}}*/
/*{{{  O - R                     "*/
/*					maybe_			     */
/*  name	token			split	keyword_fn	next */
/*  ----	-----			-----	----------	---- */

"ONLY",		TK_ONLY,		0,	0,		0,
"OPEN",		TK_OPEN,		0,	0,		0,
"OPERATOR",	TK_OPERATOR,		0,	0,		0,
"OPTIONAL",	TK_OPTIONAL,		0,	attr_keyword,	0,

"PARAMETER",	TK_PARAMETER,		0,	attr_keyword,	0,
"PAUSE",	TK_PAUSE,		0,	0,		0,
"POINTER",	TK_POINTER,		0,	attr_keyword,	0,
"PRINT",	TK_PRINT,		0,	0,		0,
"PRIVATE",	TK_PRIVATE,		0,	attr_keyword,	0,
"PROCEDURE",	TK_PROCEDURE,		0,	local_scope,	0,
"PROGRAM",	TK_PROGRAM,		0,	0,		0,
"PUBLIC",	TK_PUBLIC,		0,	attr_keyword,	0,

"READ",		TK_READ,		0,	0,		0,
"REAL",		TK_REAL,		0,	type_keyword,	0,
"RECURSIVE",	TK_RECURSIVE,		0,	uncond_keyword,	0,
"RESULT",	TK_RESULT,		0,	0,		0,
"RETURN",	TK_RETURN,		0,	0,		0,
"REWIND",	TK_REWIND,		0,	0,		0,

/*}}}*/
/*{{{  S - Z                     "*/
/*					maybe_			     */
/*  name	token			split	keyword_fn	next */
/*  ----	-----			-----	----------	---- */

"SAVE",		TK_SAVE,		0,	attr_keyword,	0,
"SELECT",	0,			TRUE,	0,		0,
"SELECTCASE",	TK_SELECT_CASE,		0,	0,		0,
"SEQUENCE",	TK_SEQUENCE,		0,	local_scope,	0,
"STOP",		TK_STOP,		0,	0,		0,
"SUBROUTINE",	TK_SUBROUTINE,		0,	local_scope,	0,

"TARGET",	TK_TARGET,		0,	attr_keyword,	0,
"THEN",		TK_THEN,		0,	0,		0,
"TO",		TK_TO,			0,	0,		0,
"TYPE",		TK_TYPE,		0,	type_keyword,	0,

"USE",		TK_USE,			0,	USE_keyword,	0,

"WHERE",	TK_WHERE,		0,	0,		0,
"WHILE",	TK_WHILE,		0,	0,		0,
"WRITE",	TK_WRITE,		0,	0,		0,
/*}}}*/

/*{{{  HPF keywords              "*/
/*					maybe_			     */
/*  name	token			split	keyword_fn	next */
/*  ----	-----			-----	----------	---- */

/*--------*/
/* HPF 3. */
/*--------*/
"ALIGN",	TK_ALIGN,		0,	ALIGN_keyword,	0,
"DISTRIBUTE",	TK_DISTRIBUTE,		0,	DISTRIBUTE_keyword,  0,
"DYNAMIC",	TK_DYNAMIC,		0,	attr_keyword,	0,
"INHERIT",	TK_INHERIT,		0,	attr_keyword,	0,
"ONTO",		TK_ONTO,		0,	ONTO_keyword,	0,
"PROCESSORS",	TK_PROCESSORS,		0,	attr_keyword,	0,
"REALIGN",	TK_REALIGN,		0,	ALIGN_keyword,	0,
"REDISTRIBUTE",	TK_REDISTRIBUTE,	0,	DISTRIBUTE_keyword,  0,
"TEMPLATE",	TK_TEMPLATE,		0,	attr_keyword,	0,
"WITH",		TK_WITH,		0,	WITH_keyword,	0,

/*--------*/
/* HPF 4. */
/*--------*/
"ENDFORALL",	TK_END_FORALL,		0,	0,		0,
"FORALL",	TK_FORALL,		0,	0,		0,
"INDEPENDENT",	TK_INDEPENDENT,		0,	uncond_keyword,	0,
"NEW",		TK_NEW,			0,	0,		0,
"PURE",		TK_PURE,		0,	uncond_keyword,	0,

/*--------*/
/* HPF 6. */
/*--------*/
"EXTRINSIC",	TK_EXTRINSIC,		0,	uncond_keyword,	0,

/*----------------------------------------------*/
/* HPF 7. 					*/
/* (N.B. "SEQUENCE" is already an F90 keyword.)	*/
/*----------------------------------------------*/
"NO",		0,			TRUE,	0,		0,
"NOSEQUENCE",	TK_NO_SEQUENCE,		0,	local_scope,	0,

0,		0,			0,	0,		0
/*}}}*/
};
/*}}}*/

static Keyword  *keytab [HASHSIZE]; /* pointers to keyword table */

#define  TRAILING_COMMENT	TRUE	/* argument of 'process_comment()' */
#define  COMMENT_LINE		FALSE	/*     "     "      "      "       */

bool  impl_stmt = FALSE;
int   nwords_to_keyword;

bool   f77_stmt_continued (),  process_comment ();
char  *analyse_stmt (),  *get_free_form_stmt (),  *get_fixed_form_stmt (),
      *get_f77_stmt_line (),  *scan_num_and_kind_param (),
      *GET_NONWHITE_AFTER_PARENS ();
/*}}}*/

/*{{{  DESCRIPTION -- Lexing procedure;  limitations;  files*/
#if 0

/*{{{  Lexing procedure*/
Lexing procedure
----------------

1. A complete stmt (incl all continuation lines) is read verbatim
into the buffer 'textbuf []'.
	IDs are not converted to upper case, and multiple blanks are
not condensed to a single one, to avoid corrupting Hollerith strings
(which are hard to detect when the stmt is being read in!).

2. 'yylex()' maintains 2 global pointers:
	'start_tk' -- points to the first char of the next token,
which is about top be read;
	'end_tk' -- points to the char immediately following the end
of the last token read.  This is set by the individual token-reading
functions ('read_name ()', etc) just before they return control to
'yylex ()'.  The next time 'yylex ()' is called to read a token, it
starts scanning for the next non-white from position 'end_tk';
when it has found it, it position is recorded in 'startp' before
the token is read.

'start_tk' is used by the 'errmsg()' and 'warning()' routines to
determine the lineno of the errmsg.  Thus, if a token is split across
multiple liens, the reported lineno is that of the *start* of the
token rather than its end (as it would be if 'end_tk' were used).

3.  The token-reading functions (e.g. 'read_name ()') convert
letters in IDs & keywords to upper case in place in the 'textbuf []'.
/*}}}*/

/*{{{  Limitations*/
Limitations
-----------

1. Whitespace is significant in fixed source form.


2. 'analyse_stmt ()' should scan stmt beginning with word FORMAT
more carefully, so can tell, e.g., that the following is a FORMAT stmt
containing a Hollerith:

	REAL FORMAT (100,100)
	FORMAT (6h) = (2)

while this is an assignment to an array element:

	FORMAT (6,h) = (2)


3.  Holleriths in FORMAT stmts require particular care.  If you
can't detect them lexically, then '!' (in fixed & free form) is
taken as a comment delimiter, while in fact it may be a char within
a Hollerith.  Likewise, '&' (in free form) is taken as a continuation
char, while again it may be within a Hollerith!
	E.g. consider:

9000	FORMAT (8 h12&  !90,2x)      ! Format statement, with Hollerith
				     ! string "12&  !90".
versus:

9001	FORMAT (8,h12&  !90,2x)      ! Trailing comment "!90, 2x)".
                     &) = 2          ! Assignment to element FORMAT (8, h12).
/*}}}*/

/*{{{  Files*/
Files
-----
(1) Without 'ida_directory':

idagen  x.f    d1/d2/x.f
  -->   x.ida  d1/d2/x.ida


(2) With 'ida_directory' 'dir' (either relative or absolute pathname):

idagen -d dir  x.f        d1/d2/x.f
  -->          dir/x.ida  dir/d1/d2/x.ida

Problem:
-- *** If 'src_filename' is absolute, will result in illegal filename. ***
E.g.:
              /usr/jhm/x.f -->  dir//usr/jhm/x.f

-- if 'src_filename' is relative, 'ida_directory' must contain same
directory structure.  Ok, seems sensible to preserve orig src directory
structure in 'ida' directory!


(3) If have 'x.f' and 'x.f90' in same directory, 2nd 'x.ida' will overwrite
the first!
Can we do something about this?

Since src_filename is stored in 'ida' file, 'ida_filename' doesn't have
to be closely related to it.  *But* it should be derivable from it.

Hmm! Add '.ida' to full filename?  But name may be too long!
Anyway, 'f77' will produce clashing '.o' files!


(4) In 'ida_filename', should we write 'src_filename' with or without path??

More flexible *without* path.  To do this:
-- In 'idagen', must have a ptr 'rootp' to point to start of last
component of name (after last '/', if any!) -- OK!

-- In 'idaproc', have to re-create the leading part of pathname from that
of the ida file!  E.g.:

idaproc  -d ida_dir  a/b/c/x.ida  (which contains src_filename == x.f90)
    -->              a/b/c/x.f90
(A BIT HARDER!! )
/*}}}*/

#endif
/*}}}*/


#ifdef V_IDA
/*{{{  main (int argc, char *argv[])  -- IDA version*/
main (argc, argv)		/* IDA version */
/*-------------*/
	int  argc;
	char *argv [];
/*--------------------------------------------------------------*/
/* Command line options						*/
/* --------------------						*/
/* N.B. each '-' can be followed by any number of options	*/
/*	-d,-D	directory for 'ida' files			*/
/*	-f,-F	free source format  (new in F90!)		*/
/*	-x,-X	fixed source format (like F77!)			*/
/*	-h,-H	give help information				*/
/*	-w,-W	supress warnings				*/
/*--------------------------------------------------------------*/
{
/*** HERE WE ALLOW STD INPUT FOR TESTING!! ******/
	extern int  yydebug;
	bool   specified_src_form = FALSE;
	char  *ida_directory = NULL;
	int    dir_len = 0;

	init_keytab ();
	init_intrintab ();
	init_primaries ();
	init_HPF_records ();
	create_unit_record (EXTERNAL_LEVEL, 0);

	/*{{{  defaults     *** SET yydebug = TRUE for parser debugging ****/
	/*----------*/
	/* defaults */
	/*----------*/
	yydebug = FALSE;
	give_warnings = TRUE;
	/*}}}*/
	/*{{{  process command line options (starting '-')*/
	/*{{{  <<< Options -- notes >>>*/
	/*----------------------------------------------------------------------
	Options -- notes:
	-----------------
	
	(i)  We allow '-' within the option list, so that successive 'option
	words'  (i.e. '-' followed by option letters) needn't be separated
	by spaces (e.g. so that "-f-w" is legal).
	
	We also allow '-' to be repeated (e.g. "--"), and don't insist
	that '-' is followed by a letter, to avoid being irritatingly fussy!
	(Anyway, this permissiveness seems to be standard practise for Unix
	commands, e.g. try 'ls'!).
	
	(ii)  We don't check for or complain about repetition of option letters,
	again to avoid irritating fussiness!  (Again this seems to be standard
	practice for Unix commands.)
	
	(iii)  We even allow the incompatible options '-f' and '-x' (for free and
	fixed source form respectively) to both be specified, in which case 
	the final one takes effect.  The is mainly to allow for cases like:
	
		ida  -x *.f77  -f *.f    # all ".f77" files are fixed source form
		                         # and all ".f" files free source form
	
	where "*.f77" expands to no files at all, in which case '-x' and '-f'
	become consecutive after filename substitution.
	----------------------------------------------------------------------*/
	/*}}}*/
	
	while (--argc && *(*++argv) == '-') {
		int  c;
		
		while (c = *++(*argv)) {
			switch (c) {
			case 'd' :  case 'D' :
				/*{{{  name of directory for 'ida' files*/
				/*------------------------------------------------------*/
				/* -d option:  directory for 'ida' files.		*/
				/* Now we expect a space followed by a pathname.	*/
				/*------------------------------------------------------*/
				if (*++(*argv)) {
					fputs ("Expect a space after the \'-d\' option.", stderr);
					goto  abort;
				}
				
				if ( ! --argc || **++argv == '-') {
					fputs ("Expect a directory path after the \'-d\' option.", stderr);
					goto  abort;
				}
				
				/*---------------*/
				/* read filename */
				/*---------------*/
				ida_directory = *argv;
				dir_len = strlen (ida_directory) + 1;		/* +1 for '/' */
				
				goto  next_word;
				/*}}}*/
	
			case 'f' :  case 'F' :
				/*----------------------*/
				/* free source form	*/
				/*----------------------*/
				free_source_form = TRUE;
				specified_src_form = TRUE;
				break;
	
			case 'x' :  case 'X' :
				/*----------------------*/
				/* fixed source form	*/
				/*----------------------*/
				free_source_form = FALSE;
				specified_src_form = TRUE;
				break;
	
			case 'h' :  case 'H' :
				helpinfo ();
				exit (2);	/* neither a failed nor a */
						/* successful execution! */
	
			case 'w' :  case 'W' :
				/*----------------------*/
				/* supress warnings	*/
				/*----------------------*/
				give_warnings = FALSE;
				break;
	
			case '-' :
				break;
	
			default :
				fprintf (stderr, "Illegal option -%c.", c);
				goto  abort;
			}
		}
	
	next_word :
		continue;
	}
	/*}}}*/
	/*{{{  if no source file specified, give errmsg & return *** MODIFY FOR TESTING****/
	/*********  PROPER VERSION **********
	
	if (! argc) {
		fputs ("No input file specified.", stderr);
		goto  abort;
	}
	
	*********/
	
	/***** THE FOLLOWING ALLOWS I/P FROM stdin FOR TESTING... */
	
	/*--------------------------------------------------------------*/
	/* We must initialise the following global vars for the first	*/
	/* line of input!						*/
	/*--------------------------------------------------------------*/
	if (! argc) {
		need_label = TRUE;
		lineno = 0;
		ifp    = stdin;
		yyparse ();
		exit (0);
	}
	/*******************************/
	/*}}}*/
	fputs ("Creating the 'ida' database files...\n\n", stderr);
	for (  ;  argc;  argc--, argv++) {
		/*{{{  decls*/
		char  *extn_p,  *root_p,
		      *ida_filename;		/* o/p file, "xxx.ida" */
		int    root_len;
		long   src_modify_time;		/* data type may be system-dependent */
		/*}}}*/
		
		/*{{{  is it an option ('-f' or '-x')?*/
		/*--------------------------------------------------------------*/
		/* Check for a '-f' or '-x' option among the filenames.		*/
		/*								*/
		/* N.B.  If this argument *is* an option (starting '-') then we	*/
		/* must be within the list of filenames, as options preceding	*/
		/* it are processed above.					*/
		/*	See the comments preceding the earlier 'options' code	*/
		/* for a discussion of why we allow isolated & repeated '-',	*/
		/* repeated option letters, and even incompatible options	*/
		/* (e.g. "-f -x").						*/
		/*--------------------------------------------------------------*/
		if (**argv == '-') {
			int  c;
		
			while (c = *++(*argv)) {
				switch (c) {
				case 'f' :  case 'F' :
					/*----------------------*/
					/* free source form	*/
					/*----------------------*/
					free_source_form = TRUE;
					specified_src_form = TRUE;
					break;
		
				case 'x' :  case 'X' :
					/*----------------------*/
					/* fixed source form	*/
					/*----------------------*/
					free_source_form = FALSE;
					specified_src_form = TRUE;
					break;
		
				case '-' :
					break;
		
				default :
					fprintf (stderr, "Illegal option -%c.", c);
					goto  abort;
				}
			}
		
			continue;	/* go to next argument */
		}
		/*}}}*/
		/*{{{  read 'src_filename' & open 'ida_filename'*/
		/*--------------------------------------------------------------*/
		/* Read 'src_filename' (the next Fortran input file) and open	*/
		/* 'ida_filename' (the corresponding output file).		*/
		/*								*/
		/* N.B. 'src_filename' is a global variable!			*/
		/*--------------------------------------------------------------*/
		src_filename = *argv;
		
		/*{{{  set 'root_p' ('src_filename' without path) & 'extn_p' (ptr to its last '.', if any)*/
		/*--------------------------------------------------------------*/
		/* Set pointers 'root_p' and 'extn_p' as follows:		*/
		/*								*/
		/* root_p -- pointer to the start of the 'src_filename', after	*/
		/*	skipping over any leading directory path component	*/
		/*	(if there is one);					*/
		/* extn_p -- pointer to the final '.' in '*root_p' (if there is	*/
		/*	one).							*/
		/*--------------------------------------------------------------*/
		
		for (extn_p = NULL,  root_p = src_filename;  **argv;  (*argv)++) {
			switch (**argv) {
			case  DIR_PATH_SEP :
				root_p = (*argv + 1);	/* ptr to char after the '/' */
				extn_p = NULL;		/* if 'extn_p' was set it pointed */
							/* to a '.' in a directory name,  */
							/* not the actual source filename */
				break;
				
			case  '.' :
				extn_p = *argv;		/* ptr to the '.' */
				break;
			}
		}
		/*}}}*/
		
		/*{{{  disallow 'src_filename' with extension ".ida"*/
		if (extn_p  && str_match (extn_p, ".ida")) {
			fprintf (stderr, "Cannot specify an `ida\' file as input: \"%s\".\n",
			                  src_filename);
			exit (1);
		}
		/*}}}*/
		
		if (! specified_src_form) {
			/*{{{  deduce required source form from filename extension*/
			/*--------------------------------------------------------------*/
			/* The required 'source form' isn't explicitly specified	*/
			/* (by "-f" or "-F"), so deduce it from the filename extension	*/
			/* (i.e. its fixed source form unless the filename ends ".f90").*/
			/*--------------------------------------------------------------*/
			
			free_source_form = (extn_p  &&  str_match (extn_p, ".f90"));
			/*}}}*/
		}
		
		/*{{{  if can't read 'src_filename', give errmsg & exit.  Find 'src_modify_time'.*/
		if ( ! freopen (src_filename, "r", stdin)) {
			fprintf (stderr, "Cannot open file \"%s\".\n", src_filename);
			exit (1);
		}
		src_modify_time = modify_time (src_filename);
		/*}}}*/
		
		/*{{{  if corresponding ".ida" file ('ida_filename') doesn't exist, create it*/
		/*--------------------------------------------------------------*/
		/* Compose the 'ida_filename'.  This is 'src_filename' with the	*/
		/* following modifications:					*/
		/* -- if 'ida_directory' is specified, it is prefixed to	*/
		/*	'src_filename';						*/
		/* -- if 'src_filename' has an extension it's replaced by	*/
		/*	".ida", otherwise ".ida" is appended to it.		*/
		/*--------------------------------------------------------------*/
		if (extn_p) *extn_p = '\0';	/* temporarily end 'src_filename' before '.' */
		
		root_len = strlen (src_filename);
		
		ida_filename = emalloc (dir_len + root_len + 5);	/* +5 for ".ida" */
		if (ida_directory) {
			strcpy (ida_filename,  ida_directory);
			*(ida_filename + dir_len - 1) = DIR_PATH_SEP;
		}
		strcpy (ida_filename + dir_len,  src_filename);
		strcpy (ida_filename + dir_len + root_len, ".ida");
		
		if (extn_p) *extn_p = '.';	/* restore the final '.' in 'src_filename' */
		
		/*--------------------------------------------------------------*/
		/* If file 'ida_filename' doesn't exist, create it.		*/
		/*	If it does exist, then:					*/
		/*  (i) if it's the 'ida' file corresponding to 'src_filename',	*/
		/* go onto the next source file.				*/
		/*  (ii) if it's an 'ida' file (i.e.its 1st line begins "ida: ")*/
		/* but it *doesn't* correspond to 'src_filename', open it for	*/
		/* writing (which will over-write the old contents);		*/
		/*  (iii) if it *isn't* an 'ida' file (according to the above	*/
		/* criterion), give an errmsg and quit (as we don't wish to	*/
		/* over-write a non-'ida' file!).				*/
		/*								*/
		/* N.B. *** The following code calls functions 'file_exists ()'	*/
		/* and 'modify_time ()' which make Unix system calls (and hence	*/
		/* aren't very portable!). ***					*/
		/*--------------------------------------------------------------*/
		if (file_exists (ida_filename)) {
			/*{{{  if it's up-to-date, continue;  if it isn't an 'ida' file, quit*/
			/*------------------------------------------------------------------
			File 'ida_filename' already exists.  Read its first line into
			'textbuf' and perform the following checks :
			
			(i)  *Is* it an 'ida' file?  It's accepted as such if its first line
			begins:
				"<<ida>>: "
			(Perhaps we should make this test more rigorous!).  If not, give an
			errmsg and quit (as we don't want to overwrite a non-'ida' file!).
			
			(ii)  If it's an 'ida' file, does it relate to this source file?
			To pass this test the 'src_filename' *and* its 'last modify time'
			must be as written in the first line of the 'ida' file.  If so, this
			'ida' file is used and the source file isn't processed.  Otherwise,
			the source file is processed and the 'ida' file overwritten.
			
			N.B. (TEXTBUF_LEN == 5242).  We assume that this is long enough to
			contain the first line of any valid 'ida' file, which is simply:
				"ida: src_filename last_modify_time \n"
			------------------------------------------------------------------*/
			FILE  *filep;		/* temporary file pointer for reading 'ida' file */
			
			if ( ! (filep = fopen (ida_filename, "r")) ) {
				fprintf (stderr, "\"%s\":  *** file \"%s\" exists but cannot be read.  Please change its permissions, rename it or remove it.\n",
				                 src_filename,  ida_filename);
				exit (1);
			}
			
			/*------------------------------------*/
			/* Read its first line into 'textbuf' */
			/*------------------------------------*/
			if (! fgets (textbuf, (TEXTBUF_LEN - 1), filep)) {
				/*------------------------------------------------------*/
				/* Empty 'ida' file!  Forget it and create a new one.	*/
				/*------------------------------------------------------*/
				fclose (filep);
				goto  create_ida_file;
			}
			fclose (filep);
			
			if (strncmp (textbuf, IDA_PREFIX, IDA_PREFIX_LEN)) {
				fprintf (stderr, "\"%s\":  *** \"%s\" is corrupt or not an `ida\' database file.  Please rename or remove it and start again.\n",
				                 src_filename,  ida_filename);
				exit (1);
			}
			else if (ida_file_up_to_date (textbuf, root_p, src_modify_time)) {
				fprintf (stderr, "\"%s\":  \"%s\" is up-to-date.\n",
				                  src_filename, ida_filename);
				continue;		/* process next source file */
			}
			/*--------------------------------------------------------------*/
			/* ... else, overwrite this old 'ida' file with a new one...	*/
			/*--------------------------------------------------------------*/
			/*}}}*/
		}
		
		create_ida_file :
		
		if ( ! freopen (ida_filename, "w", stdout)) {
			fprintf (stderr, "\"%s\":  *** cannot open or create output file \"%s\".\n",
			                  src_filename, ida_filename);
			exit (1);
		}
		
		fprintf (stderr, "\"%s\":  creating \"%s\".\n", src_filename, ida_filename);
		
		free (ida_filename);		/* no longer needed! */
		/*}}}*/
		
		/*{{{  write a special 1st line to the 'ida' file*/
		/*--------------------------------------------------------------*/
		/* Write special first line to the 'ida' file, viz:		*/
		/*								*/
		/* "<<ida>>: v0.3 src_filename modify_time \n"			*/
		/*								*/
		/* N.B. Although 'src_modify_time' is of type 'long', we write	*/
		/* it with '%d' rather than '%ld'.  This doesn't matter, as the	*/
		/* only effect of specifying '%ld' is to prefix the output	*/
		/* number by 'L'.						*/
		/*--------------------------------------------------------------*/
		printf ("%s%s%s %d \n", IDA_PREFIX,  IDA_VERSION,  root_p, src_modify_time);
		/*}}}*/
		
		/*}}}*/

		/*{{{  init certain global vars for the 1st line of input*/
		/*--------------------------------------------------------------*/
		/* We must initialise the following global vars for the first	*/
		/* line of input!						*/
		/*--------------------------------------------------------------*/
		need_label = TRUE;
		lineno = 0;
		ifp    = stdin;
		/*}}}*/

		yyparse ();
	}
	exit (0);	/* successful execution! */
	
abort :
	fputs ("  (Type \'ida -h\' for help.)\n", stderr);
	exit (1);
}
/*}}}*/
/*{{{  helpinfo ()                    -- IDA version*/
helpinfo ()
/*-------*/
/*---------------------------*/
/*  Output help information  */
/*---------------------------*/
/*{{{  appearance of output*/
/*-----------------------------------------------------------------
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
| Appearance of 'help' o/p |
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ida:  Inter-procedural Dependency Analyser (version 0.2)

Usage:
        ida  [-d <directory>]  [-fxhw]  <file>  [[-fx] <file>]...

(where [] encloses an optional item, []... encloses an optional item
that may be repeated, and <..> denotes a name supplied by the user.)

The options are:

  -d <directory>
        ".ida" database files are stored in directory <directory>
        rather than the current directory.

  -f    Free source form input (i.e. the new Fortran 90 source form).

  -x    Fixed source form input (i.e. Fortran 77 source form).
        (Without either the '-f' or '-x' options, the default is
        free source form for filenames ending ".f90" and fixed source
        form otherwise).

  -h    Give help information.

  -w    Suppress warnings.

The options may be specified individually or in groups, in any order
and either case.

`<file>'(s) must contain the source code for just one Fortran program.
It is not essential for all program units of the program to be present.

When `ida' is running, the command `help' can be entered to obtain
information about `ida's commands.

-----------------------------------------------------------------*/
/*}}}*/
{
	fputs ("\nida:  Inter-procedural Dependency Analyser (version 0.2)\n\nUsage:\n        ida  [-d <directory>]  [-fxhw]  <file>  [[-fx] <file>]...\n\n(where [] encloses an optional item, []... encloses an optional item\nthat may be repeated, and <..> denotes a name supplied by the user.)\n\nThe options are:\n\n  -d <directory>\n        \".ida\" database files are stored in directory <directory>\n        rather than the current directory.\n\n  -f    Free source form input (i.e. the new Fortran 90 source form).\n\n  -x    Fixed source form input (i.e. Fortran 77 source form).\n        (Without either the \'-f\' or \'-x\' options, the default is\n        free source form for filenames ending \".f90\" and fixed source\n        form otherwise).\n\n  -h    Give help information.\n\n  -w    Suppress warnings.\n\nThe options may be specified individually or in groups, in any order\nand either case.\n\n`<file>\'(s) must contain the source code for just one Fortran program.\nIt is not essential for all program units of the program to be present.\n\nWhen `ida\' is running, the command `help\' can be entered to obtain\ninformation about `ida\'s commands.\n\n",
	        stderr);
}
/*}}}*/

#else
/*{{{  main (int argc, char *argv[])  -- ADAPT version*/
main (argc, argv)		/* ADAPT version */
/*-------------*/
	int  argc;
	char *argv [];
/*--------------------------------------------------------------*/
/* Command line options						*/
/* --------------------						*/
/* N.B. each '-' can be followed by any number of options	*/
/*	-d,-D	directory for output files	*** MODS ***	*/
/*	-f,-F	free source format  (new in F90!)		*/
/*	-x,-X	fixed source format (like F77!)			*/
/*	-h,-H	give help information				*/
/*	-o,-O	output file name	*** NEW ONE ***		*/
/*	-w,-W	supress warnings				*/
/*--------------------------------------------------------------*/
{
#define  MAX_FILENAME_LEN	64	/* a reasonable maximum! */
	/*{{{  decls*/
	extern int  yydebug;
	bool     specified_src_form = FALSE,  specified_out_file = FALSE;
	char    *out_dir = NULL;
	char    *def_proc_names = "$P1\0$P2\0$P3\0$P4\0$P5\0$P6\0$P7";
	int      dir_len = 0,  dim,  len,  rank;
	Array   *arp;
	Symbol  *sp;
	FILE    *adapt_dat_fp;
	/*}}}*/


	/*{{{  open & read file "adapt.dat" (** WORK ON IT; COMMENTED OUT FOR NOW ***)*/
	/*-------------------
	
	if ((adapt_dat_fp = fopen ("adapt.dat", "r")) == NULL) {
		fputs ("Cannot open \"adapt.dat\"\n", stderr);
		exit (1);
	}
	fputs ("\n\n", stderr);
	read_adapt_dat (adapt_dat_fp);
	
	-------------------*/
	/*}}}*/
	init_keytab ();
	init_intrintab ();
	init_primaries ();
	create_unit_record (EXTERNAL_LEVEL, 0);

	/*{{{  defaults     *** SET yydebug = TRUE for parser debugging ****/
	/*----------*/
	/* defaults */
	/*----------*/
	yydebug = FALSE;	/*** SET TO 'TRUE' TO DEBUG PARSER! ***/
	give_warnings = TRUE;
	check_stack   = FALSE;
	stack_len     = 250000;
	/*}}}*/
	/*{{{  process command line options (starting '-')*/
	/*{{{  <<< Options -- notes >>>*/
	/*----------------------------------------------------------------------
	Options -- notes:
	-----------------
	
	(i)  We allow '-' within the option list, so that successive 'option
	words'  (i.e. '-' followed by option letters) needn't be separated
	by spaces (e.g. so that "-f-w" is legal).
	
	We also allow '-' to be repeated (e.g. "--"), and don't insist
	that '-' is followed by a letter, to avoid being irritatingly fussy!
	(Anyway, this permissiveness seems to be standard practise for Unix
	commands, e.g. try 'ls'!).
	
	(ii)  We don't check for or complain about repetition of option letters,
	again to avoid irritating fussiness!  (Again this seems to be standard
	practice for Unix commands.)
	
	(iii)  We even allow the incompatible options '-f' and '-x' (for free and
	fixed source form respectively) to both be specified, in which case 
	the final one takes effect.  The is mainly to allow for cases like:
	
		ida  -x *.f77  -f *.f    # all ".f77" files are fixed source form
		                         # and all ".f" files free source form
	
	where "*.f77" expands to no files at all, in which case '-x' and '-f'
	become consecutive after filename substitution.
	----------------------------------------------------------------------*/
	/*}}}*/
	
	while (--argc && *(*++argv) == '-') {
		int  c;
		
		while (c = *++(*argv)) {
			switch (c) {
			case 'c' :  case 'C' :
				/*{{{  set (check_stack = TRUE)*/
				/*--------------------------------------------------------------*/
				/* -c option:  check for stack overflow (and eventually other	*/
				/* things as well!).						*/
				/*--------------------------------------------------------------*/
				check_stack = TRUE;
				break;
				/*}}}*/
	
			case 'd' :  case 'D' :
				/*{{{  name of directory for output files*/
				/*------------------------------------------------------*/
				/* -d option:  directory for 'ida' files.		*/
				/* Now we expect a space followed by a pathname.	*/
				/*------------------------------------------------------*/
				if (*++(*argv)) {
					fputs ("Expected a space after the \'-d\' option.", stderr);
					goto  abort;
				}
				
				if ( ! --argc || **++argv == '-') {
					fputs ("Expected a directory path after the \'-d\' option.", stderr);
					goto  abort;
				}
				
				/*---------------*/
				/* read filename */
				/*---------------*/
				out_dir = *argv;
				dir_len = strlen (out_dir) + 1;		/* +1 for '/' */
				
				goto  next_word;
				/*}}}*/
	
			case 'f' :  case 'F' :
				/*{{{  free source form*/
				/*----------------------*/
				/* free source form	*/
				/*----------------------*/
				free_source_form = TRUE;
				specified_src_form = TRUE;
				break;
				/*}}}*/
	
			case 'x' :  case 'X' :
				/*{{{  fixed source form*/
				/*----------------------*/
				/* fixed source form	*/
				/*----------------------*/
				free_source_form = FALSE;
				specified_src_form = TRUE;
				break;
				/*}}}*/
	
			case 'h' :  case 'H' :
				/*{{{  call 'helpinfo ()'*/
				/*--------------------------------------*/
				/* -h option:  give 'help' message	*/
				/*--------------------------------------*/
				helpinfo ();
				exit (2);	/* neither a failed nor a */
						/* successful execution! */
				/*}}}*/
	
			case 'o' :  case 'O' :
				/*{{{  name of o/p file  *** NOT SURE IF TERMINATED OK!*/
				/*--------------------------------------------------------------*/
				/* -o option: now expect a space followed by a filename.	*/
				/* If the output file can be created, attach it to 'stdout'.	*/
				/*--------------------------------------------------------------*/
				if (*++(argv [0])) {
					fputs ("Expected a space after \'-o\' (name of output file) option.\n",
					        stderr);
					exit (1);
				}
				
				if (--argc == 0 || (*++argv) [0] == '-') {
					fputs ("Expected a filename after the \'-o\' option.\n", stderr);
					exit (1);
				}
				
				/*--------------------------------------------------------------*/
				/* report error if output filename is too long or file already	*/
				/* exists.							*/
				/*--------------------------------------------------------------*/
				if ((len = strlen (*argv)) > MAX_FILENAME_LEN) {
					fprintf (stderr, "Output filename \"%s\" too long -- maximum of %d characters allowed!\n",
						          *argv,  MAX_FILENAME_LEN);
					exit (1);
				}
				if (file_exists (*argv)) {
					fprintf (stderr, "Output file \"%s\" already exists -- choose another output filename.\n",
					                  *argv);
					exit (1);
				}
				
				/*------------------*/
				/* open output file */
				/*------------------*/
				if ( ! freopen (*argv, "w", stdout)) {
					fprintf (stderr, "Cannot create output file \"%s\"\n", *argv);
					exit (1);
				}
				*argv += (len - 1);	/* move back to char before the '\0', to */
							/* terminate the enclosing 'while' loop. */
				specified_out_file = TRUE;
				break;
				/*}}}*/
				
			case 'p' :  case 'P' :
				/*{{{  'pN' (N == 1-7): shape for default proc array "$PN"*/
				/*------------------------------------------------------*/
				/* '-pN' option  (N == 1 -- 7).				*/
				/* Specifies shape for default processor array "$PN".	*/
				/*------------------------------------------------------*/
				c = *++(*argv);		/* next letter after '-p' in same word */
				
				if (! c  ||  c < '1'  ||  c > '7') {
					fputs ("Expected a digit in the range 1-7 immediately after \'-p\'.",
					        stderr);
					goto  abort;
				}
				if (*++(*argv)) {
					fprintf (stderr, "Expected a space after the \'-p%c\' option.", c);
					goto  abort;
				}
				
				rank = (c - '0');
				
				if (default_proc_sp [rank]) {
					fprintf (stderr, "Duplicated \'-p%c\' option.\n", c);
					exit (1);
				}
				
				sp = default_proc_sp [rank] = (Symbol *) ecalloc (sizeof (Symbol));
				
				sp->name = def_proc_names + 4 * (rank-1);	/* to point to "$PN",
										   where (N = rank) */
				sp->attrs = AT_PROCESSORS;
				
				arp = create_array ();
				for (dim = rank;   dim;   --dim) {
					int  extent;
				
					if ( ! --argc  ||  ((c = (*++argv) [0])  &&  ! isdigit (c)) ) {
						fprintf (stderr, "Expected %d %s after the \'-p%d\' flag.\n",
						         rank, (rank == 1 ? "number" : "numbers"), rank);
						exit (1);
					}
					sscanf (*argv, "%d", &extent);
					install_bounds (arp, NULL_Sym, NULL_Expr, int_lit_primary (extent));
				
					/*------------------------------------------------------*/
					/* move '*argv' to point to final digit of number	*/
					/*------------------------------------------------------*/
					while ((c = *++(*argv))  &&  isdigit (c));
					if (c) {
						fprintf (stderr, "Expected a space after parameter \'%d\' of the \'-p%d\' option.\n",
						                  extent, rank);
						exit (1);
					}
				}
				attach_array (sp->arp, arp);
				
				--(*argv);	/* to move back to final digit */
				
				break;
				
				/*}}}*/
	
			case 's' :  case 'S' :
				/*{{{  set 'stack_len'*/
				/*------------------------------*/
				/* -s option:  set 'stack_len'	*/
				/*------------------------------*/
				c = *++(*argv);		/* next letter after '-s' in same word */
				
				if (c  &&  ! isdigit (c)) {
					fputs ("Expected a number after the \'-s\' option.", stderr);
					goto  abort;
				}
				else if (! c) {
					if ( ! --argc  ||  ((c = (*++argv) [0])  &&  ! isdigit (c)) ) {
				
						fputs ("Expected a number after the \'-s\' option.", stderr);
						goto  abort;
					}
				}
				
				sscanf (*argv, "%d", &stack_len);
				
				/*------------------------------------------------------*/
				/* move '*argv' to point to final digit of number	*/
				/*------------------------------------------------------*/
				while ((c = *++(*argv))  &&  isdigit (c))
					;
				
				--(*argv);	/* to move back to final digit */
				
				break;
				/*}}}*/
	
			case 'w' :  case 'W' :
				/*{{{  set (give warnings = FALSE)*/
				/*------------------------------*/
				/* -w option:  supress warnings	*/
				/*------------------------------*/
				give_warnings = FALSE;
				break;
				/*}}}*/
	
			case '-' :
				break;
	
			default :
				fprintf (stderr, "Illegal option -%c.", c);
				goto  abort;
			}
		}
	
	next_word :
		continue;
	}
	/*}}}*/
	/*{{{  if no source file specified, give errmsg & return*/
	if (! argc) {
		fputs ("No input file specified.", stderr);
		goto  abort;
	}
	
	
	/***** THE FOLLOWING ALLOWS I/P FROM stdin FOR TESTING...
	*--------------------------------------------------------------*
	* We must initialise the following global vars for the first	*
	* line of input!						*
	*--------------------------------------------------------------*
	if (! argc) {
		need_label = TRUE;
		lineno = 0;
		ifp    = stdin;
		yyparse ();
		exit (0);
	}
	*******************************/
	/*}}}*/

	init_HPF_records ();	/* do this *after* reading any '-pN' options */

	for (  ;  argc;  argc--, argv++) {
		/*{{{  decls*/
		char  *extn_p,  *root_p,  *out_filename;
		int    root_len;
		/*}}}*/
		
		/*{{{  is it an option ('-f' or '-x')?*/
		/*--------------------------------------------------------------*/
		/* Check for a '-f' or '-x' option among the filenames.		*/
		/*								*/
		/* N.B.  If this argument *is* an option (starting '-') then we	*/
		/* must be within the list of filenames, as options preceding	*/
		/* it are processed above.					*/
		/*	See the comments preceding the earlier 'options' code	*/
		/* for a discussion of why we allow isolated & repeated '-',	*/
		/* repeated option letters, and even incompatible options	*/
		/* (e.g. "-f -x").						*/
		/*--------------------------------------------------------------*/
		if (**argv == '-') {
			int  c;
		
			while (c = *++(*argv)) {
				switch (c) {
				case 'f' :  case 'F' :
					/*----------------------*/
					/* free source form	*/
					/*----------------------*/
					free_source_form = TRUE;
					specified_src_form = TRUE;
					break;
		
				case 'x' :  case 'X' :
					/*----------------------*/
					/* fixed source form	*/
					/*----------------------*/
					free_source_form = FALSE;
					specified_src_form = TRUE;
					break;
		
				case '-' :
					break;
		
				default :
					fprintf (stderr, "Illegal option -%c.", c);
					goto  abort;
				}
			}
		
			continue;	/* go to next argument */
		}
		/*}}}*/
		/*{{{  read 'src_filename' & open 'out_filename'*/
		/*--------------------------------------------------------------*/
		/* Read 'src_filename' (the next Fortran input file) and open	*/
		/* 'out_filename' (the corresponding output file).		*/
		/*								*/
		/* N.B. 'src_filename' is a global variable!			*/
		/*--------------------------------------------------------------*/
		src_filename = *argv;
		
		/*{{{  set 'root_p' ('src_filename' without path) & 'extn_p' (ptr to its last '.', if any)*/
		/*--------------------------------------------------------------*/
		/* Set pointers 'root_p' and 'extn_p' as follows:		*/
		/*								*/
		/* root_p -- pointer to the start of the 'src_filename', after	*/
		/*	skipping over any leading directory path component	*/
		/*	(if there is one);					*/
		/* extn_p -- pointer to the final '.' in '*root_p' (if there is	*/
		/*	one).							*/
		/*--------------------------------------------------------------*/
		
		for (extn_p = NULL,  root_p = src_filename;  **argv;  (*argv)++) {
			switch (**argv) {
			case  DIR_PATH_SEP :
				root_p = (*argv + 1);	/* ptr to char after the '/' */
				extn_p = NULL;		/* if 'extn_p' was set it pointed */
							/* to a '.' in a directory name,  */
							/* not the actual source filename */
				break;
				
			case  '.' :
				extn_p = *argv;		/* ptr to the '.' */
				break;
			}
		}
		/*}}}*/
		
		if (! specified_src_form) {
			/*{{{  deduce required source form from filename extension*/
			/*--------------------------------------------------------------*/
			/* The required 'source form' isn't explicitly specified	*/
			/* (by "-f" or "-F"), so deduce it from the filename extension	*/
			/* (i.e. its fixed source form unless the filename ends ".f90"	*/
			/* or ".h90" or ".hpf90").					*/
			/*--------------------------------------------------------------*/
			
			free_source_form = (extn_p  &&  (str_match (extn_p, ".f90")  ||
			                                 str_match (extn_p, ".h90")  ||
			                                 str_match (extn_p, ".hpf90")) );
			/*}}}*/
		}
		
		/*{{{  if can't read 'src_filename', give errmsg & exit.*/
		if ( ! freopen (src_filename, "r", stdin)) {
			fprintf (stderr, "Cannot open file \"%s\".\n", src_filename);
			exit (1);
		}
		/*}}}*/
		
		if (! specified_out_file) {
			/*{{{  if corresponding output file ('out_filename') doesn't exist, create it*/
			/*--------------------------------------------------------------*/
			/* Compose the 'out_filename'.  This is 'src_filename' with the	*/
			/* following modifications:					*/
			/* -- if 'out_dir' is specified, it's prefixed to		*/
			/*	'src_filename';						*/
			/* -- if 'src_filename' has an extension it's replaced by	*/
			/*	"-OUT.f", otherwise "-OUT.f" is appended to it.		*/
			/*--------------------------------------------------------------*/
			if (extn_p) *extn_p = '\0';	/* temporarily end 'src_filename' before '.' */
			
			root_len = strlen (src_filename);
			
			out_filename = emalloc (dir_len + root_len + 7);	/* +7 for "-OUT.f" */
			if (out_dir) {
				strcpy (out_filename,  out_dir);
				*(out_filename + dir_len - 1) = DIR_PATH_SEP;
			}
			strcpy (out_filename + dir_len,  src_filename);
			strcpy (out_filename + dir_len + root_len, "-OUT.f");
			
			if (extn_p) *extn_p = '.';	/* restore the final '.' in 'src_filename' */
			
			/*--------------------------------------------------------------*/
			/* Create file 'out_filename' if it doesn't already exist.	*/
			/* If it does exist, give an errmsg and quit (as we don't want	*/
			/* to over-write an existing file!).				*/
			/*								*/
			/* N.B. *** The following code calls functions 'file_exists ()'	*/
			/* which makes Unix system calls (and hence isn't very		*/
			/* portable!). ***						*/
			/*--------------------------------------------------------------*/
			if (file_exists (out_filename)) {
				fprintf (stderr, "\"%s\":  *** output file \"%s\" already exists -- choose another output filename.\n",
				                  src_filename, out_filename);
				exit (1);
			}
			
			/*------------------*/
			/* open output file */
			/*------------------*/
			if ( ! freopen (out_filename, "w", stdout)) {
				fprintf (stderr, "\"%s\":  *** cannot create output file \"%s\".\n",
				                  src_filename, out_filename);
				exit (1);
			}
			
			fprintf (stderr, "\"%s\": creating \"%s\".\n", src_filename, out_filename);
			
			free (out_filename);		/* no longer needed! */
			/*}}}*/
		}
		
		/*}}}*/

		/*{{{  init certain global vars for the 1st line of input*/
		/*--------------------------------------------------------------*/
		/* We must initialise the following global vars for the first	*/
		/* line of input!						*/
		/*--------------------------------------------------------------*/
		need_label = TRUE;
		lineno = 0;
		ifp    = stdin;
		/*}}}*/

		yyparse ();
	}
	exit (has_errors ? 1 : 0);
	
abort :
	fputs ("  (Type \'adapt -h\' for help.)\n", stderr);
	exit (1);
}
/*}}}*/
/*{{{  helpinfo ()                    -- ADAPT version*/
helpinfo ()		/* ADAPT version */
/*-------*/
/*---------------------------*/
/*  Output help information  */
/*---------------------------*/
/*{{{  appearance of output*/
/*-----------------------------------------------------------------
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
| Appearance of 'help' o/p |
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

adapt:  HPF -> Fortran 90 translator (version 0.1)

Usage:
        adapt  [-d <directory>]  [-o <file>]  [-cfxhw]  [-s <number>]
               [-p [1234567] <number> <number>... ]  <file>  [[-fx] <file>]...

(where [] encloses an optional item, []... encloses an optional item
that may be repeated, and <..> denotes a name or number supplied by the user.)

The options are:

  -d <directory>
        If there is no '-o' option (to specify an explicit output
        path and filename)  output files are stored in directory
        <directory> rather than the current directory.

  -o <file>
        Specifies the output filename.

  -c    Compile code to check that temporary arrays do not exceed the
        amount of memory reserved for them (which can be specified
        by the '-s' option).

  -f    Free source form input (i.e. the new Fortran 90 source form).

  -x    Fixed source form input (i.e. Fortran 77 source form).
        (Without either the '-f' or '-x' options, the default is
        free source form for filenames ending ".f90", ".h90" or
        ".hpf90" and fixed source form otherwise).

  -h    Give help information.

  -w    Suppress warnings.

  -s <number>
        The amount of memory reserved for temporary arrays, in words.
        (Default 250000).

  -p1 <n1>
  -p2 <n1> <n2>
  ...
  -p7 <n1> <n2> ... <n7>
        These options specify the shapes of default HPF abstract
        processor arrays for use as distribution targets in HPF
        DISTRIBUTE directives without an ONTO clause  (i.e. without
        an explicit target processor arrangement).  For example:
                -p2 5 6
        specifies that the default 2-dimensional processor array has
        shape (5,6).  If the '-pN' option is not specified, the default
        shape for an N-dimensional processor array is:
                (/ 1, i=1,N /)
        that is, it contains just one processor!

The options may be specified individually or in groups, in any order
and either case.

-----------------------------------------------------------------*/
/*}}}*/
{
	fputs ("\nadapt:  HPF -> Fortran 90 translator (version 0.1)\n\nUsage:\n        adapt  [-d <directory>]  [-o <file>]  [-cfxhw]  [-s <number>]\n               [-p [1234567] <number> <number>... ]  <file>  [[-fx] <file>]...\n\n(where [] encloses an optional item, []... encloses an optional item\nthat may be repeated, and <..> denotes a name or number supplied by the user.)\n\nThe options are:\n\n  -d <directory>\n        If there is no \'-o\' option (to specify an explicit output\n        path and filename)  output files are stored in directory\n        <directory> rather than the current directory.\n\n  -o <file>\n        Specifies the output filename.\n\n  -c    Compile code to check that temporary arrays do not exceed the\n        amount of memory reserved for them (which can be specified\n        by the \'-s\' option).\n\n  -f    Free source form input (i.e. the new Fortran 90 source form).\n\n  -x    Fixed source form input (i.e. Fortran 77 source form).\n        (Without either the \'-f\' or \'-x\' options, the default is\n        free source form for filenames ending \".f90\", \".h90\" or\n        \".hpf90\" and fixed source form otherwise).\n\n  -h    Give help information.\n\n  -w    Suppress warnings.\n\n  -s <number>\n        The amount of memory reserved for temporary arrays, in words.\n        (Default 250000).\n\n  -p1 <n1>\n  -p2 <n1> <n2>\n  ...\n  -p7 <n1> <n2> ... <n7>\n        These options specify the shapes of default HPF abstract\n        processor arrays for use as distribution targets in HPF\n        DISTRIBUTE directives without an ONTO clause  (i.e. without\n        an explicit target processor arrangement).  For example:\n                -p2 5 6\n        specifies that the default 2-dimensional processor array has\n        shape (5,6).  If the \'-pN\' option is not specified, the default\n        shape for an N-dimensional processor array is:\n                (/ 1, i=1,N /)\n        that is, it contains just one processor!\n\nThe options may be specified individually or in groups, in any order\nand either case.\n\n",
	        stderr);
}
/*}}}*/

#endif

/*{{{  init_keytab ()*/
init_keytab ()
/*----------*/

/*------------------------------------------------------*/
/*    Link keywords into keyword hash table ('keytab')	*/
/*------------------------------------------------------*/
{
	register Keyword  *kp;
	register unsigned  hashval;

	for (kp = keywords;  kp->name;  kp++) {
		hashval = hash (kp->name);
		kp->next = keytab [hashval];
		keytab [hashval] = kp;
	}
}
/*}}}*/

/*{{{  int  yylex () -- done HPF*/
int yylex ()
/*---------*/
{
	register char  *ptr;
	register int    c;
	static bool  after_stmt_separator = FALSE,
	             need_next_stmt  = TRUE,
	             lookfor_complex = TRUE;
	int  token;

	/*------------------------------------------------------*/
	/* Check whether we need to input a new stmt, or are at	*/
	/* the start of a new stmt following a ';'.		*/
	/*	The following errmsgs are given for error	*/
	/* conditions at stmt boundaries:			*/
	/* -- stmt starts with ';'				*/
	/*	"no statement for ';' to terminate"		*/
	/* -- label without stmt (or followed by ';')		*/
	/*	"no statement following label '%d'"		*/
	/* -- syntax error at start of stmt			*/
	/*	"syntax error at start of stmt, at token '%s'"	*/
	/* -- stmt ends ';'	-- OK!				*/
	/*------------------------------------------------------*/
start :
	while (after_stmt_separator) {
		/*{{{  read label (if any) and set 'end_tk'*/
		
		after_stmt_separator = FALSE;
		
		/*--------------------------------------------------------------*/
		/* Skip successive blanks and statement separators (';').	*/
		/*--------------------------------------------------------------*/
		for (ptr = end_tk+1;  ((c = *ptr) == ' ' || c == '\t' || c == ';');  ptr++)
			;
		
		if (! c)
			/*------------------------------------------------------*/
			/* No statement follows the statement separator (which	*/
			/* is allowed!).  Read the next stmt.			*/
			/*------------------------------------------------------*/
			need_next_stmt = TRUE;
		
		else {
			/*------------------------------------------------------*/
			/* 'analyse_stmt ()' reads the stmt label (if any),	*/
			/* scans enough of the stmt to determine its type	*/
			/* (setting 'nwords_to_keyword' accordingly), and	*/
			/* returns a pointer to the start of the statement body	*/
			/* (after skipping over the label, if any).		*/
			/*	If the stmt body is empty (i.e. there is only	*/
			/* a label), it returns a pointer to the ';' terminator	*/
			/* if there is one, or NULL otherwise.  In these cases	*/
			/* it also gives an errmsg, so just go ahead and read	*/
			/* the next stmt.					*/
			/*------------------------------------------------------*/
			start_tk = ptr;			/* for errmsgs in 'analyse_stmt()' */
		
			stmt_startp = end_tk = analyse_stmt (start_tk);
							/* returns pointer to start */
							/* of stmt body, or ';'     */
			if (! end_tk)
				need_next_stmt = TRUE;
			else if (*end_tk == ';')
				after_stmt_separator = TRUE;
		
			/* ...else (after_stmt_separator = need_next_stmt = FALSE)... */
		
			else if (in_HPF_dir)
				return  TK_HPF;		/* signalling start of HPF directive */
		}
		/*}}}*/
	}
	while (need_next_stmt) {
		/*{{{  input next stmt, read its label (if any), and set 'end_tk'*/
		
		while ( ! (start_tk = get_stmt ()) ) {		/* end-of-file */
			if ( ! (ifp = close_file ()) )		/* file wasn't 'INCLUDE'd */
				return  TK_EOF;			/* end of input! */
		}
		
		/*--------------------------------------------------------------*/
		/* 'analyse_stmt ()' reads the stmt label (if any), scans	*/
		/* enough of the stmt to determine its type (setting		*/
		/* 'nwords_to_keyword' accordingly), and returns a pointer to	*/
		/* the start of the statement body (after skipping over the	*/
		/* label, if any).						*/
		/*	If the stmt body is empty (i.e. there is only a label),	*/
		/* it returns a pointer to the ';' terminator if there is one,	*/
		/* or NULL otherwise.  In these cases it also gives an errmsg,	*/
		/* so just go ahead and read the next stmt.			*/
		/*--------------------------------------------------------------*/
		
		if (end_tk = analyse_stmt (start_tk)) {	   /* returns pointer to start */
							   /* of stmt body, or ';'     */
			need_next_stmt = FALSE;
			stmt_startp    = end_tk;
			if (*end_tk == ';') {
				after_stmt_separator = TRUE;
				goto  start;
			}
			if (in_HPF_dir)
				return  TK_HPF;		/* signalling start of HPF directive */
		}
		/*}}}*/
	}

	/*--------------------*/
	/* get next non-white */
	/*--------------------*/
	ptr = end_tk;
	get_nonwhite_from_here (c, ptr);

	start_tk = ptr;		/* global ptr for 'read' & 'errmsg' functions */

	/*------------------------------------------------------*/
	/*	Now (c == *ptr) is the first char of the next	*/
	/* token in the input line.  Scan the token and return	*/
	/* its value.  Also, set 'end_tk' to point to the char	*/
	/* after the token, in order to save this pointer for	*/
	/* the next call to 'yylex' and for 'yyerror ()'.	*/
	/* ('end_tk' is set either in a 'read_' function, if	*/
	/* used, or explicitly otherwise.)			*/
	/*------------------------------------------------------*/
	/*{{{  read the next token*/
	if (isalpha (c)) {
		/*{{{  identifier, BOZ literal or string starting with kind param*/
		/*----------------------*/
		/*  Identifier		*/
		/*----------------------*/
		token = read_word_or_boz_or_kind_string (start_tk);	/* sets 'end_tk' */
		lookfor_complex = FALSE;
		/*}}}*/
	}
	else if (isdigit (c)) {
		/*{{{  number starting with digit, or string starting with kind param*/
		/*------------------------------*/
		/*  Number starting with digit	*/
		/*------------------------------*/
		
		token = read_num_or_kind_string (start_tk, TRUE);	/* sets 'end_tk' */
		lookfor_complex = FALSE;
		/*}}}*/
	}
	else {
		ptr++;			/* to point to char after 'c' */
	
		switch (c) {
		/*{{{  number or identifier starting with '.'*/
		/*----------------------------------------------*/
		/*  Number or identifier starting with '.'	*/
		/*----------------------------------------------*/
		
		case  '.' :
			if (isdigit (*ptr)) {
				token = read_num_or_kind_string (start_tk, TRUE);
									/* this sets 'end_tk'*/
				lookfor_complex = FALSE;
			}
			else if (isalpha (*ptr)) {
				token = read_dotted_word (start_tk);	/* this sets 'end_tk'*/
				lookfor_complex = TRUE;
			}
			else {
				token = c;
				end_tk = ptr;
			}
			break;
		/*}}}*/
		/*{{{  string*/
		/*--------------*/
		/*  String	*/
		/*--------------*/
		
		case  '\'' :
		case  '\"' :
			token = read_string (start_tk);		/* sets global ptr 'end_tk' */
			lookfor_complex = FALSE;
			break;
		/*}}}*/
		/*{{{  ( or (/*/
		/*--------------*/
		/*  (  or  (/	*/
		/*--------------*/
		
		case '(' :
			if (*ptr == '/') {
				token = TK_OPEN_AC;
				end_tk = ptr + 1;
			}
			else if (lookfor_complex)
				token = read_complex_num_or_paren (start_tk);
								    /* this sets 'end_tk' */
			else {
				token = c;
				end_tk = ptr;
			}
		
			if (token == '(')
				nparens++;
			lookfor_complex = TRUE;
			break;
		
		
		/*}}}*/
		/*{{{  * or ***/
		/*--------------*/
		/*  *  or  **	*/
		/*--------------*/
		
		case  '*' :
			switch (*ptr) {
			case '*' :
				token = TK_POWER_OP;
				end_tk = ptr + 1;
				break;
		
			default :
				token = c;
				end_tk = ptr;
				break;
			}
			lookfor_complex = TRUE;
			break;
		
		/*}}}*/
		/*{{{  / or /) or // or /=*/
		/*------------------------------*/
		/*  /  or  /)  or  //  or  /=	*/
		/*------------------------------*/
		
		case  '/' :
			switch (*ptr) {
			case ')' :
				token = TK_CLOSE_AC;
				end_tk = ptr + 1;
				lookfor_complex = FALSE;
				break;
		
			case '/' :
				token = TK_CONCAT_OP;
				end_tk = ptr + 1;
				lookfor_complex = TRUE;
				break;
		
			case '=' :
				token = TK_REL_OP;   yylval.str = ".NE.";
				end_tk = ptr + 1;
				lookfor_complex = TRUE;
				break;
		
			default :
				token = c;
				end_tk = ptr;
				lookfor_complex = TRUE;
				break;
			}
			break;
		/*}}}*/
		/*{{{  : or ::*/
		/*--------------*/
		/*  :  or  ::	*/
		/*--------------*/
		
		case  ':' :
			switch (*ptr) {
			case ':' :
				token = TK_DBL_COLON;
				end_tk = ptr + 1;
				lookfor_complex = FALSE;
				break;
		
			default :
				token = c;
				end_tk = ptr;
				lookfor_complex = TRUE;
				break;
			}
			break;
		/*}}}*/
		/*{{{  < or <=*/
		/*--------------*/
		/*  <  or  <=	*/
		/*--------------*/
		
		case  '<' :
			switch (*ptr) {
			case '=' :
				yylval.str = ".LE.";
				end_tk = ptr + 1;
				break;
		
			default :
				yylval.str = ".LT.";
				end_tk = ptr;
				break;
			}	
			token = TK_REL_OP;
			lookfor_complex = TRUE;
			break;
		/*}}}*/
		/*{{{  = or == or =>*/
		/*----------------------*/
		/*  =  or  ==  or  =>	*/
		/*----------------------*/
		
		case  '=' :
			switch (*ptr) {
			case '=' :
				token = TK_REL_OP;   yylval.str = ".EQ.";
				end_tk = ptr + 1;
				lookfor_complex = TRUE;
				break;
		
			case '>' :
				token = TK_PTR_ASSIGN;
				end_tk = ptr + 1;
				lookfor_complex = FALSE;
				break;
		
			default :
				token = c;
				end_tk = ptr;
				lookfor_complex = TRUE;
				break;
			}
			break;
		/*}}}*/
		/*{{{  > or >=*/
		/*--------------*/
		/*  >  or  >=	*/
		/*--------------*/
		
		case  '>' :
			switch (*ptr) {
			case '=' :
				yylval.str = ".GE.";
				end_tk = ptr + 1;
				break;
		
			default :
				yylval.str = ".GT.";
				end_tk = ptr;
				break;
			}
			token = TK_REL_OP;
			lookfor_complex = TRUE;
			break;
		/*}}}*/
		/*{{{  \n or ;  (end of statement)*/
		/*------------------------------------------------------*/
		/*  End of statement ('\n' or statement separator ';')	*/
		/*------------------------------------------------------*/
		
		case  '\0' :	/*--- i.e '\n' in input ---*/
		case  ';' :
		
			token = '\n';
			end_tk = start_tk;	/* so 'yyerror ()' knows we're at EOL,	*/
						/* & 'skip_rest_of_stmt()' starts here!	*/
			if (! c)
				need_next_stmt = TRUE;
			else
				after_stmt_separator = TRUE;
			lookfor_complex = FALSE;
			break;
		/*}}}*/
		/*{{{  any other allowed single character, viz:  % ) + - , _*/
		/*--------------------------------------*/
		/*  Any other allowed single character	*/
		/*--------------------------------------*/
		case  '%' :
		case  ')' :
		case  '+' :
		case  '-' :
		case  ',' :
		case  '_' :	/* This is *illegal* except in names, but we don't */
				/* want to report it as such (as in next section). */
			token = c;
			end_tk = ptr;
			if (token == ')')
				nparens--;
		
			lookfor_complex = TRUE;
			break;
		/*}}}*/
		/*{{{  any other (illegal) character*/
		default :
			/*--------------------------------------------------------
			This reports illegal characters.
				Of the printable ones, none are in the 'Fortran
			character set' except for '?' and '$', which have no defined
			use in Fortran stmts at the moment.  (Note, any char that *does*
			have a defined use but cannot start a token, e.g. '_', is
			filtered out above, so it's reported as a "syntax error"
			rather than as an "illegal character".)
				The hex value is given for non-printing ones.
			(Currently '\n' and '\t' are the only control chars we permit!).
		
			N.B.
			----
			(1) Sun-OS 'f77' accepts (at least) '^M', which is part of '\n'
			in DOS files but is extraneous in Unix files.  Perhaps it would
			be nice to accept these, but I'm not sure of a good strategy for
			what to accept and what to reject.  'NAG f90' rejects everything,
			as we do!
		
			(2) Here we rely on 'yyerror()' being called, otherwise
			'suppress_yyerror_msg' won't be reset (so the next syntax
			error won't be reported!).
			--------------------------------------------------------*/
			token = c;
			end_tk = ptr;
			if (isprint (c))
				errmsg ("illegal character \'%c\' in statement", c);
			else
				/*----------------------------------------------*/
				/* N.B. We cast 'c' to an 'unsigned char' and	*/
				/* then 'unsigned int' to zero higher order	*/
				/* bits (i.e. to undo any sign extension that	*/
				/* may occur).					*/
				/*----------------------------------------------*/
				errmsg ("illegal character (hex value \'%X\') in statement",
					 ((unsigned int) (unsigned char) c) );
			suppress_yyerror_msg = TRUE;
			break;
		/*}}}*/
		}
	}
	/*}}}*/

	return token;
}
/*}}}*/

/*{{{  <<< line handling functions >>>*/
/*{{{  COMMENT -- handling lines exceeding limits for length / continuations*/
/*------------------------------------------------------------------------
Handling lines that exceed the limits on length / no. of continuations
----------------------------------------------------------------------
The limits are:

Free:   132 chars/line;  39 continuation lines (viz. 5280 chars total)
Fixed:   72 chars/line;  19 continuation lines.

The following messages are output and actions taken if these limits are
exceeded:

Free source form:
-----------------
For a STATEMENT LINE exceeding 132 chars:

	errmsg ("line >132 characters long - the remainder
		(namely 'ABCDE') is ignored")

... OR, if there are more than 5 extra characters, the errmsg ends...

		(starting 'ABCDE') is ignored")

Trailing whitespace is ignored in this count.

For a stmt exceeding 39 continuation lines:

	errmsg ("more than 39 continuation lines - the remainder are ignored")

For a COMMENT or trailing comment that makes a line longer than 132 chars:

	warning ("comment line >132 characters long")

but we process and output all of the comment (since it's annoying to
reject a program just because the comments are too long!)

Fixed source form:
-----------------
The corresponding messages for STATEMENTs are:

	warning ("statement extends beyond column 72 - the remainder
		(namely 'ABCDE') is ignored")
or
		(starting 'ABCDE') is ignored")

	errmsg ("more than 19 continuation lines - the remainder are ignored")

THERE ARE NO ERROR MESSAGES FOR COMMENTS -- DON'T CHECK THEIR LENGTH!!

There is an extra message if '\t' is used in cols 1-6:

	warning ("using TAB to reach statement field is non-standard")
------------------------------------------------------------------------*/
/*}}}*/
/*{{{  COMMENT -- fixed form:  treatment of errors in cols 1-6!!*/
/*----------------------------------------------------------------
Contents of cols 1-6
~~~~~~~~~~~~~~~~~~~~
Normal cases:
~~~~~~~~~~~~~
cols 1-5    col 6         example          interpretation
--------    -----         -------          --------------
<     >     cont char     <     *XYZ>      continuation stmt [1]

<     >	    ' '           <      XYZ>      Comment, blank line or new stmt
                                              (must read further to decide)
<     >     '0'           <     0   >      New stmt

label       ' ' or '0'    <10000 XYZ>      Labelled new stmt
  "        '\t' col<=6    <100 '\t'>            "     "   "
  "        '\n' col<=6    <100 '\n'>            "     "   "  [2]
N.B.
----
[1] If there's a syntax error in the line, perhaps errmsg should mention
	that it's a continuation line (because maybe it's supposed
	to be a new stmt but was started in the wrong column!)
[2] If this stmt isn't continued, or is otherwise empty, then
	'analyse_line ()' gives the message "no stmt following label",
	so the line-reading functions needn't report this error.

Error cases:
~~~~~~~~~~~~
In the following cases, skip the line and assume that the prev stmt
line *isn't* a continuation:

cols 1-5    col 6         example          error message
--------    -----         -------          -------------
nondigits   anything      <1000* XYZ"      [2]
    "      '\n' col<6     <ABC '\n'>        "

<     >     non-print     <     '\1'XYZ>   [3]
or label

label       cont char     "1000 *XYZ"      [4]

Errmsgs:
--------
[2] errmsg ("nondigit \'%c\' in label (columns 1-5) - line skipped");
 OR errmsg ("nondigit (hex value \'%X\') in label (columns 1-5) - line skipped");
[3] errmsg ("illegal character (hex value \'%X\') in column 6 - line skipped");
[4] errmsg ("line has both label and continuation character (\'%c\' in
             column 6) - line skipped");

N.B:  The line reading functions don't check whether the label is >0,
as we choose not to skip the statement for that reason alone.
If the label is 0, an errmsg will be reported by 'analyse_stmt ()'.

Summary
~~~~~~~
-- Only a syntactically correct continuation is recognised as such;
	i.e only one that has cols 1-5 blank and col 6 non-blank.
-- As soon as we see non-blanks in cols 1-5 or 7-72, continuation = FALSE;
-- Input at least up to column 6 (or up to '\n' or '!' if earlier)
	even if we know earlier that  (continuation == FALSE), as this
	simplifies the processing.
------------------------------------------------------------------*/
/*}}}*/
/*{{{  COMMENT -- fixed form:  contents of 'next_label []'*/
/*----------------------------------------------------------------
	On return from 'f77_stmt_continued ()', 'next_label' is
always terminated by '\0'.  It contains a VERBATIM copy of the
start of the next statement line, as follows:

	if (EOF)
		'\0';
	else if (the first 6 cols are blank) {
		the 1st 6 cols (i.e. ' ''s and possibly '\t'),
		followed immediately by the first non-white char
		found in cols 7-72 of the input line
		(ignoring leading whitespace).
	}
	else {
		*-------------------------------------------------------*
		* there is at least one non-white in the label or col 6 *
		*-------------------------------------------------------*
		the 1st 6 cols
		--OR--
		up to and including a '\n' or '!' if before col 7.
	}

We keep a verbatim copy because cols 1-6 are checked, and errmsgs
generated if necessary, later on by 'get_f77_stmt_line ()'.
----------------------------------------------------------------*/
/*}}}*/


/*{{{  char *get_free_form_stmt ()  -- DONE HPF*/
char *get_free_form_stmt ()
/*-----------------------*/
/*--------------------------------------------------------------
	Reads the next *free format* statement, including continuation
lines, into 'textbuf', terminated by '\0' rather than '\n'.
If at EOF, '\0' is put at the start of 'textbuf'.  The input text
is copied verbatim (e.g. it isn't converted to upper case, and
sequences of whitespace characters are preserved) so as not to
corrupt Hollerith strings, which cannot be readily identified at
this stage.
	The function returns 'textbuf', or NULL if at EOF.
It *doesn't* set the global pointers 'start_tk' or 'end_tk'.
	Obviously, in order to reach the next statement or
continuation line, intervening blank and comment lines are read
and processed (i.e. output).  Also, any trailing comment at the
end of a statement or continuation line is processed.  A warning
is given if commentary causes a line to exceed MAX_LINELEN characters,
and an error message is given if statement text exceeds MAX_LINELEN
chars or there is illegal text after a contn char ('&').  Also,
an errmsg is given if there are > MAX_CONTNS continuation lines,
and the rest are skipped (see below).  Note that such 'lexical'
warnings and errmsgs for continuation lines may cause messages
to have out-of-sequence line numbers, as a subsequent 'syntax'
or 'semantic' message may refer to an earlier line of the statement
than a 'lexical' message.
--------------------------------------------------------------*/
/*{{{  *** Handling of > MAX_CONTN lines ****/
/*----------------------------------------------------------------
Handling of > MAX_CONTN lines
-----------------------------
	If there are > MAX_CONTN continuation lines, the function
returns after reading just the first MAX_CONTN continuation lines
into 'textbuf' -- as though there are no more (perhaps causing a
syntax error to be generated by the parser).  This situation is
signalled by setting the static variable 'skip_contns' to TRUE.
	The *next time* this function is called, successive
continuation lines are skipped until the next new stmt is reached,
which is then input as normal.
	Superfluous continuation lines are handled this way (rather
than being skipped at the end of the previous fn call, which handles
the statement to which they belong),  so that they can be written at
the *start* of 'textbuf' and then overwritten (rather than being
written in 'textbuf' after the first MAX_CONTN contn lines, which
may overflow 'textbuf'!)
----------------------------------------------------------------*/
/*}}}*/
{
	/*{{{  decls*/
	
	register int    c,  col;
	register char  *ptr;	/* posn in 'textbuf' where next char written */
	
	static bool  in_string   = FALSE,
	             skip_contns = FALSE;
	
	bool   HPF_dir_line,	/* if TRUE, the *current* line has the HPF directive
				      prefix.  (Cf. 'in_HPF_dir' means that the 1st
				      line of this stmt (which may be continued)
				      has an HPF directive prefix.)  */
	       expect_contn,	/* if TRUE, expect a continuation line */
	       found_stmt,	/* if TRUE, found the first line of a stmt / directive*/
	       stmt_line;	/* if TRUE, the current line is part of a stmt / 
	       			    directive (i.e. *not* a blank or comment line) */
	
	int    n_contns,	/* the number of continuation lines so far,
				   including current one. */
	       string_delim;	/* the current string delimiter ('\'' or '\"') */
	/*}}}*/

	/*{{{  initialisation*/
	
	in_HPF_dir = FALSE;
	use_current_lineno = TRUE;	/* for errmsgs */
	
	ptr = textbuf;
	found_stmt = FALSE;
	n_contns   = 0;		/* ...even if 'skip_contns == TRUE', as we don't  */
				/* count continuation lines in that circumstance! */
	if (skip_contns)
		expect_contn = TRUE;
	else {
		expect_contn = FALSE;
		in_string    = FALSE;
	}
	/*}}}*/
	while ( !found_stmt || expect_contn) {	/* read next i/p line */
		col = 0;		/* column no. of last char read */
		lineno++;		/* undo this if at EOF! */
		stmt_line = HPF_dir_line = FALSE;
					/* passed to 'process_comment()' */
find_first_nonwhite:
		/*{{{  scan to 1st nonwhite to see if begins a stmt*/
		/*--------------------------------------------------------------*/
		/* Skip initial whitespace to find the first nonwhite char 'c'.	*/
		/* (Don't worry about its column number!)			*/
		/*--------------------------------------------------------------*/
		while ((c = readchar (col)) == ' ' || c == '\t')
			;
		
		/*--------------------------------------------------------------*/
		/* Now 'c' is the first nonwhite char, which appeared in column	*/
		/* 'col' (which may be > MAX_LINELEN).				*/
		/*--------------------------------------------------------------*/
		switch (c) {
		case  '\n' :
			continue;		/* read next line */
		
		case  '!' :
			/*------------------------------------------------------*/
			/* The 2nd arg of the next call is equivalent to:	*/
			/*	COMMENT_LINE      -- if (HPF_dir_line == FALSE)	*/
			/*	TRAILING_COMMENT  -- if (HPF_dir_line == TRUE)	*/
			/*------------------------------------------------------*/
			if (HPF_dir_line = process_comment (col, HPF_dir_line)) {
				col += 4;		/* ...because the above read "HPF$" */
				goto  find_first_nonwhite;  /* find 1st nonwhite after prefix */
			}
			else
				continue;		/* a comment;  read next line */
		
		case  EOF :
			/*{{{  set *ptr = '\0';  return NULL if no stmt or 'textbuf' otherwise*/
			/*-------------*/
			/* End-of-file */
			/*-------------*/
			*ptr = '\0';		/* in case (ptr == textbuf) */
			if (col == 1)
				lineno--;	/* to undo the earlier increment of 'lineno'! */
			if (expect_contn)
				errmsg ("expected a continuation line - found end-of-file");
			
			goto  finished;
			/*}}}*/
		}
		
		if (col > MAX_LINELEN) {
			/*{{{  whitespace up to and possibly beyond col MAX_LINELEN!*/
			/*--------------------------------------------------------------*/
			/* Whitespace up to and possibly beyond col MAX_LINELEN!	*/
			/*--------------------------------------------------------------*/
			goto  line_too_long;
			/*}}}*/
		}
		else if (expect_contn) {	/* expecting a continuation line */
			/*{{{  give errmsg for isolated '&', or if string doesn't start with '&'*/
			if (! in_HPF_dir  &&  HPF_dir_line) {
				/*------------------------------------------------------*/
				/* Found an HPF directive line when expecting a		*/
				/* continuation of a stmt.  Give an errmsg and skip	*/
				/* the directive.					*/
				/*------------------------------------------------------*/
				errmsg ("an HPF directive must not appear within a continued statement");
				goto  skip_line;
			}
			else if (in_HPF_dir  &&  ! HPF_dir_line) {
				/*------------------------------------------------------*/
				/* Expected an HPF directive to be continued, but found	*/
				/* an ordinary stmt.  Give an errmsg, but treat it as a	*/
				/* continuation of the directive!			*/
				/*------------------------------------------------------*/
				errmsg ("a continuation of an HPF directive must start with the directive prefix \'!HPF$\'");
				HPF_dir_line = TRUE;		/* pretend it is, for convenience! */
			}
			
			expect_contn = FALSE;	/* found continuation line */
			n_contns++;		/* the number of continuation lines so far */
			found_stmt   = TRUE;
			stmt_line    = TRUE;
			
			if (c == '&') {
				/*------------------------------------------------------*/
				/* Don't write the '&' - it's not part of the stmt!	*/
				/*------------------------------------------------------*/
				if (in_string) {
					c = readchar (col);
					goto  stmt_string;
				}
				else {
					/*{{{  check that '&' isn't isolated, and write ' ' if necessary*/
					/*--------------------------------------------------------------*/
					/* Continuation line starting with '&', and not in a string.	*/
					/*	We must check that there is another non-blank char (not	*/
					/* '!') on the line, at or before column MAX_LINELEN, as it's	*/
					/* illegal for a line to contain only '&' (as it's ambiguous	*/
					/* whether this is an initial '&' -- meaning end-of-stmt, or a	*/
					/* final '&' -- meaning stmt continued).			*/
					/*--------------------------------------------------------------*/
					
					/*--------------------------------------------------------------*/
					/* Copy input whitespace to 'textbuf' until the next nonwhite	*/
					/* character after '&' ('c') is found, stopping if column	*/
					/* MAX_LINELEN is exceeded.					*/
					/*--------------------------------------------------------------*/
					while (((c = readchar (col)) == ' ' || c == '\t')  &&  (col <= MAX_LINELEN))
						*ptr++ = c;
					
					/*--------------------------------------------------------------*/
					/* Now 'c' is the next nonwhite char; col may be > MAX_LINELEN.	*/
					/*--------------------------------------------------------------*/
					if (c == '\n' || c == EOF || c == '!' || col > MAX_LINELEN) {
						errmsg ("isolated \'&\' is illegal - end of statement assumed!");
						switch (c) {
						case '\n' :
						case EOF :
							goto  end_of_line;
					
						case '!' :
							process_comment (col, (stmt_line || HPF_dir_line));
							goto  end_of_line;
					
						default :
							goto  line_too_long;
						}
					}
					
					/*------------------------------------------------------*/
					/* ... otherwise, process 'c' in the normal stmt loop!	*/
					/*------------------------------------------------------*/
					/*}}}*/
				}
			}
			else if (in_string)
				errmsg ("a continuation of a string must start with an \'&\'");
			
			else
				/*------------------------------------------------------*/
				/* must ensure that there's a ' ' after the last token	*/
				/*------------------------------------------------------*/
				if ( *(ptr-1) != ' ' )
					*ptr++ = ' ';
			/*}}}*/
		}
		else if (c == '&') {		/* NOT expecting a continuation line */
			if (HPF_dir_line)
				errmsg ("unexpected continuation of HPF directive");
			else
				errmsg ("unexpected continuation line");
			goto  skip_line;
		}
		
		/*--------------------------------------------------------------*/
		/* If we've got this far then we've found the next stmt or HPF	*/
		/* directive line!  'c' is it's first significant  (i.e.	*/
		/* non-white, non-'&', non-'!', etc) character (after the "HPF$"*/
		/* prefix if this is an HPF directive), and is processed by the	*/
		/* main statement loop (to handle special cases!)		*/
		/*--------------------------------------------------------------*/
		in_HPF_dir = HPF_dir_line;
		found_stmt = TRUE;
		stmt_line  = TRUE;
		if (in_string)
			goto  stmt_string;
		else
			goto  stmt_text;
		/*}}}*/
stmt_text:
		/*{{{  stmt -- not in string*/
		/*--------------------------------------------------------------*/
		/* Process statement text.  On entering this loop we've already	*/
		/* input, but not processed, a non-white character 'c' in	*/
		/* column 'col'.						*/
		/*--------------------------------------------------------------*/
		while (col <= MAX_LINELEN) {
			switch (c) {
			case  '\n' :
			case EOF :
				goto  end_of_line;
				
			case '!' :
				process_comment (col, (stmt_line || HPF_dir_line));	/* trailing comment */
				goto  end_of_line;
		
			case '\'' :
			case '\"' :
				*ptr++ = string_delim = c;
				in_string = TRUE;
				c = readchar (col);
				goto  stmt_string;
		
			case '&' :
				/*------------------------------------------------------*/
				/* continuation char ('&') -- don't write it to textbuf */
				/*------------------------------------------------------*/
				expect_contn = TRUE;
				c = readchar (col);
				goto  after_ampersand;
		
			default :
				*ptr++ = c;
				break;
			}
			c = readchar (col);
		}	
		goto  line_too_long;
		
		/*}}}*/
stmt_string:
		/*{{{  stmt -- in string*/
		/*--------------------------------------------------------------*/
		/* Process a character string.  On entering this loop we've	*/
		/* already input, but not processed, a char 'c' in column 'col'.*/
		/*--------------------------------------------------------------*/
		while (col <= MAX_LINELEN) {
			if (c == '\n' || c == EOF)
				goto  end_of_line;
				
			*ptr++ = c;
			if (c == string_delim) {
				in_string = FALSE;
				c = readchar (col);
				goto  stmt_text;
			}
			else if (c == '&') {
				/*------------------------------------------------------*/
				/* This is only a continuation character if it's the	*/
				/* last nonwhite on the line (at least up to col. 133).	*/
				/* N.B. Don't goto 'after_ampersand', since we can't	*/
				/* complain of "illegal trailing text after an '&'" in	*/
				/* a string context!					*/
				/*------------------------------------------------------*/
				char  *ampersand_posn = (ptr - 1);
		
				while ( ((c = readchar (col)) == ' ' || c == '\t')  &&
				        (col <= MAX_LINELEN) )
					*ptr++ = c;
		
				if (c == '\n' || c == EOF || col > MAX_LINELEN) {
					/*----------------------------------------------*/
					/* Whitespace to end-of-line or up to and incl.	*/
					/* col MAX_LINELEN, so '&' *is* a continuation!	*/
					/*----------------------------------------------*/
					ptr = ampersand_posn;
					expect_contn = TRUE;
					if (c == '\n' || c == EOF)
						goto  end_of_line;
					else
						goto  line_too_long;
				}
				/*------------------------------------------------------*/
				/* ... else read & copied whitespace up to a nonwhite	*/
				/* char 'c' in col <= MAX_LINELEN, so '&' *isn't* a	*/
				/* continuation!  Process 'c' in the next loop iteration*/
				/*------------------------------------------------------*/
				continue;	/* already have next char 'c'! */
			}
			c = readchar (col);
		}	
		goto  line_too_long;
		
		/* N.B. Need to be careful in the 'line_too_long' section, as if '!' is the
		first nonwhite char seen after col 132, we don't want it to be interpreted
		as a comment (and just a warning given) -- comments aren't allowed when
		we're 'in_string'!! */
		/*}}}*/
after_ampersand:
		/*{{{  don't write to 'textbuf', and ignore string delimiters!*/
		/*--------------------------------------------------------------*/
		/* On entry, 'c' is the next char *after* a continuation char	*/
		/* ('&').  Don't write anything to 'textbuf'.  First, skip	*/
		/* blanks, then check whether the next nonwhite char is legal	*/
		/* after an '&' and take the appropriate action.		*/
		/* N.B. Only check the column no. if we find illegal trailing	*/
		/* text.							*/
		/*--------------------------------------------------------------*/
		while (c == ' ' || c == '\t')
			c = readchar (col);
		
		switch (c) {
		case '\n' : case EOF :
			goto  end_of_line;
		
		case '!' :
			process_comment (col, (stmt_line || HPF_dir_line));
			goto  end_of_line;
			/*------------------------------------------------------*/
			/* N.B. We still process comments even if they start	*/
			/* in col >132.						*/
			/*------------------------------------------------------*/
		}
		
		/*--------------------------------------------------------------*/
		/* Now we know that an error has occurred (i.e. illegal text	*/
		/* after an an '&'), so just read and discard text until the	*/
		/* '\n' -- don't bother processing comments, and don't give an	*/
		/* errmsg if the line is too long!				*/
		/*--------------------------------------------------------------*/
		errmsg ("illegal text after a continuation character (\'&\')");
		goto  skip_line;
		/*}}}*/
line_too_long:
		/*{{{  skip rest of line, and give errmsg if text beyond col 132*/
		/*--------------------------------------------------------------*/
		/* Column > 132.						*/
		/* ------------							*/
		/* We already have the next char 'c' (which may be anything,	*/
		/* including '\n') in column 'col' > 132.			*/
		/*	Don't write anything to 'textbuf'.  Firstly, skip	*/
		/* whitespace, as we don't want to complain that the stmt is	*/
		/* too long simply because of trailing whitespace.		*/
		/*--------------------------------------------------------------*/
		while (c == ' ' || c == '\t')
			c = readchar (col);	/* (keep track of 'col' in case    */
						/* 'process_comment()' is called). */
		
		if (c == '\n' || c == EOF)
			goto  end_of_line;
		
		else if ((c == '!') && !in_string) {  /* comments aren't recognised in a string */
			process_comment (col, (stmt_line || HPF_dir_line));
			goto  end_of_line;
		}
		else {
			/*------------------------------------------------------*/
			/* Report the error!					*/
			/* N.B. Now that we know that an error has occurred,	*/
			/* don't bother looking for comments!			*/
			/*------------------------------------------------------*/
			char tempbuf [6];
			int  i = 0;
		
			/*------------------------------------------------------*/
			/* copy <=5 chars from the next nonwhite into 'tempbuf'	*/
			/*------------------------------------------------------*/
			while (c != '\n'  &&  c != EOF  &&  i < 5) {
				tempbuf [i++] = c;
				c = getc (ifp);
			}
			tempbuf [i] = '\0';
		
			/*------------------------------------------------------*/
			/* Give an errmsg (whose exect form depends on whether	*/
			/* there are any more nonwhite chars in the line, for	*/
			/* exactness!)						*/
			/*------------------------------------------------------*/
			while (c == ' ' || c == '\t')
				c = getc (ifp);
		
			errmsg ("line >%d characters long - the remainder (%s %q) is ignored",
			         MAX_LINELEN,
			         ((c == '\n' || c == EOF) ? "namely" : "starting"),  tempbuf);
		
			goto  skip_line;	/* N.B. 'c' may be '\n' or EOF! */
		}
		/*}}}*/
skip_line:
		/*{{{  just skip the rest of the i/p line*/
		/*--------------------------------------------------------------*/
		/* Just skip the rest of the input line.			*/
		/* N.B. The last input char 'c' may already be '\n' or EOF.	*/
		/*--------------------------------------------------------------*/
		while (c != '\n' && c != EOF)
			c = getc (ifp);
		
		goto  end_of_line;
		/*}}}*/
end_of_line:
		/*{{{  set 'stmt_endp/lineno', & reset 'skip_contn' if necessary.*/
		if (stmt_line) {
			/*------------------------------------------------------*/
			/* Set 'stmt_endp[]' and 'stmt_lineno[]', to record	*/
			/* the lineno of the text in this portion of 'textbuf'.	*/
			/* Only do this if in 'stmt_line', otherwise		*/
			/* 'stmt_lineno[]' will be overwritten by subsequent	*/
			/* blank and comment lines!				*/
			/*------------------------------------------------------*/
			stmt_endp [n_contns]   = ptr - 1;	/* the last char written */
			stmt_lineno [n_contns] = lineno;
		}
		
		if (skip_contns) {
			ptr = textbuf;	/* to overwrite the last contn. line read! */
			found_stmt = FALSE;
			n_contns   = 0;		/* we don't count these contn lines! */
			if (!expect_contn) {
				/*----------------------------------------------*/
				/* no more contns -- now we can read next stmt! */
				/*----------------------------------------------*/
				skip_contns = FALSE;
				in_string   = FALSE;
			}
		}
		else if (expect_contn  &&  (n_contns == MAX_CONTNS)) {
			/*------------------------------------------------------*/
			/* Already have the maximum allowed no. of continuation	*/
			/* lines!  We return the current form of the stmt, as	*/
			/* if it's finished, and skip the remaining continuation*/
			/* lines on the *next* call to this function, before	*/
			/* reading the next stmt.  (The reason for this		*/
			/* organisation is given above!)			*/
			/*------------------------------------------------------*/
			errmsg ("more than %d continuation lines - the remainder are ignored",
			         MAX_CONTNS);
			skip_contns = TRUE;
			goto  finished;
		}
		/*}}}*/
	}
finished:
	/*{{{  reset 'use_current_lineno', and return*/
	
	*ptr = '\0';
	stmt_endp [n_contns] = ptr;	/* replace (ptr-1) as the value of  */
					/* 'stmt_endp', to include the '\0' */
	use_current_lineno = FALSE;
	
	return  (found_stmt ? textbuf : NULL);	/* return NULL if EOF & no stmt */
	/*}}}*/
}
/*}}}*/

/*{{{  char *get_fixed_form_stmt () */
char *get_fixed_form_stmt ()
/*------------------------*/
/*{{{  DESCRIPTION*/
/*--------------------------------------------------------------
	Reads the next *fixed format* (i.e. F77-like) statement,
including continuation lines, into 'textbuf', terminated by '\0'
rather than '\n'.  If there's a statement label, it is checked
and installed by 'lookup_or_install_label ()'  (which sets the global
variables 'stmt_lp', 'stmt_label' & 'stmt_newlabel'), but *isn't*
copied into 'textbuf'.  The 'stmt_...' variables are initialised
to 0 if there isn't a label.  The input text is copied verbatim
(e.g. it is not converted to upper case, and sequences of multiple
whitespace characters are preserved) so as not to corrupt Hollerith
strings, which cannot be readily identified at this stage.

	The function returns 'textbuf', or NULL if at EOF.
It *doesn't* set the global pointers 'start_tk' or 'end_tk'.

	Before returning, the label field of the next stmt line
is read into a static array 'next_label []' and inspected to
determine whether this line is continued.  This is done by
'f77_stmt_continued()'.  In doing this, intervening blank and comment
lines are processed (i.e. output straight away), as are any trailing
comments on lines of the current stmt.  Checks are made of the
line-length (up to 72 chars) and of the validity of the label
and continuation fields (cols 1-6).  'f77_stmt_continued()' is
solely responsible for incrementing 'lineno', as it reads the start
of *all* lines.

N.B. If this function returns 'textbuf', then it's guaranteed that
either:
-- (textbuf [0] != '\0'), i.e. the statement isn't empty, or
-- (stmt_lp != NULL), i.e. there is a valid, non-duplicated, label,
	even though there is no statement following it.
--------------------------------------------------------------*/
/*}}}*/
{
	register char  *ptr;	 /* current posn of EOL ('\0') in 'textbuf' */
	bool   found_first_line; /* if TRUE, found 1st line of a stmt */
	int    n_contns;	 /* # contns so far, incl current one. */


	stmt_lp    = (Label *) NULL;
	stmt_label = stmt_newlabel = 0;
	in_HPF_dir = FALSE;

	use_current_lineno = TRUE;	/* global, for errmsgs and warnings */
	found_first_line   = FALSE;
	n_contns = 0;
	/*------------------------------------------------------*/
	/* Look for the next 'acceptable' first line of a stmt.	*/
	/*------------------------------------------------------*/
	while ( !found_first_line) {
		while (need_label) {
			/*{{{  get start of next (non-continuation) stmt line*/
			/*----------------------------------------------------------------
				'next_label []' *doesn't* contain the start of the next stmt.
			(Normally it contains its first 6 chars and, if they're blank,
			the first non-blank character before col 73).  This happens only
			at start of a file, or if the previous line was skipped because
			cols 1-6 were unintelligible.
				Note that in this case it's an error for the line to be
			a continuation (i.e. to have a char in column 6).  If it is, ignore
			it and continue reading until a non-continuation line is found!
			----------------------------------------------------------------*/
			if (need_label = f77_stmt_continued ()) {
				/*--------------------------------------------------------
				Next stmt line is a continuation -- discard it!
			
				N.B. In this case, 'next_label []' contains *exactly*
				<     X>, i.e. blanks in cols 1-5, a non-blank in col 6
				and '\0' in col 7.  It explicitly *doesn't* contain '\n',
				so it's safe to search for it here in the input.
				(If 'next_label []' *did* contain the '\n', then searching
				for the next one here would both skip a line unnecessarily
				and upset the line numbering, as 'lineno' is only incremented
				in 'f77_stmt_continued()'!)
				--------------------------------------------------------*/
				register int  c;
			
				errmsg ("unexpected continuation line");
				while ((c = getc (ifp)) != '\n'  &&  c != EOF)
					;
			}
			/*}}}*/
		}
		ptr = get_f77_stmt_line (textbuf);
		if (ptr > textbuf || stmt_lp) {
			/*{{{  found 1st line of stmt / HPF directive*/
			/*--------------------------------------------------------------*/
			/* Text was copied into 'textbuf' or a label was installed, so	*/
			/* we've found the first line of a statement or HPF directive.	*/
			/*--------------------------------------------------------------*/
			found_first_line = TRUE;
			
			stmt_endp [0]   = ptr - 1;	/* the last char written */
			stmt_lineno [0] = lineno;
			/*}}}*/
		}
		else if ( !ptr) {
			/*{{{  EOF;  return NULL*/
			/*----------------------------------------------------------------
			EOF, so return NULL.
				The following actions are performed at EOF:
			-- 'f77_stmt_continued()', which reads the stmt label field, sets
			(next_label [0]='\0') for EOF -- and in no other case;
			-- then, the next call to 'get_f77_stmt_line()' returns NULL
			if 'next_label [0] == '\0'.
			/*--------------------------------------------------------------*/
			
			return  NULL_str;
			/*}}}*/
		}
		else {	/* ptr == textbuf && !stmt_lp */
			/*{{{  line skipped as errors in cols 1-6, so try again*/
			/*----------------------------------------------------------------
			'get_f77_stmt_line()' wrote no text to 'textbuf' and didn't
			install a label.
				This means that either:
			
			-- 'f77_stmt_continued ()' found nonwhite chars in cols 1-6 of the
			line, but 'get_f77_stmt_line ()' found errors therein so the line
			was skipped, OR
			
			--'f77_stmt_continued ()' found an HPF directive prefix, signalling
			the start of an HPF directive, but there was no text in the
			directive body.
			
			Hence, start again.
				Note that in this circumstance we don't allow a continuation.
			----------------------------------------------------------------*/
			need_label = TRUE;
			/*}}}*/
		}
	}
	/*---------------------------*/
	/* Append continuation lines */
	/*---------------------------*/
	while (f77_stmt_continued ()) {
		if (++n_contns <= MAX_F77_CONTNS) {
			/*{{{  read line into 'textbuf'*/
			/*----------------------------------------------------------------
				Expecting a continuation line.  Note that, since
			'f77_stmt_continued()' has returned TRUE, we know that the label
			field is valid so the line won't be skipped!  Nor can it be EOF.
			However, it could be an empty continuation, so 'ptr' may not advance!
			----------------------------------------------------------------*/
			ptr = get_f77_stmt_line (ptr);
			
			stmt_endp [n_contns]   = ptr - 1;	/* the last char written */
			stmt_lineno [n_contns] = lineno;
			/*}}}*/
		}
		else {
			/*{{{  too many continuations; skip line*/
			/*----------------------------------------------------------------
			Too many continuation lines.  Give an errmsg for if this is the
			first such line, and skip to end of line.
			
			N.B. As above, since this is a continuation line, 'next_label []'
			contains *exactly*  "     X", i.e. blanks in cols 1-5, a non-blank
			in col 6, and '\0' in col 7.  In particular, it *doesn't contain
			'\n' -- so it's safe for this function to search for it in the input.
			(If 'next_label []' *did* contain the '\n', then searching for
			the next one here would upset the line numbering, as 'lineno' is
			only incremented in 'f77_stmt_continued()'!)
			----------------------------------------------------------------*/
			register int  c;
			
			if (n_contns == (MAX_F77_CONTNS + 1))
				errmsg ("more than %d continuation lines - the remainder are ignored",
				         MAX_F77_CONTNS);
			
			while ((c = getc (ifp)) != '\n'  &&  c != EOF)
				;
			/*}}}*/
		}
	}
	/*{{{  reset 'stmt_endp[]' & 'use_current_lineno'*/
	
	stmt_endp [n_contns] = ptr;	/* replace (ptr-1) as the value of */
					/* 'stmt_endp', to include the '\0' */
	use_current_lineno = FALSE;
	/*}}}*/

	return  textbuf;	/* if EOF, NULL was returned earlier */
	
}
/*}}}*/
/*{{{  bool  f77_stmt_continued () -- think it's ok*/
bool  f77_stmt_continued ()
/*-----------------------*/
/*{{{  DESCRIPTION*/
/*--------------------------------------------------------------
	This function reads enough of the input to determine
whether the last stmt line read by 'get_f77_stmt_line ()' will
be continued.  This means reading at least cols 1-6 of the next
line, and if they're blank, up to the first non-blank char of the
statement body (in cols 7-72).  If the line turns out to be a blank
or comment line, it is output and the above actions are repeated
for successive lines until the next 'statement line' is found.
	The function returns a boolean result indicating whether
the previous line is continued.
	The characters input from the stmt line are stored verbatim
in a buffer ('next_label'), terminated by '\0', and are later copied
to 'textbuf' by 'get_f77_stmt_line()' when it processes the next
line.  Note that if '\n' is read it's copied straight into
'next_label'; it's no good replacing it by '\0', because that's
the normal end-of-string marker (which here does NOT signify
end-of-line!).  If at EOF, 'next_label [0]' is set to '\0'.
(See the earlier COMMENT for a full description of the contents
of 'next_label[]'.)
	No error messages are written about the contents of the
stmt line -- again this is done by 'get_f77_stmt_line ()' when
it reads the whole line.
	As we've said, in order to reach the next stmt line,
intervening blank and comment lines are read and processed
(i.e. output).  A warning is generated only in one exceptional
case: if a line is blank up to col 72 and contains nonwhite chars
thereafter.  Such warnings will be out-of-sequence with any error
messages generated when the current statement is parsed!  However,
no other warnings are generated; in particular, there are none for
fixed-form comments (even if they are overlength!!).
	This function is solely responsible for incrementing
'lineno' in the case of fixed-form input.
/*--------------------------------------------------------------*/
/*}}}*/
{
	register char  *labelp;
	register int    c,  col;
	bool   label_field_occupied;


read_next_line :

	lineno++;		/* undo this if at EOF!!! */
	labelp = next_label;
	col = 0;		/* column no. of last char read */
	start_col = 0;		/* set non-0 only for a new unlabelled stmt,
				   whose 1st sig char is in col 'start_col' */
	label_field_occupied = FALSE;
		
	/*------------------------------------------------------*/
	/* scan cols 1-6, and return if we can determine from	*/
	/* this whether the next line is a continuation.	*/
	/*------------------------------------------------------*/
	/*{{{  1st char of line*/
	/*----------------------*/
	/* first char of line	*/
	/*----------------------*/
	switch (c = readchar (col)) {
	case  EOF :
		*labelp = '\0';		/* i.e. set  next_label [0] == '\0' */
		lineno--;		/* undoes earlier increment of 'lineno'! */
		return  FALSE;		/* not a continuation! */
		
	case  'c' :
	case  'C' :
	case  '*' :
	case  '!' :
		if ( ! process_comment (col, COMMENT_LINE))
			goto  read_next_line;	/* a comment, copied out */
	
		/*------------------------------------------------------*/
		/* ...else read "CHPF$" at the start of the next line,	*/
		/* signifying an HPF directive.  It's a continuation	*/
		/* directive if the next i/p character (i.e. in col 6)	*/
		/* is anything other than ' ', '0', '\t', '\n' or EOF!	*/
		/*------------------------------------------------------*/
		next_label [0] = 'C';		/* ...signifying an HPF directive */
	
		switch (c = readchar (col)) {	/* char in column 6 */
		case  '\n' :
		case  EOF  :
			goto  read_next_line;	/* an isolated HPF directive prefix
						   -- ignore it! */
		case  ' '  :
		case  '0'  :
		case  '\t' :
			start_col = 6;		/* col from which last char read! */
			return  FALSE;		/* not a continuation! */
	
		default :
			/*----------------------------------------------*/
			/* Found "CHPF$x" at start of next line, which	*/
			/* is an HPF continuation directive.  However,	*/
			/* only accept it as such if we're already in	*/
			/* an HPF directive.  Otherwise give an errmsg	*/
			/* and discard the line.			*/
			/*----------------------------------------------*/
			if (in_HPF_dir) {
				start_col = 6;	/* col from which last char read! */
				return  TRUE;	/* a continuation directive! */
			}
			else {
				errmsg ("unexpected continuation of HPF directive");
				while ((c = getc (ifp)) != '\n'  &&  c != EOF)
					;
				goto  read_next_line;
			}
		}
	}
	/*}}}*/
	/*{{{  process label & continuation char (cols 1-6)*/
	/*--------------------------------------------------------------*/
	/* Process label and continuation char (columns 1-6).		*/
	/* Note that if '\n', EOF or '!' appear in this field, they're	*/
	/* written into 'next_label []' (except that EOF -> '\n').	*/
	/*--------------------------------------------------------------*/	
	if (c == '\t')
		col = 6;
	*labelp++ = c;
	
	while ((col < 6)  &&  (c != '!')  &&  (c != '\n')) {	/* in label */
		/*----------------------------------*/
		/* this applies only to cols 1-5... */
		/*----------------------------------*/
		if ( ! (c == ' ' || c == '\t'))
			label_field_occupied = TRUE;
	
		/*----------------------------------*/
		/* this applies only to cols 2-6... */
		/*----------------------------------*/
		c = readchar (col);
		if (c == EOF)
			c = '\n';
		else if (c == '\t')
			col = 6;
	
		*labelp++ = c;
	}
	/*}}}*/
	/*{{{  col 6, or '!', '\t' or '\n'*/
	/*--------------------------------------------------------------*/
	/* column 6, or earlier due to a premature end of label if 'c'	*/
	/* is '!', '\t' or '\n'.  All chars read so far have been	*/
	/* written to 'next_label []' (including '!' and '\n').		*/
	/*--------------------------------------------------------------*/
	if (  label_field_occupied || c == '0' ||
	    ! (c == '\n' || c == '\t' || isprint (c)) ) {
		/*------------------------------------------------------*/
		/* ...last condition is TRUE if illegal char in col 6!	*/
		/*------------------------------------------------------*/
		*labelp = '\0';
		return  FALSE;		/* Not a contn. (New stmt or garbage!) */
	}
	
	switch (c) {	/* 'c' is char in col 6, or '\n', '\t' or '!' */
	case  '\n' :
		/*----------------------*/
		/* a blank line		*/
		/*----------------------*/
		goto  read_next_line;
	
	case  ' ' :
	case  '\t' :
		/*------------------------------------------------------*/
		/* it MAY BE a blank or comment line -- investigate	*/
		/* the statement body (cols 7-72).			*/
		/*------------------------------------------------------*/
		break;
	
	case  '!' :
		/*------------------------------------------------------*/
		/* If '!' is in column 6 it's a continuation char, so	*/
		/* return TRUE.  Otherwise it's either a comment line	*/
		/* (which is written out) or an HPF directive prefix.	*/
		/* In the latter case it *isn't* a continuation, so	*/
		/* return FALSE.					*/
		/*------------------------------------------------------*/
		if (col < 6) {
			if ( ! process_comment (col, COMMENT_LINE))
				goto  read_next_line;	/* a comment, copied out */
	
			next_label [0] = 'C';	/* ...signifying an HPF directive */
			start_col = col + 4;	/* col from which last char read */
			return  FALSE;		/* not a continuation */
		}
		/* ...else fall through to 'default' case... */
	default :
		/*------------------------------------------------------*/
		/* We must be in column 6, with columns 1-5 blank, and	*/
		/* 'c' is a valid continuation character.		*/
		/*------------------------------------------------------*/
		*labelp = '\0';
		if (in_HPF_dir) {
			errmsg ("unexpected continuation line");
			while ((c = getc (ifp)) != '\n'  &&  c != EOF)
				;
			goto  read_next_line;
		}
		else
			return  TRUE;	/* a continuation line! */
	}
			
	
	/*}}}*/

	/*------------------------------------------------------*/
	/* cols 1-6 blank, so find 1st sig char in col > 6	*/
	/*------------------------------------------------------*/
	/*{{{  stmt (cols 7-72)*/
	/*----------------------------------------------------------------
	Cols 1-6 blank, so find the first nonwhite char in col > 6.
		If any nonwhite char except '!', '\n' or EOF is seen
	in cols 7-72, this line is a new stmt.  In that case, save
	the char in '*labelp' and its column number in 'start_col',
	and return.
		Otherwise, skip the remainder of the line and try the
	next one (giving a warning if text starts in col > 72).
	----------------------------------------------------------------*/
	while ((c = readchar (col)) == ' ' || c == '\t')
		;
	
	/* ... now 'c' is the 1st nonwhite char ('col' may be > 72) */
	
	if (c == '\n' || c == EOF)
		/*----------------------*/
		/* a blank line		*/
		/*----------------------*/
		goto  read_next_line;
	
	else if (c == '!') {
		/*------------------------------------------------------*/
		/* A pure comment line -- copy it out.	 Note that we	*/
		/* don't complain if '!' starts in col > 72!		*/
		/*------------------------------------------------------*/
		if ( ! process_comment (col, COMMENT_LINE))
			goto  read_next_line;	/* a comment, copied out */
	
		next_label [0] = 'C';	/* ...signifying an HPF directive */
		start_col = col + 4;	/* col from which last char read */
		return  FALSE;		/* not a continuation */
	}
	else if (col <= 72) {
		/*------------------------*/
		/* This is a proper stmt! */
		/*------------------------*/
		*labelp++ = c;
		*labelp   = '\0';
		start_col = col;
		return  FALSE;		/* a new stmt, not a continuation */
	}
	else {
		/*------------------------------------------------------*/	
		/* The first non-white char is in col > 72 (i.e. it's	*/
		/* commentary).  Write a warning that the line is	*/
		/* ignored, scan to its end, and try the next one!	*/
		/*							*/
		/* N.B.: We know that 'c' isn't '\n' or EOL.		*/
		/*------------------------------------------------------*/	
		warning ("line has text only beyond column 72, and is ignored");
		while ((c = getc (ifp)) != '\n'  &&  c != EOF)
			;
		
		goto  read_next_line;
	}
	/*}}}*/
}
/*}}}*/
/*{{{  char *get_f77_stmt_line (ptr) */
char *get_f77_stmt_line (ptr)
/*-------------------------*/
	register char  *ptr;

/*{{{  DESCRIPTION*/
/*--------------------------------------------------------------
	This function inputs the next fixed-format non-blank
and non-comment line into 'textbuf', starting at position 'ptr',
and terminated by '\0' rather than '\n'.

	The function returns a pointer to the terminating '\0',
or NULL for EOF.  If cols 1-6 of the next stmt line contain errors,
an errmsg is given, the line is skipped, and (textbuf [0] == '\0').

	If ((ptr == textbuf) && !stmt_lp), then we're expecting a
*new* statement or EOF, not a continuation.  The first part of the
statement, comprising at least the label and the continuation column
(col 6), is obtained from the 'next_label []' buffer, where it was
placed by a prior call to 'f77_stmt_continued()'.  'next_label []'
contains either '\0' for EOF, or an exact copy of cols 1-6 of the
input line.  In the latter case, if there's a label it's installed
by 'lookup_or_install_label ()', which sets the global variables
'stmt_lp', 'stmt_label' and 'stmt_newlabel'.  It *isn't* written
to 'textbuf' (as it is for free format input).  The remainder of
the statement is copied from the input to 'textbuf' verbatim (e.g.
it isn't converted to upper case, and sequences of multiple whitespace
characters are preserved) so as not to corrupt Hollerith strings,
which cannot be readily identified at this stage.  If there are any
errors in cols 1-6  (viz, nondigits in cols 1-5, or a non-printing
char in col 6, or both a label AND a continuation char) then the
statement is skipped, and on return '*ptr' (i.e. textbuf [0])
just contains '\0'.

	If ((ptr > textbuf) || stmt_lp), then we already have the
start of a statement.  In this case, the last call to
'f77_stmt_continued()' showed that the forthcoming line is a
*continuation*.  There's no point inspecting 'next_label []', which
must have the form <     X>, and we start copying input into 'textbuf'
straight away.
--------------------------------------------------------------*/
/*}}}*/
{
	register char *labelp = next_label;
	register int   c,  col;
	static bool  in_string;
	static int   end_col,  string_delim;

	in_HPF_dir = (*labelp == 'C');


	if (! in_HPF_dir  &&  start_col) {
		/*{{{  cols 1-6 blank;  get 1st sig char from 'next_label[]'*/
		/*--------------------------------------------------------------
		This is the start of a new statement with cols 1-6 blank.
			We must still scan 'next_label []', because:
		-- we give a warning message if '\t' is found in cols 1-6;
		-- since cols 1-6 are blank, 'next_label []' also contains
			the first non-white char of the stmt body.
		The first nonwhite char was found in column 'start_col' (a global
		variable).  It can't be '\n' or '!', otherwise the line is a
		blank or comment line and would have been processed already.
		Also (start_col <= 72) for the same reason.  We must read this
		character into variable 'c' before going to the main loop which
		inputs the stmt body.
		/*--------------------------------------------------------------*/
		in_string = FALSE;
		while ((c = *labelp++) == ' ')
			;
		
		if (c == '\t') {
			warning ("using TAB to reach the statement field is non-standard");
			c = *labelp++;
		}
		
		col = start_col;
		
		/*----------------------------------------------------------------
		'c' is now the first significant char of the stmt body, in
		column 'col', as described above.
		----------------------------------------------------------------*/
		/*}}}*/
	}
	else if (in_HPF_dir  &&  (ptr == textbuf)) {
		/*{{{  start of an HPF directive*/
		/*----------------------------------------------------------------
		Get 1st significant char of stmt body
		-------------------------------------
		Before entering the next loop to read cols 7-72 we input the
		first *non-white* character 'c' in col >= 7.
			We skip leading whitespace, as we don't want to
		put whitespace at the start of 'textbuf'.
		
		N.B. Don't worry about exceeding col 72, as the 'stmt loop'
		will handle that!  Also note that the 1st nonwhite char 'c'
		may be a '\n', EOF or '!', in any column >= 7!
		----------------------------------------------------------------*/
		col = start_col;	/* column no. of last i/p char */
		while ((c = readchar (col)) == ' ' || c == '\t')
			;
		/*}}}*/
	}
	else if (ptr > textbuf || stmt_lp) {
		/*{{{  expect continuation;  read next i/p char*/
		/*--------------------------------------------------------------
			If (ptr > textbuf) or a label has been installed
		(stmt_lp != NULL), then the last call to 'f77_stmt_continued()'
		showed that the next line is a *continuation*.  There's no point
		inspecting 'next_label []', which must contain <     X>, so we
		start copying input into 'textbuf' straight away.
			In the previous input line, '\n' occurred in column
		'end_col' (i.e. column 'end_col' corresponds to '\0' in 'textbuf').
		If (ptr > textbuf) then we're *within* a stmt (as is usually the case),
		so we pad-out the line with blanks in columns (end_col -- 72) inclusive
		before reading the continuation line (in case we're within a
		Hollerith string where they're significant).  For the same reason.
		we don't skip whitespace in col >= 7 of the current continuation
		line, but rather copy it into 'textbuf',
			However, if (ptr == textbuf) then we're still at the start
		of the stmt (i.e. only a label has been installed), so we don't need
		to add whitespace to 'textbuf'.  Indeed, to do so would be erroneous,
		as we assume that there isn't whitespace at the start of 'textbuf'.
		----------------------------------------------------------------*/
		if (ptr > textbuf) {
			for (col = (73 - end_col);  col;  col--)
				*ptr++ = ' ';
		
			col = 6;
			c = readchar (col);
		}
		else {
			/*----------------------------------------------------------------
			Get 1st significant char of stmt body
			-------------------------------------
			Before entering the next loop to read cols 7-72 we input the
			first *non-white* character 'c' in col >= 7.
				We skip leading whitespace, as we don't want to
			put whitespace at the start of 'textbuf'.
		
			N.B. Don't worry about exceeding col 72, as the 'stmt loop'
			will handle that!  Also note that the 1st nonwhite char 'c'
			may be a '\n', EOF or '!', in any column >= 7!
			----------------------------------------------------------------*/
			col = 6;
			while ((c = readchar (col)) == ' ' || c == '\t')
				;
		}
		
		
		/* Now we go to the main loop which reads input for the stmt body... */
		/*}}}*/
	}
	else if ( ! *labelp)
		return  NULL;		/* EOF! */

	else {
		/*{{{  label, or '0'/non-print in col 6;  analyse 'next_label[]'*/
		/*----------------------------------------------------------------
		This line starts a new statement and bears a label (valid or not)
		and/or has a '0' or a non-print char in col 6.
			'next_label []' contains cols 1-6.  If there are any errors,
		give an errmsg and skip the line (returning '\0' in 'textbuf [0]').
		Otherwise, write the label, if any, at the start of 'textbuf'
		followed by one blank, then input up to the first non-white
		character 'c' in col >= 7 (which may be a '\n', EOF or '!'),
		ready for the main statement loop.
		----------------------------------------------------------------*/
		bool  cont_char,  illegal_label;
		int   illegal_char,  label,  n_digits_in_label;
		
		in_string = FALSE;	/* as new stmt */
		
		/*{{{  read cols 1-6 (i.e. label & continuation char) from 'next_label'*/
		/*---------------------------------------------------------------*/
		/* read cols 1-6 (label and continuation char) from 'next_label' */
		/*---------------------------------------------------------------*/	
		label = n_digits_in_label = 0;
		illegal_label = FALSE;
		
		c = *labelp++;
		col = 1;	/* N.B. (c != '\t'), as cols 1-6 aren't blank! */
		
		while ((col < 6) && (c != '!') && (c != '\n')) {	/* in label */
		
			/*-----------------------------*/
			/* this applies to cols 1-5... */
			/*-----------------------------*/
			if (isdigit (c)) {
				n_digits_in_label++;
				label = 10 * label + (c - '0');
			}
			else if ( ! (c == ' ' || c == '\t' || illegal_label) ) {
				illegal_label = TRUE;
				illegal_char = c;
			}
			/*-----------------------------*/
			/* this applies to cols 2-6... */
			/*-----------------------------*/
			c = *labelp++;
			if (c == '\t') {
				col = 6;
				warning ("using TAB to reach the statement field is non-standard");
			}
			else
				col++;
		}
		/*}}}*/
		
		/*{{{  if there's an error, write errmsg & skip line*/
		/*----------------------------------------------------------------
		Now 'c' is the character in column 6, or earlier due to a premature
		end of label if 'c' is '!', '\t' or '\n'.
			If there's an error in cols 1-6, write an errmsg and
		goto 'skip_line', which skips the rest of the line, puts '\0'
		in 'textbuf [0]', and returns 'textbuf'.
		
		N.B. If 'next_label []' contains '\n', then that is the current
		value of 'c'.  The code for line-skipping checks 'c' and does
		nothing if (c == '\n').  This is important, since if the '\n' was
		missed, then:  (i) the next line would we skipped, even though it
		may be legal, and  (ii) 'lineno' wouldn't be incremented, so would
		be too low by 1.
		----------------------------------------------------------------*/
		cont_char = !(c == ' ' || c == '0' || c == '\t' || c == '\n' ||
		                                                  (c == '!' && col < 6));
		if (illegal_label) {
			/*------------------------------------------------------*/
			/* Illegal char(s) in label.  Skip the line and just	*/
			/* return '\0' in 'textbuf[0]'!				*/
			/*------------------------------------------------------*/
			begin_errmsg ("nondigit ");
			if (isprint (illegal_char))
				cont_errmsg ("\'%c\'", illegal_char);
			else
				cont_errmsg ("(hex value \'%X\')",
				              ((unsigned int) (unsigned char) illegal_char) );
			end_errmsg (" in label (columns 1-5) - line skipped");
		
			goto  skip_line;
		}
		else if ( ! (isprint (c) || c == '\n' || c == '\t')) {
			/*------------------------------------------------------*/
			/* Illegal non-printing char in column 6.  Skip line	*/
			/*------------------------------------------------------*/
			errmsg ("illegal character (hex value \'%X\') in column 6 - line skipped",
			            ((unsigned int) (unsigned char) c) );
		
			goto  skip_line;
		}
		else if (n_digits_in_label  &&  cont_char) {
			/*------------------------------------------------------*/
			/* Line has both label and continuation char.  Skip it.	*/
			/*------------------------------------------------------*/
			errmsg ("line has both label and continuation character (\'%c\' in column 6) - line skipped",
			         c);
		
			goto  skip_line;
		}
		/*}}}*/
		
		/*{{{  if there's a label, check it's non-0 and install it*/
		/*--------------------------------------------------------------
		At this point we know that there isn't an error in cols 1-6.
		They must have one of the forms:
		
			<label >,  <label0>  or  <     0>
		
			If there is a label, check that it's valid (i.e. non-zero).
		If so, look it up or install it, giving an errmsg if it has been
		set already.
		
		N.B.  'lookup_or_install_label ()' sets 'stmt_lp', stmt_label' and
		'stmt_newlabel'.
		----------------------------------------------------------------*/
		if (label)
			lookup_or_install_label (label, SET_LABEL);    /* sets 'stmt_lp', etc*/
		
		else if (n_digits_in_label)
			errmsg ("label must be non-zero");
		/*}}}*/
		
		/*{{{  if cols 1-6 contain '!' or '\n', process them and return*/
		/*--------------------------------------------------------------
			Test for a premature end-of-label, i.e. a '!' before col 6
		or a '\n' at col <= 6.
		
		N.B. There must be a label prior to the '!' or '\n', otherwise the
		line would be a blank or comment line and would have been output
		already.
		----------------------------------------------------------------*/
		if (c == '!' && col != 6) {
			process_comment (col, TRAILING_COMMENT);
			end_col = 73;		/* don't pad with ' ' if line continued */
			*ptr = '\0';
			return  ptr;
		}
		else if (c == '\n') {
			end_col = 73;		/* don't pad with ' ' if line continued */
			*ptr = '\0';
			return  ptr;
		}
		
		/*--------------------------------------------------------------
		N.B. At this stage, the only form of cols 1-6 that we haven't
		explicitly tested for is:
			<     0>
		i.e. spaces in cols 1-5, and '0' in col 6.  This doesn't need any
		special action -- it just indicates the start of a new stmt -- so
		just go ahead and read the input.
		----------------------------------------------------------------*/
		/*}}}*/
		
		/*{{{  read the 1st nonwhite char in stmt body (cols 7-72)*/
		/*----------------------------------------------------------------
		Get 1st significant char of stmt body
		-------------------------------------
		Before entering the next loop to read cols 7-72 we input the
		first *non-white* character 'c' in col >= 7.
			We skip leading whitespace as we don't want to put
		whitespace at the start of 'textbuf'.
		
		N.B. Don't worry about exceeding col 72, as the 'stmt loop'
		will handle that!  Also note that the 1st nonwhite char 'c'
		may be a '\n', EOF or '!', in any column >= 7!
		----------------------------------------------------------------*/
		
		while ((c = readchar (col)) == ' ' || c == '\t')
			;
		/*}}}*/
		/*}}}*/
	}

	/*{{{  input stmt body (cols 7-72)*/
	/*----------------------------------------------------------------
	input statement body (columns 7-72)
	-----------------------------------
	Now just copy the input to 'textbuf' until the end-of-line.
	
	N.B. The next significant char 'c' is already available
	----------------------------------------------------------------*/
	while (col <= 72) {
		if (c == '\n' || c == EOF) {
			/*------------------------------------------------
			End-of-line.
				Save the column number in static var 'end_col',
			so that if the line is continued it can be padded-out
			with blanks in columns (end_col -- 72) inclusive
			(in case we're in a Hollerith string where blanks
			are significant).
			------------------------------------------------*/
			end_col = col;
			*ptr = '\0';
			return  ptr;
		}
		else if (in_string) {
			/*------------------------------------------------
			ignore the meaning of all special chars (eg. '!')
			except the string delimiter.
			------------------------------------------------*/
			*ptr++ = c;
			if (c == string_delim)
				in_string = FALSE;
		}
		else if (c == '\'' || c == '\"') {
			*ptr++ = string_delim = c;
			in_string = TRUE;
		}
		else if (c == '!') {
			/*------------------------------------------------
			Assume that:
			 (i) we're not in a Hollerith (so the line needn't be
			padded with blanks if continued);
			 (ii) a comment always separates tokens.
				Hence, set 'end_col' to 73, and replace the
			comment by a single space if within a statement.
			------------------------------------------------*/
			process_comment (col, TRAILING_COMMENT);
			end_col = 73;
			if (ptr > textbuf) *ptr++ = ' ';
			*ptr = '\0';
			return  ptr;
		}
		else
			*ptr++ = c;
	
		c = readchar (col);
	}
	/*}}}*/
	/*{{{  cols >72  -- give warning if non-white, skip rest of line & return*/
	/*----------------------------------------------------------------
	columns >72
	-----------
	Skip whitespace, since don't want to complain that a stmt is
	too long simply because of trailing whitespace!
	
	N.B. The next char 'c' is already available!
	----------------------------------------------------------------*/
	end_col = 73;
	while (c == ' ' || c == '\t')
		c = getc (ifp);
	
	if (c == '\n' || c == EOF) {
		/*------------------------------------------------------*/
		/* OK -- just whitespace after col 72.  Now return!	*/
		/*------------------------------------------------------*/
		*ptr = '\0';
		return  ptr;
	}
	else {
		/*--------------------------------------------------------
		Warn about extra characters beyond col 72.
		N.B. we ignore the special meaning of any chars (e.g. '!',
		'\'') in this field.  We even give a warning if '!' is the
		first sig char in col > 72, as the stmt text may be a
		string, where '!' doesn't denote a comment!
		--------------------------------------------------------*/
		char tempbuf [6];
		int  i = 0;
		
		/*--------------------------------------------------------
		Copy next 5 chars, or upto '\n' if earlier, into 'tempbuf'.
		--------------------------------------------------------*/
		do {
			tempbuf [i++] = c;
			c = getc (ifp);
		} while (c != '\n'  &&  c != EOF  &&  i < 5);
		tempbuf [i] = '\0';
	
		/*--------------------------------------------------------
		Give an errmsg (whose exect form depends on whether there
		are any more nonwhite chars in the line, for exactness!)
		--------------------------------------------------------*/
		while (c == ' ' || c == '\t')
			c = getc (ifp);
	
		warning ("statement extends beyond column 72 - the remainder (%s %q) is ignored",
			((c == '\n' || c == EOF) ? "namely" : "starting"),  tempbuf);
	
		goto  skip_line;
	}
	/*}}}*/

skip_line :
	while ( ! (c == '\n' || c == EOF))
		c = getc (ifp);

	*ptr = '\0';
	return  ptr;
}
/*}}}*/

/*{{{  bool  process_comment (in_col, trailing) -- DONE HPF */
bool  process_comment (in_col, trailing)
/*------------------------------------*/
	register int  in_col;	/* column no. of LAST input char */
	bool  trailing;		/* TRUE for trailing comment */
/*----------------------------------------------------------------
	A comment delimiter has just been read in column 'in_col'
of the source line (i.e. a '!', or, in fixed source form, one of
the characters 'C', 'c' or '*' in col 1).  This function copies
the comment, up to the end-of-line, directly to the output, except
that tabs are converted to single spaces.
	It keeps track of the column numbers of the input and output
lines.  If the input line is too long, a warning is issued for free
source form (provided that the extra chars are not just whitespace),
but not for fixed; in either case the whole comment is still output.
On output, the comment is split up into lines 80 columns long, in
accordance with the requirements of F77, except that a new comment
line is not written if it would contain only trailing whitespace.
----------------------------------------------------------------*/
/*{{{  COMMENT -- how it works*/
/*------------------------------------------------------------------
How it works
------------
	At the start of section 'copy_comment', the function has
buffered-up  (i.e. read from the input line but not yet written to the
output line) 'nspaces' blanks followed by the character 'c' (which
may be anything).

Section 'copy_comment' writes out 'nspaces' blanks, followed by
the character 'c', then copies through the input up to the
end-of-line, all the while keeping a track of the input and
output line lengths.  It branches to 'input_too_long' or
'output_too_long' when necessary.

The behaviour of the macro 'quit_if_only_nonwhite_left' is described
below.  The rest is obvious.
------------------------------------------------------------------*/

/*}}}*/
/*{{{  #define  quit_if_only_whitespace_left*/
/*----------------------------------------------------------------
The behaviour of the macro 'quit_if_only_whitespace_left' is as
follows.  If it is invoked with 'c' NOT equal to a whitespace or
'\n', it performs no action and returns immediately.  If 'c' is
a whitespace char (' ' or '\t'), the input is read up to the next
non-white char, which is assigned to 'c', and 'nspaces' is
incremented accordingly.  Finally, if  c=='\n', the macro branches
to the 'end' label.
----------------------------------------------------------------*/ 
#define  quit_if_only_whitespace_left(c,nspaces)			\
		if (c == ' ' || c == '\t') {				\
			nspaces++;					\
			while ((c = readchar (in_col)) == ' ' || c == '\t') \
				nspaces++;				\
		}							\
		if (c == '\n' || c == EOF)				\
			goto  end

/*}}}*/

{
	/*{{{  decls*/
	register int  c, out_col;	/* column no. of NEXT output char */
	char  *ptr,  prefix_buf [4];
	int    i,  nspaces;
	bool   check_input_len = free_source_form;
				/* don't bother for fixed source form! */
	/*}}}*/
	
	/*{{{  read first 4 chars; if ("HPF$" && ! trailing) return TRUE, else copy out*/
	/*----------------------------------------------------------------
	Read the first 4 characters after the comment character  (checking
	the input column number if necessary).  If they're "HPF$" (in upper
	or lower case return TRUE.  Otherwise copy them out, read the next
	input character 'c', initialise (nspaces == 0), and goto 'copy_comment'.
	----------------------------------------------------------------*/
	if (check_input_len  &&  (in_col > MAX_LINELEN)) {
		if (trailing)
			warning ("trailing comment makes line >%d characters long",
			          MAX_LINELEN);
		else
			warning ("comment line >%d characters long", MAX_LINELEN);
		check_input_len = FALSE;
	}
	
	for (i = 4,  ptr = prefix_buf;   i;   --i,  ++ptr) {
		*ptr = c = readchar (in_col);
	
		switch (c) {
		case  '\n' :
		case  EOF :
			return  FALSE;    /* output nothing if <4 chars in comment!*/
		}
		if (check_input_len  &&  (in_col > MAX_LINELEN)  &&
		                                      ! (c == ' ' || c == '\t')) {
			if (trailing)
				warning ("trailing comment makes line >%d characters long", MAX_LINELEN);
			else
				warning ("comment line >%d characters long", MAX_LINELEN);
			check_input_len = FALSE;
		}
	}
	
	if ((prefix_buf [0] == 'h' || prefix_buf [0] == 'H')  &&
	    (prefix_buf [1] == 'p' || prefix_buf [1] == 'P')  &&
	    (prefix_buf [2] == 'f' || prefix_buf [2] == 'F')  &&
	    (prefix_buf [3] == '$')) {
	
		if ( ! trailing)
			return  TRUE;		/* this is an HPF directive! */
		else
			errmsg ("HPF directive must not be a trailing comment");
				/* ... and continue processing as though it
				       isn't an HPF directive */
	}
	out_col = 1;
	writechar (out_col, 'C');
	for (i = 4,  ptr = prefix_buf;   i;   --i,  ++ptr)
		writechar (out_col, *ptr);
	
	nspaces = 0;
	c = readchar (in_col);		/* the 5th char of the comment! */
	
	/* ... now goto section copy_comment */
	/*}}}*/
	
copy_comment:
	/*{{{  */
	/*----------------------------------------------------------------
	Output 'nspaces' blanks, followed by 'c', then copy the remainder
	of the input line to the output.  Before any char is written,
	the output column is checked and a new output comment line started
	if necessary (i.e. if the output column is 81 and there are still
	nonwhite chars left).  Also, once the blanks are written, we keep
	track of the input line length and give a warning if it exceeds
	MAX_LINELEN.
	----------------------------------------------------------------*/
	
	for (;  nspaces;  nspaces--) {
		if (out_col == 81)
			goto  output_too_long;
		else
			writechar (out_col, ' ');
	}
	
	while (TRUE) {
		if (c == '\n' || c == EOF)
			goto  end;
	
		else if (out_col == 81)
			goto  output_too_long;
	
		else if (check_input_len  &&  (in_col > MAX_LINELEN))
			goto  input_too_long;
	
		else
			writechar (out_col, c);
	
		c = readchar (in_col);
	}
	
	/*}}}*/
		
input_too_long:
	/*{{{  */
	/*--------------------------------------------------------------*/
	/* Check that the current char is nonwhite, or that there are	*/
	/* nonwhite chars remaining in the comment, and if so, give	*/
	/* warning message that input line is too long.			*/
	/*--------------------------------------------------------------*/
	quit_if_only_whitespace_left (c, nspaces);
	
	if (trailing)
		warning ("trailing comment makes line >%d characters long", MAX_LINELEN);
	else
		warning ("comment line >%d characters long", MAX_LINELEN);
	check_input_len = FALSE;
	
	goto  copy_comment;
	/*}}}*/


output_too_long:
#ifdef  V_ADAPT
	/*{{{  */
	/*--------------------------------------------------------------*/
	/* Check that there are more nonwhite chars still to be written	*/
	/* (because we don't want to start a new comment line that will	*/
	/* just contain blanks!), and, if so, start a new comment line.	*/
	/*--------------------------------------------------------------*/
	quit_if_only_whitespace_left (c, nspaces);
	
	putchar ('\n');
	
	out_col = 1;
	writechar (out_col, 'C');
	goto  copy_comment;
	/*}}}*/
#else
	/*{{{  */
	/*--------------------------------------------------------------*/
	/* Check that there are more nonwhite chars still to be written	*/
	/* (because we don't want to start a new comment line that will	*/
	/* just contain blanks!), and, if so, start a new comment line.	*/
	/*--------------------------------------------------------------*/
	quit_if_only_whitespace_left (c, nspaces);
	
	out_col = 1;
	writechar (out_col, 'C');
	goto  copy_comment;
	/*}}}*/
#endif

end:
#ifdef  V_ADAPT
	putchar ('\n');
#endif
	return  FALSE;		/* i.e. not an HPF directive */
}

#undef  quit_if_only_whitespace_left
/*}}}*/

/*{{{  skip_rest_of_stmt () */
skip_rest_of_stmt ()
/*----------------*/
/*----------------------------------------------------------------
	Called by 'yyerror()' and 'read_format()' after a syntax
error has been found, to skip the rest of the statement, including
continuation lines and subsequent statements following ';' on
the same line.  The function advances through 'textbuf' from
'end_tk' to the end-of-line (signalled by *ptr == '\0').  'end_tk'
may already point to the '\0'.
	On return, 'end_tk' is left pointing to the '\0' (i.e.
the end-of-line) in 'textbuf'.

N.B. It's tempting to stop if a ';' is found, so that we only
skip the rest of *this* statement, not subsequent statements
separated by ';'.   We don't do this because of the complication
that  ';' might be within a string or Hollerith string.
	In any case, this function isn't called if we're already
at a ';' when the syntax error is found (i.e. if the error is
"incomplete statement (terminated by ';')").  In that case,
'yylex ()' resumes at the char after the ';' regardless of the error.
----------------------------------------------------------------*/
{
	register char  *ptr = end_tk;

	while (*ptr++)
		;

	end_tk = ptr - 1;	/* points to the terminating '\0' */
}
/*}}}*/
/*}}}*/

/*{{{  <<< token reading functions >>>  */
/*--------------------------------------------------------------*/
/* Token reading functions called directly by 'yylex ()'	*/
/* (and from elsewhere when indicated).				*/
/*--------------------------------------------------------------*/
/*{{{  int  read_word_or_boz_or_kind_string (startp)  */
int  read_word_or_boz_or_kind_string (startp)
/*-----------------------------------------*/
	char  *startp;		/* pointer to the start of a word */
/*--------------------------------------------------------------*/
/*	This function handles all tokens starting with a letter.*/
/* As well as identifiers and keywords, this category also	*/
/* includes:							*/
/*  --  strings starting with a named kind param;		*/
/*  --  BOZ literal constants.					*/
/*--------------------------------------------------------------*/
{
	register char  *ptr;
	register int    c;
	register unsigned  hashval;
	char  *endp;
	int    endc,  len,  token;

	/*{{{  scan the word and put '\0' after its last char in 'textbuf'*/
	/*--------------------------------------------------------------*/
	/*	Scan the word in 'textbuf' and put a '\0' delimiter	*/
	/* immediately after it (at position 'end_tk'), saving the	*/
	/* overwritten char in 'endc' to be restored before the function*/
	/* returns.  If the word is too long (> MAX_ID_LEN chars), put	*/
	/* the '\0' after the first MAX_ID_LEN chars and give a warning,*/
	/* but still scan 'textbuf' to the actual end of the		*/
	/* (overlength) identifier.					*/
	/*--------------------------------------------------------------*/
	ptr = startp;
	c   = *ptr;
	if (islower (c)) c = *ptr = toupper (c);
	hashval = 0;
	
	do {
		hashval = c + (31 * hashval);
		c = *++ptr;
		if (islower (c)) c = *ptr = toupper (c);
	} while (isidchar (c));
	
	hashval = hashval % HASHSIZE;
	
	end_tk = endp = ptr;	/* to point to 'c', the next char after 'word'.	*/
				/* ('end_tk' is a global pointer).		*/
				
	len = ptr - startp;			
	if ( (len <= MAX_ID_LEN)  ||  ((len == MAX_ID_LEN+1) &&
	                        (c == '\''||c == '\"') && (*(ptr-1) == '_')) )
		/*------------------------------------------------------*/
		/* If 'word' ends '_' and is followed by ' or ",  it's	*/
		/* the kind parameter of a string.  In that case the	*/
		/* final '_' *isn't* part of 'word', so 'len' is 1 more	*/
		/* than it should be.  The above test allows for this!	*/
		/*------------------------------------------------------*/
		delimit_word_in_textbuf (endc, endp);
	
	else {
		/*------------------------------------------------------*/
		/* 'word' is too long!  Give a warning and rescan it,	*/
		/* putting '\0' after MAX_ID_LEN chars and re-computing	*/
		/* 'hashval' for the newly-defined word, but leaving	*/
		/* 'c' and 'end_tk' unchanged.				*/
		/*------------------------------------------------------*/
		len = MAX_ID_LEN;
		endp = startp + MAX_ID_LEN;
		delimit_word_in_textbuf (endc, endp);
		warning ("identifier is too long; only %d characters are significant (viz. %q)",
		          MAX_ID_LEN,  startp);
	
		for (ptr = startp,  hashval = 0;   ptr < endp;   ptr++)
			hashval = *ptr + (31 * hashval);
	
		hashval = hashval % HASHSIZE;
	}
	/*}}}*/

	/*------------------------------------------------------*/
	/*	Usually 'word' is an identifier.  However, it	*/
	/* might be a kind parameter for a string, or a single	*/
	/* letter that introduces a BOZ literal constant. Check	*/
	/* these possibilities and take the appropriate action.	*/
	/*	At this point:					*/
	/* endc == *ptr == *endp :				*/
	/*	char after significant part (limited to 31 chars)*/ 
	/* c == *end_tk :					*/
	/*	char after the *full* 'word' (even if len > 31).*/ 
	/*------------------------------------------------------*/
	token = 0;
	if (c == '\'' || c == '\"') {
		if (len == 1) {
			/*{{{  may be a BOZ literal const (if 'word' is B,O or Z)*/
			/*------------------------------------------------------*/
			/*  The word MAY introduce a BOZ literal constant, viz:	*/
			/*	B'...'  or  B"..."  or				*/
			/*	O'...'  or  O"..."  or				*/
			/*	Z'...'  or  Z"...".				*/
			/*------------------------------------------------------*/
			int  base = 0;
			
			switch (*startp) {
			case 'B' :	base = 2;	break;
			case 'O' :	base = 8;	break;
			case 'Z' :	base = 16;	break;
			}
			
			if (base) {
				/*------------------------------------------------------*/
				/* A BOZ literal!  Before calling 'read_boz_num ()' we	*/
				/* must restore the delimiter (' or ") in 'textbuf',	*/
				/* as the BOZ constant - including its delimiters - may	*/
				/* be output in errmsgs by 'read_boz_num ()'.		*/
				/*							*/
				/* N.B. It's also restored again later, but that's ok	*/
				/* -- the same thing is simply done twice!!		*/
				/*------------------------------------------------------*/
				restore_textbuf (endc, endp);		/* restore the ' or " ('c') */
				token = read_boz_num (ptr, base);	/* 'ptr' points to ' or " */
									/* Resets 'emd_tk' */
			}
			/*}}}*/
		}
		else if ( *(end_tk-1) == '_') {      /* 'word' ends '_' */
			/*{{{  'word' is a kind parameter for a string*/
			/*--------------------------------------------------------------*/
			/*	'word' is a named kind parameter followed by a string.	*/
			/*	Remove the final '_' from the word (if it's there --	*/
			/* it may have been omitted because the word is overlength)	*/
			/* and lookup the kind value.  Then read and return the string.	*/
			/*								*/
			/* N.B. If 'word' isn't too long, 'endp' points to the position	*/
			/* in 'textbuf' originally occupied by the string delimiter,	*/
			/* which is now occupied by '\0'.  The code below restores the	*/
			/* delimiter in 'textbuf' and changes 'endp'.			*/
			/*	If 'word' *is* overlength, 'endp' points to an earlier	*/
			/* character position.						*/
			/*	In either case, 'textbuf' contains the string delimiter	*/
			/* when 'read_string ()' is called, which is necessary as the	*/
			/* string - including its delimiters - is saved or output in	*/
			/* errmsgs by 'read_string ()'.					*/
			/*--------------------------------------------------------------*/
			Symbol  *sp;
			
			if (endp == end_tk) {
				/*------------------------------------------------------*/
				/* The word *isn't* too long, and is incorrectly	*/
				/* terminated by '_'.					*/
				/*	We therefore re-scan it, omitting the final '_',*/
				/* to re-compute its 'hashval' and re-position the '\0'	*/
				/* so it can be looked-up.  However, 'end_tk' is left	*/
				/* unchanged.						*/
				/*------------------------------------------------------*/
				restore_textbuf (endc, endp);	/* restore the ' or " */
				endp--;				/* now 'endp' points to '_' */
				len--;				/* now 'len' omits the '_' */
				delimit_word_in_textbuf (endc, endp);
			
				for (ptr = startp,  hashval = 0;   ptr < endp;   ptr++)
					hashval = *ptr + (31 * hashval);
			
				hashval = hashval % HASHSIZE;
			}
			
			sp = lookup_or_install_id (startp, len, hashval);   /* Don't put 'sp'   */
									  /* on parser stack! */
			check_kind_name (TY_CHARACTER, sp);
			
			token = read_string (end_tk);	/* 'end_tk' points to string delimiter. */
							/* This function resets the global 'end_tk'.*/
			/*}}}*/
		}
	}
	if (! token  &&  nwords_to_keyword  &&  ! nparens  &&
	                 ! --nwords_to_keyword)
		token = lookup_keyword (startp, endc, hashval);
			/* resets 'end_tk' if a multiple keyword */

	if (! token) {
		yylval.sym = lookup_or_install_id (startp, len, hashval);
		token = TK_ID;
	}

	restore_textbuf (endc, endp);
	
	return  token;
}
/*}}}*/
/*{{{  int  read_dotted_word (startp) */
int  read_dotted_word (startp)
/*--------------------------*/
	char  *startp;		/* pointer to the start of 'dotted word' */
/*--------------------------------------------------------------*/
/*	On entry, 'startp' points to a '.' and the next char	*/
/* is a letter, indicating the start of a 'dotted word' (viz.	*/
/* a string of letters starting and ending with a '.').		*/
/*--------------------------------------------------------------*/
{
	register char  *ptr;
	register int    c;
	register unsigned  hashval;
	bool   only_letters;
	char  *endp;
	int    endc,  len,  token;

	/*{{{  scan the letters of the 'dotted word' (upto & incl the final '.')*/
	/*--------------------------------------------------------------*/
	/*	Scan the word in 'textbuf' and put a '\0' delimiter	*/
	/* immediately after it (at position 'endp'), saving the	*/
	/* overwritten char in 'endc' to be restored before the function*/
	/* returns.  If the word is too long (> MAX_ID_LEN chars between*/
	/* the initial and final '.'), put the '\0' after the first	*/
	/* MAX_ID_LEN chars and give a warning, but still scan 'textbuf'*/
	/* to the actual end of the (overlength) identifier.		*/
	/*--------------------------------------------------------------*/
	int  n_idchars;
	
	only_letters = TRUE;		/* only letters are allowed! */
	hashval = '.';
	ptr = startp+1;
	
	c = *ptr;
	if (islower (c)) c = *ptr = toupper (c);
	
	do {
		if (! isupper (c))
			only_letters = FALSE;
		hashval = c + (31 * hashval);
		c = *++ptr;
		if (islower (c)) c = *ptr = toupper (c);
	} while (isidchar (c));
	
	if (c == '.') {
		hashval = '.' + (31 * hashval);
		ptr++;			/* points to next char after word */
	}
	
	end_tk = endp = ptr;		/* points to next char after word */
	
	hashval   = hashval % HASHSIZE;
	len       = ptr - startp;			/* total length incl '.'s */
	n_idchars = (c == '.') ? len - 2 : len - 1;	/* length excluding '.'s */
	
	if (n_idchars <= MAX_ID_LEN)		/* N.B. omit '.'s from this count */
		delimit_word_in_textbuf (endc, endp);
	
	else {
		/*------------------------------------------------------*/
		/* 'word' is too long!  Give a warning and rescan it,	*/
		/* putting '\0' after MAX_ID_LEN chars and re-computing	*/
		/* 'hashval' for the newly-defined word, but leaving	*/
		/* 'c' and 'end_tk' unchanged.				*/
		/* N.B. In this case the final '.' is omitted from the	*/
		/* truncated word!					*/
		/*------------------------------------------------------*/
		len  = MAX_ID_LEN + 1;		/* new 'len' (including initial '.') */
		endp = startp + len;
		delimit_word_in_textbuf (endc, endp);
		warning ("operator name is too long; only %d letters are significant (viz. %q)",
		          MAX_ID_LEN, startp);
	
		only_letters = TRUE;		/* only letters are allowed! */
		for (ptr = startp+1,  hashval = '.';   ptr < endp;   ptr++) {
			hashval = *ptr + (31 * hashval);
			if (! isupper (*ptr)) only_letters = FALSE;
		}
		hashval = hashval % HASHSIZE;
	}
	/*}}}*/

	/*------------------------------------------------------*/
	/* give errmsg if 'word' (incl. any overlength chars)	*/
	/* doesn't end with a dot.				*/
	/*------------------------------------------------------*/
	if (c != '.')
		errmsg ("a name beginning with a '.' must end with one: %q",
		         startp);

	/*----------------------------------------------*/
	/* give errmsg if 'word' contains non-letters.	*/
	/*----------------------------------------------*/
	else if ( !only_letters)
		errmsg ("operator name %q must contain only letters between the initial and final \'.\'",
		         startp);

	/*------------------------------------------------------*/
	/* Lookup (and if necessary install) 'word'.		*/
	/* N.B. Here, we search 'keytab' irrespective of the	*/
	/* value 'nwords_to_keyword', as relational operators	*/
	/* like ".EQ." are classified as keywords and may	*/
	/* appear in expressions (unlike other keywords!)	*/
	/*------------------------------------------------------*/
	if ( !(token = lookup_dotted_keyword (startp, endc, hashval)) ) {

		yylval.sym = lookup_or_install_id (startp, len, hashval);
		token = TK_DEFINED_OP;
	}
	restore_textbuf (endc, endp);
	
	return  token;
}
/*}}}*/
/*{{{  int  read_num_or_kind_string (startp,  write_yylval) -- [& elsewhere]*/
int  read_num_or_kind_string (startp,  write_yylval)
/*------------------------------------------------*/
	char  *startp;		/* pointer to the start of a number */
	bool   write_yylval;	/* if TRUE put value on parser stack */
/*--------------------------------------------------------------
	Called by 'yylex()' & 'read_complex_num_or_paren()' to read
an integer or real number (including its kind parameter, if any),
or a string preceded by a kind parameter.
	On input 'startp' points to a digit or a '.' followed by
a digit, i.e. a valid first character of a number.
	The function returns TK_INT_LIT_CONST, TK_REAL_LIT_CONST
or TK_DP_LIT_CONST.  If 'write_yylval' is TRUE, then for the
first it puts the integer's value in 'yylval.num', while for the
last two it makes a copy of the number's lexeme and puts it in
'yylval.str'.

	Small complications are:
-- The number may be a kind value prefixed to a string.  In that
case the string is also read and TK_CHAR_LIT_CONST returned.
-- The number may be an integer immediately followed by a
'dotted name' (e.g. "2.EQ."), which may look like a real number
with an exponent until the 2nd char after the '.' is read!
----------------------------------------------------------------*/
{
	char  *ptr;
	int    c,  token,  val;

	token = read_num (startp, &val);

	ptr = end_tk;		/* global 'end_tk' was set by 'read_num()'*/
	c   = *ptr;

	/*------------------------------------------------------*/
	/*	Now 'ptr' points to the next char after the	*/
	/* number ('c').  If (c == '_'), it could be followed	*/
	/* by a kind parameter or a string...			*/
	/*------------------------------------------------------*/
	if (c == '_'  &&  ( *(ptr+1) == '\'' || *(ptr+1) == '\"' )) {
		/*{{{  'num' is the kind value of a string*/
		/*-----------------------------------------------*/
		/* 'num' is the (prefix) kind value of a string. */
		/*-----------------------------------------------*/
		if (token == TK_INT_LIT_CONST)
			check_kind_val (TY_CHARACTER, val);
		else {
			delimit_word_in_textbuf (c, ptr);
			errmsg ("invalid kind parameter %q; it must be an integer",
			         startp);
			restore_textbuf (c, ptr);
		}
		
		ptr++;			/* to point to the string delimiter */
		token = read_string (ptr);
		/*}}}*/
	}
	else {
		/*{{{  a number;  set 'yylval' & check kind param if necessary*/
		/*--------------------------------------------------------------*/
		/* Terminate number by '\0' so '*start_tk' is its lexeme, for	*/
		/* the purpose of errmsgs, saving the number in a string, etc.	*/
		/*								*/
		/* N.B.  We must use the local variable 'ptr' rather than the	*/
		/* global variable 'end_tk' to store the char position after	*/
		/* 'num', where '\0' is to be written, as 'end_tk' is reset by	*/
		/* 'read_and_check_kind_param ()'.				*/
		/*--------------------------------------------------------------*/
		
		*ptr = '\0';	/* 'c' stores the char formerly in this posn */
		
		if (write_yylval) {
			if (token == TK_INT_LIT_CONST)
				yylval.num = val;
		
			else {
				/*----------------------------------------------*/
				/* Save the real literal constant in a string	*/
				/* and assign it to 'yylval.str'.		*/
				/*----------------------------------------------*/
				char  *s;
				
				s = emalloc (ptr - startp + 1);	/* +1 for the '\0' */
				strcpy (s, startp);
				yylval.str = s;
			}
		}
		
		if (c == '_') {
			/*------------------------------------------------------*/
			/* There follows the kind parameter of 'num'.		*/
			/*							*/
			/* N.B. 'ptr' points to '_', so we pass (ptr+1), which	*/
			/* points to the start of the kind parameter, to	*/
			/* 'read_and_check_kind_param()'.			*/
			/*------------------------------------------------------*/
			if (token == TK_DP_LIT_CONST)
				errmsg ("cannot specify a kind parameter for a double precision constant: %q",
				         startp);
		
			read_and_check_kind_param ((ptr + 1), token, startp);
		}
		
		*ptr = c;	/* restore the 'textbuf' */
		/*}}}*/
	}

	return  token;
}
/*}}}*/
/*{{{  int  read_string (startp)                            -- [& elsewhere]*/
int  read_string (startp)
/*---------------------*/
	char  *startp;		/* pointer to initial delimiter */
/*--------------------------------------------------------------*/
/*	Called by 'yylex ()', 'read_num_or_kind_string ()',	*/
/* 'read_word_or_boz_or_kind_string ()' and 'read_format ()'	*/
/* to read a string.						*/
/*	On entry, 'startp' points to the initial string		*/
/* delimiter character (a quote or apostrophe).  Before		*/
/* returning, the function resets the global pointer 'end_tk'	*/
/* to point to the next char after the string.			*/
/*--------------------------------------------------------------*/
{
	register char *ptr;
	register int   c,  delim;
	char    *endp;

	for (delim = *startp,  ptr = startp + 1;   c = *ptr;   ptr++) {
		/*----------------------------------------------*/
		/* quit loop at 'delim' character, unless the	*/
		/* following character is also 'delim'!		*/
		/*----------------------------------------------*/
		if ((c == delim)  &&  (*++ptr != delim))
			goto  finished_string;
	}
	/*{{{  missing closing delimiter.  Give errmsg*/
	/*--------------------------------------------------------------*/
	/* Reached end-of-line without seeing a closing delimiter!	*/
	/* Give an errmsg like the following:				*/
	/*								*/
	/* (<= 6 chars seen)						*/
	/*	character string ''hello' has no closing delimiter (')	*/
	/*								*/
	/* (> 6 chars seen)						*/
	/*	character string beginning ''hello' has no closing	*/
	/*	delimiter (')						*/
	/*								*/
	/* (but still place a copy of the string, such as it is, on	*/
	/* the parser stack).						*/
	/*								*/
	/* N.B. The error may be that the *wrong* sort of closing	*/
	/* delimiter was supplied (e.g. "....') but don't bother	*/
	/* checking for or suggesting this in the errmsg!		*/
	/*--------------------------------------------------------------*/
	endp = startp + 6;
	
	delimit_word_in_textbuf (c, endp);
	
	errmsg ("character string %s%q has no closing delimiter (%c)",
	         ((ptr - startp < 6) ? "" : "beginning "),  startp,  delim);
	
	restore_textbuf (c, endp);
	/*}}}*/
	
finished_string :
	end_tk = endp = ptr;	/* pointer to char after the string */
	
	delimit_word_in_textbuf (c, endp);

	ptr = emalloc (endp - startp + 1);	/* +1 for the terminal '\0' */
	strcpy (ptr, startp);
	yylval.str = ptr;

	restore_textbuf (c, endp);

	return  TK_CHAR_LIT_CONST;
}
/*}}}*/
/*{{{  int  read_complex_num_or_paren (startp) */
int  read_complex_num_or_paren (startp)
/*-----------------------------------*/
	char  *startp;		/* points to '(', which may start complex num*/
/*----------------------------------------------------------------
	Called by 'yylex ()'.  On entry, 'startp' points to a '('
which might start a complex number.  This function scans the
following text and returns the token 'TK_COMPLEX_LIT_CONST' if it
IS a complex number, or '(' otherwise.  It also resets the global
pointer 'end_tk' to point to the next character after the recognised
token.
	The complex number may be scanned twice.  Initially 
we cannot use 'read_num_or_kind_string ()' to scan the real and
imaginary parts of the putative complex number, as it writes error
messages and increments 'end_tk', both of which are undesirable
if the scanned text ISN'T a complex value  (as the error messages
would be repeated when the text is re-scanned later).  Hence,
numbers are scanned by a special function 'scan_num_and_kind_param ()'
that doesn't increment 'end_tk' or give error messages, but
instead indicates whether messages should be generated.  If so,
and if the text IS a complex number, then 'read_num_or_kind_string ()'
is used in a second scan over the components.
----------------------------------------------------------------*/
{
	register char *ptr;
	register int   c;
	char  *s,  *real_part,  *imag_part;

	
	/*{{{  scan '( real,  imag )';  goto 'not_a_number' if no match*/
	/*--------------------------------------------------------------*/
	/* Scan '( real , imag )';  goto 'not_a_number' if no match.	*/
	/*--------------------------------------------------------------*/
	ptr = startp;		/* points to '(' */
	
	
	/*-----------*/
	/* real part */
	/*-----------*/
	get_next_nonwhite (c, ptr);
	
	if (c == '+' || c == '-') get_next_nonwhite (c, ptr);
	
	if ( ! (isdigit (c) || (c == '.' && isdigit (*(ptr+1))) ) )
		goto  not_a_number;
	
	ptr = scan_num_and_kind_param (ptr, &real_part);
	
	if ( ! ptr) goto  not_a_number;
	
	get_nonwhite_from_here (c, ptr);
	
	if (c != ',')  goto  not_a_number;
	
	
	/*-----------*/
	/* imag part */
	/*-----------*/
	get_next_nonwhite (c, ptr);
	
	if (c == '+' || c == '-') get_next_nonwhite (c, ptr);
	
	if ( ! (isdigit (c) || (c == '.' && isdigit (*(ptr+1))) ) )
		goto  not_a_number;
	
	ptr = scan_num_and_kind_param (ptr, &imag_part);
	
	if ( ! ptr) goto  not_a_number;
	
	get_nonwhite_from_here (c, ptr);
	
	if (c != ')')  goto  not_a_number;
	
	
	/*------------------------------------------------------*/
	/* ...now 'ptr' points to the final ')' character.	*/
	/*------------------------------------------------------*/
	/*}}}*/

	/*{{{  Found a complex number!  Now re-scan parts if necessary.*/
	/*--------------------------------------------------------------*/
	/*	Found a complex number!  Now check whether the real or	*/
	/* imaginary parts have to be re-scanned (either because they	*/
	/* contain a lexical error or because they have a kind		*/
	/* parameter that needs to be checked).  If so, call		*/
	/* 'read_num_or_kind_string ()' to process them.		*/
	/*--------------------------------------------------------------*/
	if (real_part)
		read_num_or_kind_string (real_part, FALSE);
	
	if (imag_part)
		read_num_or_kind_string (imag_part, FALSE);
	/*}}}*/

	/*----------------------------------------------*/
	/* Save the complex constant in a string.	*/
	/*----------------------------------------------*/
	end_tk = ++ptr;		/* to point to the char after ')' */

	delimit_word_in_textbuf (c, ptr);
	s = emalloc (ptr - startp + 1);		/* +1 for the '\0' */
	strcpy (s, startp);
	restore_textbuf (c, ptr);

	yylval.str = s;

	return  TK_COMPLEX_LIT_CONST;

not_a_number :
	end_tk = startp + 1;
	return  '(';
}
/*}}}*/


/*--------------------------------------------------------------*/
/* Ancilliary token reading functions, not called directly by	*/
/* 'yylex ()', but rather from other token-reading functions.	*/
/*--------------------------------------------------------------*/
/*{{{  int  read_num (startp, val_p)  */
int  read_num (startp, val_p)
/*-------------------------*/
	char  *startp;		/* pointer to the start of a number */
	int   *val_p;
/*----------------------------------------------------------------
	Called by 'read_num_or_kind_string ()' and
'read_and_check_kind_param ()' to read a real or integer number,
***excluding any kind parameter*** (i.e. a '_' and any subsequent text).
	On input 'startp' points to a digit or a '.' followed
by a digit, i.e. a valid first character of a number.  Before
returning the function sets the global pointer 'end_tk' to point
to the char after the number.  It returns  TK_INT_LIT_CONST,
TK_REAL_LIT_CONST or TK_DP_LIT_CONST as appropriate, and for
the first of these returns the integer's value in '*val_p'.

N.B. A small complication is that the number may be an integer
immediately followed by a 'dotted name' (e.g. "2.EQ."), which
may look like a real number with an exponent until the 2nd char
after the '.' is read!
----------------------------------------------------------------*/
{
	register char  *ptr = startp;
	register int    c = *ptr;
	int   token,  val;

	/*{{{  scan the number and evaluate its integral part*/
	/*--------------------------------------------------------------*/
	/* Scan the number and put the value of its integer part in	*/
	/* 'val'.							*/
	/*--------------------------------------------------------------*/
	/*---------------*/
	/* Integer part  */
	/*---------------*/
	token = TK_INT_LIT_CONST;
	val = 0;
	while (isdigit (c)) {
		val = 10 * val + (c - '0');
		c = *++ptr;
	}
	
	/*--------------------------------------------------------------*/
	/* Fractional part, or possibly a 'dotted_name' (e.g. ".EQ.").	*/
	/* We assume the latter if the '.' is followed by:		*/
	/*		letter   (letter | '.')				*/
	/* If so quit scanning the number.  ('ptr' will subsequently be	*/
	/* reset to point to the '.').					*/
	/*--------------------------------------------------------------*/
	if (c == '.') {
		if (isalpha (*(ptr+1))  &&  (isalpha (*(ptr+2)) || *(ptr+2) == '.'))
			goto  finished_num;
	
		token = TK_REAL_LIT_CONST;
		do
			c = *++ptr;
		while (isdigit (c));
	}
	
	/*---------------*/
	/* Exponent part */
	/*---------------*/
	if (islower (c)) c = *ptr = toupper (c);
	if (c == 'D' || c == 'E') {
		token = (c == 'D') ? TK_DP_LIT_CONST : TK_REAL_LIT_CONST;
		c = *++ptr;
		if (c == '+' || c == '-')
			c = *++ptr;
	
		/* scan exponent digits, using 'val' to count them! */
		for (val = 0;   isdigit (c);   val++, c = *++ptr)
			;
	
		if ( ! val) {
			int    endc;
	
			delimit_word_in_textbuf (endc, ptr);
			errmsg ("incomplete exponent: %q", startp);
			restore_textbuf (endc, ptr);
		}
	}
	/*}}}*/

finished_num :
	end_tk = ptr;	/* to point to the next char after the number.	*/

	if (token == TK_INT_LIT_CONST)
		*val_p = val;

	return  token;
}
/*}}}*/
/*{{{  int  read_boz_num (startp, base) */
int  read_boz_num (startp, base)
/*----------------------------*/
	char  *startp;		/* pointer to the opening string delimiter */
	int    base;
/*--------------------------------------------------------------*/
/*	Called by 'read_word_or_boz_or_kind_string ()' to read	*/
/* a BOZ literal constant, i.e.:				*/
/*		binary:		B'...'  or  B"..."		*/
/*		octal:		O'...'  or  O"..."		*/
/*	    or	hexadecimal:	Z'...'  or  Z"...".		*/
/*								*/
/*	On entry, 'base' is the base (2 for binary, 8 for octal,*/
/* 16 for hex), and 'startp' points to the opening delimiter of	*/
/* the string (a quote or apostrophe).				*/
/*	On exit, the equivalent integer value is placed in	*/
/* 'yylval.num', the global pointer 'end_tk' points to the next	*/
/* char after the BOZ constant, and TK_BOZ_LIT_CONST is		*/
/* returned.							*/
/*--------------------------------------------------------------*/
{
	register char *ptr;
	register int   c,  delim;
	char *str_base,  *num = startp - 1;	/* 'num' points to B,O or Z */
	int   n_chars,  n_illegal_chars,  val;


	/*{{{  initialise 'delim', 'num' & 'str_base' (e.g. "binary")*/
	delim = *startp;
	num   = startp - 1;	/* 'num' points to B,O or Z */
	
	switch (base) {
	case  2 :	str_base = "binary";		break;
	case  8 :	str_base = "octal";		break;
	case  16 :	str_base = "hexadecimal";	break;
	}
	/*}}}*/
	/*{{{  read the BOZ constant*/
	/*--------------------------------------------------------------*/
	/* Read the BOZ constant, keeping count of any illegal chars.	*/
	/*--------------------------------------------------------------*/
	n_chars = n_illegal_chars = val = 0;
	
	for (ptr = startp;    c = *++ptr;    n_chars++) {
		if (islower (c))
			c = *ptr = toupper (c);
	
		if (isdigit (c)  &&  (c - '0') < base)
			/*-----------------------*/
			/* a legal decimal digit */
			/*-----------------------*/
			val = base * val + (c - '0');
	
		else if ((base == 16)  &&  (c >= 'A')  &&  (c <= 'F'))
			/*----------------------*/
			/* a hex digit (A-F)	*/
			/*----------------------*/
			val = 16 * val + (10 + c - 'A');
	
		else if (c == delim) {
			/*----------------------------*/
			/* end of string -- quit loop */
			/*----------------------------*/
			ptr++;
			break;
		}
		else
			n_illegal_chars++;
	}
	/*}}}*/

	end_tk = ptr;		/* points to char after the BOZ string */

	/*-----------------------------*/
	/* write errmsgs if necessary! */
	/*-----------------------------*/
	if ( ! c) {
		/*{{{  errmsg: no closing delimiter*/
		char  *endp = num + 6;
			
		delimit_word_in_textbuf (c, endp);
		
		errmsg ("%s constant string %s%q has no closing delimiter (%c)",
		         str_base,
		         ((ptr - num <= 6) ? "" : "beginning "),  num,  delim);
		
		restore_textbuf (c, endp);
		/*}}}*/
	}
	else if ( ! n_chars) {
		/*{{{  errmsg: no digits*/
		delimit_word_in_textbuf (c, ptr);
		
		errmsg ("%s constant has no digits: %q", str_base, num);
		
		restore_textbuf (c, ptr);
		/*}}}*/
	}
	else if (n_illegal_chars) {
		/*{{{  errmsg: illegal characters*/
		delimit_word_in_textbuf (c, ptr);
		
		errmsg ("%s constant %q contains %s",
		         str_base,  num,  ((n_illegal_chars == 1) ?
		                          "an illegal character" : "illegal characters") );
		
		restore_textbuf (c, ptr);
		/*}}}*/
	}
	
	yylval.num = val;
	return  TK_BOZ_LIT_CONST;
}
/*}}}*/
/*{{{  read_and_check_kind_param (startp, tk_lit_const, strval) */
read_and_check_kind_param (startp, tk_lit_const, strval)
/*----------------------------------------------------*/
	char  *startp;		/* pointer to the start of the kind param */
	int    tk_lit_const;	/* token   } ... of the literal whose */
	char  *strval;		/* string  } kind param is being read */
/*----------------------------------------------------------------
	Called by 'read_num_or_kind_string ()' & 'lookup_dotted_keyword()'
to read and check a kind parameter following a '_' at the end
of a literal constant.  The type of literal constant is given by
'tk_lit_const', and may be TK_INT_LIT_CONST, TK_REAL_LIT_CONST,
TK_DP_LIT_CONST or TK_LOGICAL_LIT_CONST.
	On entry, 'startp' points to the first character of the
kind parameter (which hasn't been validated yet!).  The function
sets the global pointer 'end_tk' to point to the next character
after the kind parameter.

N.B.:
--  We haven't determined yet whether there actually *is* a kind
parameter -- we only know that there should be one!
--  *(startp-1), which was '_', has been set to '\0' by the caller.
This terminates '*strval', the literal constant whose kind parameter
is being read (so it can be used in errmsgs).  On return, the caller
must restore this '_'.
----------------------------------------------------------------*/
{
	register char *ptr;
	register int   c;
	int32  type;		/* type flag */

	switch (tk_lit_const) {
	case  TK_INT_LIT_CONST :	type = TY_INTEGER;	break;
	case  TK_REAL_LIT_CONST :	type = TY_REAL;		break;
	case  TK_LOGICAL_LIT_CONST :	type = TY_LOGICAL;	break;
	}

	ptr = startp;
	c   = *ptr;

	if ( isdigit (c) || (c == '.'  &&  isdigit (*(ptr+1))) ) {
		/*{{{  read & check literal constant kind parameter*/
		/*---------------------------------------------------*/
		/* Kind value given by a number.  Read and check it. */
		/*---------------------------------------------------*/
		int   token,  val;
		
		token = read_num (ptr, &val);	/* this sets global pointer 'end_tk'	*/
						/* to point to the char after the number*/
		
		if (tk_lit_const == TK_DP_LIT_CONST)
			/*------------------------------------------------------*/
			/* It's illegal to specify a kind parameter for a	*/
			/* double precision number, and the caller has already	*/
			/* given an errmsg to this effect.  Hence don't do any	*/
			/* other checks or report other errors in this case.	*/
			/*------------------------------------------------------*/
			;
		else if (token == TK_INT_LIT_CONST)
			check_kind_val (type, val);
		
		else {
			delimit_word_in_textbuf (c, end_tk);
			errmsg ("invalid kind parameter %q; it must be an integer",
			         startp);
			restore_textbuf (c, end_tk);
		}
		/*}}}*/
	}
	else if (isalpha (c)) {
		/*{{{  read & check named kind parameter*/
		/*--------------------------------------------------------------*/
		/*	Kind value given by an identifier (which should be a	*/
		/* scalar integer named constant!).  Read and check it.		*/
		/*	The reading phase is basically the same as in		*/
		/* 'read_word_or_boz_or_kind_string ()', except that:		*/
		/* -- it doesn't allow for 'name' being a letter introducing	*/
		/* a BOZ literal constant, or a kind parameter for a subsequent	*/
		/* string;							*/
		/* -- 'endp' and 'endc' are *local*.				*/
		/*--------------------------------------------------------------*/
		register unsigned  hashval;
		char    *endp;
		int      endc,  len;
		Symbol  *sp;
		
		/*{{{  scan the name and put '\0' after its last char in 'textbuf'*/
		/*--------------------------------------------------------------*/
		/*	Scan the name in 'textbuf' and put a '\0' delimiter	*/
		/* immediately after it (at position 'endp'), saving the	*/
		/* overwritten char in 'endc' to be restored before the function*/
		/* returns.  If the name is too long (> MAX_ID_LEN chars), put	*/
		/* the '\0' after the first MAX_ID_LEN chars and give a warning,*/
		/* but still scan 'textbuf' to the actual end of the		*/
		/* (overlength) identifier.					*/
		/*								*/
		/* N.B. This is identical to 'read_word_or_boz_or_kind_string()'*/
		/* except that:							*/
		/* (i) 'endp' and 'endc' are *local*;				*/
		/* (ii) we don't allow for 'name' being followed by _"string"	*/
		/*	(as it's the kind param of a preceding literal constant,*/
		/*	not of a subsequent string constant!).			*/
		/*--------------------------------------------------------------*/
		if (islower (c)) c = *ptr = toupper (c);
		hashval = 0;
		
		do {
			hashval = c + (31 * hashval);
			c = *++ptr;
			if (islower (c)) c = *ptr = toupper (c);
		} while (isidchar (c));
		
		hashval = hashval % HASHSIZE;
		
		end_tk = endp = ptr;	/* to point to 'c', the next char after 'name'.	*/
					/* ('end_tk' is a global pointer).		*/
		
		len = ptr - startp;			
		if (len <= MAX_ID_LEN)
			delimit_word_in_textbuf (endc, endp);
		
		else {
			/*------------------------------------------------------*/
			/* 'name' is too long!  Give a warning and rescan it,	*/
			/* putting '\0' after MAX_ID_LEN chars and re-computing	*/
			/* 'hashval' for the newly-defined word, but leaving	*/
			/* 'c' and 'end_tk' unchanged.				*/
			/*------------------------------------------------------*/
			len = MAX_ID_LEN;
			endp = startp + MAX_ID_LEN;
			delimit_word_in_textbuf (endc, endp);
			warning ("identifier is too long; only %d characters are significant (viz. %q)",
			          MAX_ID_LEN,  startp);
		
			for (ptr = startp,  hashval = 0;   ptr < endp;   ptr++)
				hashval = *ptr + (31 * hashval);
		
			hashval = hashval % HASHSIZE;
		}
		/*}}}*/
		
		sp = lookup_or_install_id (startp, len, hashval);	/* Don't put 'sp'   */
									/* on parser stack! */
		
		if (tk_lit_const != TK_DP_LIT_CONST)
			/*------------------------------------------------------*/
			/* N.B. It's illegal to specify a kind param for a dbl	*/
			/* precision no, and we've already given an errmsg to	*/
			/* this effect.						*/
			/*------------------------------------------------------*/
			check_kind_name (type, sp);
		
		restore_textbuf (endc, endp);
		/*}}}*/
	}
	else {
		/*----------------------------------------------*/
		/* Missing kind parameter!  Set global pointer	*/
		/* 'end_tk' to point to 'startp' (as nothing	*/
		/* has been read!).				*/
		/*----------------------------------------------*/
		end_tk = ptr;
		errmsg ("expected a kind parameter (an integer literal or named constant) after \'%s_\'",
		         strval);
	}
}
/*}}}*/

/*{{{  char  *scan_num_and_kind_param (startp, startp_p) */
char  *scan_num_and_kind_param (startp, startp_p)
/*---------------------------------------------*/
	char  *startp,  **startp_p;
/*----------------------------------------------------------------
	Called by 'read_complex_num_or_paren ()' to scan a potential
number  (i.e. to check that it's indeed a number and that it's
syntactically correct).
	On entry, 'startp' points to the first character of the
candidate number (a digit or '.'), and '*startp_p' is uninitialised.
	If the token IS a number (even with syntax errors)
the function returns a pointer to the character immediately
following it, otherwise it returns NULL (e.g. if it's a kind value
preceding a string, or is immediately followed by a 'dotted' operator).
The return value of '*startp_p' indicates whether the number has
to be re-scanned by 'read_num_or_kind_string ()' in order to report a
syntax error or check a kind parameter.  If so, then on return
'*startp_p' points to the start of the number, otherwise it is NULL. 

N.B.:  This function differs from 'read_num_or_kind_string ()' in
that it *doesn't*:
-- reset the global pointer 'end_tk';
-- evaluate the integral part of the number;
-- write any error messages, or check the validity or syntactic
	correctness of any kind parameter.
----------------------------------------------------------------*/
{
	register char  *ptr = startp;
	register int    c = *ptr;
	bool  found_kind_param = FALSE,  must_rescan = FALSE;

scan_num :
	/*{{{  scan the number*/
	/*----------------------*/
	/* Scan the number...	*/
	/*----------------------*/
	/*---------------*/
	/* Integer part  */
	/*---------------*/
	while (isdigit (c))
		c = *++ptr;
	
	/*--------------------------------------------------------------*/
	/* Fractional part, or possibly a 'dotted_name' (e.g. ".EQ.").	*/
	/* We assume the latter if the '.' is followed by:		*/
	/*		letter   (letter | '.')				*/
	/* If so return NULL, as this can't be part of a complex	*/
	/* number.							*/
	/*--------------------------------------------------------------*/
	if (c == '.') {
		if (isalpha (*(ptr+1))  &&  (isalpha (*(ptr+2)) || *(ptr+2) == '.' ))
			return  NULL_str;
		do
			c = *++ptr;
		while (isdigit (c));
	}
	
	/*---------------*/
	/* Exponent part */
	/*---------------*/
	if (islower (c)) c = *ptr = toupper (c);
	if (c == 'D' || c == 'E') {
		int   exp_len;		/* # digits in the exponent */
	
		if ((c = *++ptr) == '+' || c == '-')
			c = *++ptr;
	
		for (exp_len = 0;   isdigit (c);   exp_len++,  c = *++ptr)
			;
	
		if ( ! exp_len)
			must_rescan = TRUE;	/* syntax error! */
	}
	/*}}}*/
	
	/*------------------------------------------------------*/
	/*  If '_' follows 'num', it may be followed by a kind	*/
	/* parameter (ok) or a string (return NULL).		*/
	/* (N.B. 'c' is now the char after the number, and	*/
	/* 'ptr' points to the char after 'c').			*/
	/*------------------------------------------------------*/
	if (c == '_'  &&  !found_kind_param) {
		/*{{{  look beyond the '_' (for kind param, etc)*/
		
		if ((c = *++ptr) == '\'' || c == '\"')
			/*------------------------------------------------------*/
			/* 'num' ISN'T a number in its own right, but rather	*/
			/* the kind value for a string, and syntactically forms	*/
			/* part of the string token.  Hence return NULL		*/
			/* (meaning number not found!)				*/
			/*------------------------------------------------------*/
			return  NULL_str;
		
		
		/*----------------------------------------------------------------
		'num' should be followed by a kind parameter (numeric or named),
		which is scanned here.  'must_rescan' is set to TRUE, so the kind
		parameter's validity and syntax are checked by the caller.
		
		N.B.  If the text immediately after '_' is neither a name nor a
		number, it's interpreted as a missing kind parameter!  This function
		still returns a non_NULL result (as a number with a missing kind
		parameter is interpreted as a number!), and since 'must_rescan' is
		TRUE a suitable error message will be given when the number is
		re-scanned!
		----------------------------------------------------------------*/
		must_rescan = found_kind_param = TRUE;
		
		if (isdigit (c) || c == '.')
			/*------------------------------*/
			/* a numeric kind parameter	*/
			/*------------------------------*/
			goto  scan_num;
		
		else if (isalpha (c))
			/*------------------------------*/
			/* a named kind parameter	*/
			/*------------------------------*/
			while ((c = *++ptr) == '_' || isalnum (c))
				;
		/*}}}*/
	}

	*startp_p = must_rescan ? startp : NULL_str;
	
	return  ptr;
}
/*}}}*/


/*------------------------------*/
/* Keyword lookup functions	*/
/*------------------------------*/
/*{{{  int  lookup_keyword (word, endc, hashval) */
int  lookup_keyword (word, endc, hashval)
/*-------------------------------------*/
	char  *word;		/* pointer to the start of the word */
	int    endc;		/* the char after 'word' */
	unsigned   hashval;
/*----------------------------------------------------------------
	Called by 'read_word_or_boz_or_kind_string ()' to search
for 'word' (which *isn't* dotted) in the keyword table.  If 'word'
might be the first component of a split keyword (eg, END DO), then
the next word of input is scanned to check this, and this may be
repeated to allow for keywords that can be split into several parts
(currently only 'END BLOCK DATA').
	If 'word' IS a keyword, its token value is returned,
otherwise 0 is returned.
	On entry, the global pointer 'end_tk' points to the
char after 'word', which is temporarily replaced by '\0' (its
original value being 'endc').  If a multiple keyword is found,
'end_tk' is reset accordingly.

N.B. This function may read more input (for a split keyword)
which is why it's in file 'lex.c' rather than 'symbol.c'.
If a split keyword is found, the global variable 'end_tk' is reset.
However, 'endc' is *not* restored in 'textbuf' -- the calling
function does that.  Also, no more '\0's are placed in 'textbuf'.
----------------------------------------------------------------*/
{
	register Keyword  *kp;

	for (kp = keytab [hashval];  kp;  kp = kp->next)
		if (str_match (kp->name, word))		/* found! */
			break;

	if ( !kp)	/* keyword not found */
		return  0;

	else if (kp->maybe_split  &&  (endc == ' ' || endc == '\t')) {
		/*{{{  check whether 'word' is part of a 'split' keyword*/
		/*--------------------------------------------------------------*/
		/* Since 'kp->maybe_split' == TRUE, 'word' may be the first part*/
		/* of a split keyword.  To check this, copy it to a temporary	*/
		/* buffer ('wordbuf'), append the next word to it, and search	*/
		/* for the composite word in the keyword table.  This procedure	*/
		/* may be repeated in case a keyword is split into 3 parts	*/
		/* (which is currently only possible for 'END BLOCK DATA'!).	*/
		/*								*/
		/* We assume that:						*/
		/* (i) 'word' is currently terminated by '\0' in 'textbuf'	*/
		/*	('endc' is the original character in this position);	*/
		/* (ii) keywords contains only *letters*.			*/
		/*--------------------------------------------------------------*/
		register char  *from,  *to;
		register int    c;
		char   wordbuf [MAX_ID_LEN+1],
		      *end_wordbuf = (wordbuf + MAX_ID_LEN);	/* last char of 'wordbuf' */
		Keyword   *lkp;
		
		
		/*{{{  copy 'word' to 'wordbuf', and set 'from' & 'to'*/
		/*--------------------------------------------------------------*/
		/*	Copy the first component ('word') to a temporary buffer	*/
		/* ('wordbuf').							*/
		/*	When this has been done, 'from' and 'to' point to the	*/
		/* '\0' after 'word' in 'textbuf' and 'wordbuf' respectively.	*/
		/* 'endc' is the (whitespace) character formerly at position	*/
		/* 'from' in 'textbuf'.						*/
		/*--------------------------------------------------------------*/
		for (from = word-1,  to = wordbuf-1;   *++to = *++from;   )
			;
		/*}}}*/
		
		
		append_next_word :	/* label to jump back to if >2 components! */
		
		/*{{{  append the next word to 'wordbuf'*/
		/*--------------------------------------------------------------*/
		/* append the next word from 'textbuf' to the end of the word	*/
		/* in 'wordbuf' (up to MAX_ID_LEN chars).			*/
		/*--------------------------------------------------------------*/
		get_next_nonwhite (c, from);	/* N.B. we know that initially	*/
						/* '*from' is whitespace.	*/
		if (islower (c))
			c = *from = toupper (c);
		else if (! isupper (c))
			goto end;		/* not a word */
		
		do {
			*to++ = c;
			c = *++from;
			if (islower (c)) c = *from = toupper (c);
		} while (isupper (c)  &&  (to < end_wordbuf));
		
		if (isupper (c)) goto end;	/* combined word too long for a keyword! */
		
		/*--------------------------------------------------------------*/
		/* Now 'from and 'to' point to the char after the last word in	*/
		/* 'textbuf' and 'wordbuf' respectively.			*/
		/*--------------------------------------------------------------*/
		*to = '\0';		/* terminate new word in 'wordbuf'*/
		/*}}}*/
		
		/*--------------------------------------------------------------*/
		/* search for the new 'word' in the keyword table (using local	*/
		/* pointer 'lkp').						*/
		/*--------------------------------------------------------------*/
		for (lkp = keytab [hash (wordbuf)];  lkp;  lkp = lkp->next) {
			if (str_match (lkp->name, wordbuf)) {
				
				/*{{{  found a new 'keyword';  if it's complete, update 'kp' etc,*/
				/*--------------------------------------------------------------*/
				/* Found a new composite 'keyword'.  If it's complete (i.e. not	*/
				/* 'END BLOCK' -- see below), update 'kp', 'end_tk', etc.	*/
				/*--------------------------------------------------------------*/
				if ( ! lkp->maybe_split ) {
					/*------------------------------------------------------*/
					/* Found a new complete keyword!!  Update 'kp', 'end_tk'*/
					/* and 'endc'.						*/
					/*------------------------------------------------------*/
					kp = lkp;
					end_tk =  from;
					endc   = *from;		/* passed to 'keyword_fn ()' */
				}
				else if (c == ' ' || c == '\t')
					/*------------------------------------------------------*/
					/* The dreaded special case!  Found 'END BLOCK'		*/
					/* followed by a space, which may be 'END BLOCK DATA',	*/
					/* so append the next word and search the keytab again.	*/
					/*							*/
					/* N.B.							*/
					/* (1)  Don't change 'kp' or 'end_tk' yet, since this	*/
					/* 'lkp->token' == 0.  If we don't find the 3rd		*/
					/* component, we just return the keyword 'END'.		*/
					/*							*/
					/* (2)  If 'END BLOCK' *isn't* followed by a space,	*/
					/* again 'kp', 'end_tk', etc, are upchanged and keyword	*/
					/* 'END' is returned.  This is sensible, as 'END BLOCK'	*/
					/* isn't a keyword and we want 'yyparse ()' to treat	*/
					/* 'BLOCK' as an identifier (although it's a syntax	*/
					/* error!)						*/
					/*------------------------------------------------------*/
					goto  append_next_word;	
				/*}}}*/
				break;		/* exit 'for' loop! */
			}
		}
			
		/*}}}*/
	}
	
end:
	/*----------------*/
	/* Found keyword! */
	/*----------------*/
	if (kp->keyword_fn)
		nwords_to_keyword = (*(kp->keyword_fn)) (endc);

	/* ... else 'nwords_to_keyword' remains 0 (i.e. no more keywords!) */

	if (kp->token == TK_READ) {
		/*{{{  return 'TK_READ' if next nonwhite is '(', else 'TK_UREAD'*/
		/*----------------------------------------------------------------
		If the next nonwhite is '(' return 'TK_READ', otherwise return
		'TK_UREAD'  (for 'unbracketed READ', in which case we expect it
		to be followed by a 'format', not in parentheses).
		
		N.B. This ugly special case is mandated by the ambiguity of the
		alternatives for 'read_stmt", viz:
		
		read-stmt  =  READ  (  io-control-spec-list  )  [input-item-list]
		           |  READ  format  [,  input-item-spec-list ]
		
		where 'format' can be a char-expr, and thus bracketed!
		----------------------------------------------------------------*/
		register char  *p;
		register int    c;
		
		
		if (endc == '(')
			return  TK_READ;
		
		else if ( ! (endc == ' ' || endc == '\t'))
			return  TK_UREAD;		/* i.e. 'READ' isn't followed by '(' */
		
		p = end_tk;		/* N.B. we know that initially '*p' is whitespace */
		get_next_nonwhite (c, p);
		
		return  ((c == '(') ? TK_READ : TK_UREAD);
		/*}}}*/
	}	
	return  kp->token;
}
/*}}}*/
/*{{{  int  lookup_dotted_keyword (word, endc, hashval) */
int  lookup_dotted_keyword (word, endc, hashval)
/*--------------------------------------------*/
	char  *word;		/* pointer to the start of the word */
	int   endc;		/* the char after 'word' */
	unsigned   hashval;
/*----------------------------------------------------------------
	Called by 'yylex ()' or 'read_dotted_word ()' to search
for a 'dotted word' (e.g. ".LT.") in the keyword table.  If 'word'
IS a keyword, its token value is returned, otherwise 0 is returned.
	On entry, the global pointer 'end_tk' points to the char
after 'word', which is temporarily replaced by '\0'.  Its original
value is given by 'endc'.

N.B. This function differs from 'lookup_keyword()' -- for 'normal'
keywords -- in several respects:
(i) it's assumed that a 'dotted keyword' can't be split, in
contrast to some normal keywords (e.g. 'END DO');
(ii) 'nwords_to_keyword' isn't reset, as dotted keywords normally
occur in contexts where other keywords can't appear, viz. in
expressions (often within parentheses).
----------------------------------------------------------------*/
{
	register Keyword  *kp;

	for (kp = keytab [hashval];  kp;  kp = kp->next) {
		if (str_match (kp->name, word)) {
			/*----------------*/
			/* found keyword! */
			/*----------------*/
			switch (kp->token) {
			case TK_REL_OP :
				yylval.str = kp->name;
				break;
		
			case TK_LOGICAL_LIT_CONST :
				yylval.str = kp->name;
				if (endc == '_')
					read_and_check_kind_param (end_tk + 1,
					       TK_LOGICAL_LIT_CONST, kp->name);
				break;
			}
			return  kp->token;
		}
	}

	return  0;	/* keyword not found */
}
/*}}}*/
/*}}}*/

/*{{{  <<< functions pointed to by 'kp->keyword_fn' >>> */
/*--------------------------------------------------------------*/
/* Functions pointed to by 'kp->keyword_fn'.			*/
/*	These return a value for 'nwords_to_keyword', which	*/
/* prescribes how many words ahead the next keyword (as opposed	*/
/* to ID) is to be found, and, if appropriate, set (scope_type	*/
/* = SC_LOCAL) so that subsequent ID's have local scope numbers	*/
/* (i.e. they are *declared* rather than referenced).		*/
/*--------------------------------------------------------------*/
/*--------------*/
/* for F90...	*/
/*--------------*/
/*{{{  int  attr_keyword (next_char)  -- calls 'set_local_scope ()'*/
int  attr_keyword (next_char)
/*-------------------------*/
	int  next_char;		/* char originally at 'end_tk' */
/*----------------------------------------------------------------
	Called by 'lookup_keyword ()' after an 'attribute' keyword
'attr' (e.g. DIMENSION, PARAMETER) has been read.  It returns a
value for 'nwords_to_keyword', which depends on the text
following 'attr' as follows:

 text					nwords_to_keyword
 ----					--------------
 "attr [(..)] ,"			1
 anything else (e.g. "attr ::")		0

	It also calls 'set_local_scope ()' to set (scope_type =
SC_LOCAL), so that subsequent ID's have local scope numbers
(i.e. they're *declared* rather than referenced).

N.B.:
-----
-- The global variable 'end_tk' points to the position in
'textbuf' immediately after the 'attr' keyword.  'next_char'
is the character originally stored there, which has been
temporarily replaced by '\0'.  'end_tk' isn't changed, as
the input is just inspected, not read.

-- The full list of 'attr' keywords is as follows:
	ALLOCATABLE,  DIMENSION,  EXTERNAL,  INTENT,  INTRINSIC,
	OPTIONAL,  PARAMETER,  POINTER,  PRIVATE,  PUBLIC,  SAVE,
	TARGET,  DYNAMIC,  INHERIT,  PROCESSORS,  TEMPLATE,
----------------------------------------------------------------*/
{
	register char  *ptr;
	register int   c;

	set_local_scope ();

	/*{{{  set 'c' & 'ptr' to next non-blank after 'attr'*/
	/*------------------------------------------------------*/
	/* set 'c' to the next non-white char after 'attr', and	*/
	/* 'ptr' to point to its posn in 'textbuf'.		*/
	/*------------------------------------------------------*/
	ptr = end_tk;	/* N.B. 'next_char' is original value of '*ptr' */
	
	if (next_char == ' ' || next_char == '\t')
		get_next_nonwhite (c, ptr);
	else
		c = next_char;
	/*}}}*/

	if (c == '(')
		get_nonwhite_after_parens (c, ptr);

	return  (c == ',');		/* 1 if char is ',', else 0 */
}
/*}}}*/
/*{{{  int  local_scope (next_char)   -- calls 'set_local_scope ()'*/
int  local_scope (next_char)
/*------------------------*/
	int  next_char;		/* char originally at 'end_tk' */
/*----------------------------------------------------------------
	Called by 'lookup_keyword ()' to set (scope_type = SC_LOCAL),
so that subsequent ID's have local scope numbers (i.e. they're
*declared* rather than referenced).
	It returns the value 0 for 'nwords_to_keyword'.
----------------------------------------------------------------*/
{
	set_local_scope ();

	return  0;
}
/*}}}*/
/*{{{  int  type_keyword (next_char)  -- calls 'set_local_scope ()'*/
int  type_keyword (next_char)
/*-------------------------*/
	int  next_char;		/* char originally at 'end_tk' */
/*{{{  EXPLANATION*/
/*----------------------------------------------------------------
	Called by 'lookup_keyword ()' after a 'type' keyword (viz.
CHARACTER,  COMPLEX,  DOUBLE PRECISION,  INTEGER,  LOGICAL,  REAL
or TYPE) has been read.  It returns a value for 'nwords_to_keyword',
which depends on the text following 'type' as follows, where:

	star  =  "* (..)"
	      |  "* int_lit_const"

text					nwords_to_keyword
----					--------------
"type [(..) [(..)]] ,"			1  (In a 'type-decl', where it should
					    be followed by an 'attr', or an
					    'impl-stmt', followed by other
					    'impl-spec's)
"type star { (..)"			1  (In an 'impl-stmt')
           { , attr-list ::"		1  (In a 'type-decl', followed by attr)

	N.B. In the last case we must look for "::".  If it isn't
	present then ',' is followed by an 'id_list', not an 'attr_list'!
	(Incidentally, 'type' should be 'CHARACTER', but we don't
	bother to check for that!)

"type [(..) | star] word [(..)] letter"	1  (In a 'func-stmt'.  'word' may be
					    'FUNCTION' or a 'prefix', viz.
					     RECURSIVE, PURE or EXTRINSIC(..))
For anything else:			0

	It also calls 'set_local_scope ()' to set (scope_type =
SC_LOCAL), so that subsequent ID's have local scope numbers
(i.e. they're *declared* rather than referenced).

N.B.:
----
--  The global variable 'end_tk' points to the position in
'textbuf' immediately after the 'type' keyword.  'next_char'
is the character originally stored there, which has been
temporarily replaced by '\0'.  'end_tk' isn't changed, as the
input is just inspected, not read.

--  In a 'type-decl', the 'star' form (i.e. the old form of
CHARACTER length selector, e.g. *10, or *(10), or *(*)),
is complicated because an immediately following comma doesn't
necessarily signal an ensuing 'attr_list';  it could also be
followed by an 'id_list'.  In this case, to distinguish an 'attr_list'
(nwords_to_keyword = 1) from an 'id_list' (nwords_to_keyword = 0),
we must search for "::"  -- which may mean scanning to the EOL
if it isn't present!
-----------------------------------------------------------------*/
/*}}}*/
{
	register char  *ptr;
	register int    c;
	bool   char_star;

	set_local_scope ();

	/*{{{  set 'c' and 'ptr' to next non-blank after 'type'*/
	/*--------------------------------------------------------------*/
	/* set 'c' to the next non-white char after 'type', and 'ptr'	*/
	/* to point to its posn in 'textbuf'.				*/
	/*--------------------------------------------------------------*/
	ptr = end_tk;		/* N.B. 'next_char' is original value of '*ptr' */
	
	if (next_char == ' ' || next_char == '\t')
		get_next_nonwhite (c, ptr);
	else
		c = next_char;
	/*}}}*/

	/*{{{  scan "[(..) | *int-num | *(..)]", & set 'char_star' for last two*/
	/*--------------------------------------------------------------*/
	/* Scan "[ (..) | * int-num | *(..) ]".				*/
	/* In the last two cases set (char_star = TRUE).		*/
	/*--------------------------------------------------------------*/
	char_star = FALSE;
	
	if (c == '*') {
		char_star = TRUE;
		get_next_nonwhite (c, ptr);
	}
	
	if (c == '(')
		get_nonwhite_after_parens (c, ptr);
	
	else if (char_star  &&  isdigit (c)) {
		/*------------------------------------------------------*/
		/* Skip to the next non-blank after the int literal	*/
		/* (e.g. "CHARACTER *10")				*/
		/*------------------------------------------------------*/
		do
			c = *++ptr;
		while (isdigit (c));
		get_nonwhite_from_here (c, ptr);
	}	
	/*}}}*/

	if (c == '(')
		/*------------------------------------------------*/
		/* "type ()() | type star ()";  in an 'impl-stmt' */
		/*------------------------------------------------*/
		return  1;

	else if (c == ',') {
		if ( ! char_star)
			/*------------------------------------------*/
			/* "type [(..)] ,";  in a 'type-decl'       */
			/* followed by an 'attr', or an 'impl-stmt' */
			/*------------------------------------------*/
			return 1;

		/*{{{  "type star ,";  in a 'type-decl' & must search for "::"*/
		/*{{{  EXPLANATION*/
		/*----------------------------------------------------------------
		Found:
			"type star ,"
		
		so it's a 'type-decl' (where 'type' should be "CHARACTER").
		
		However, the ',' may be followed by either an 'attr-list' (return 1)
		*OR* an 'id-list' (return 0)  (thanks to another Fortran anomaly, viz.
			"CHARACTER * {num | (...)}"
		can be followed by an optional comma in a 'type-decl'!)
		To distinguish these cases we must search for a "::", which
		always follows an 'attr-list' but never occurs in an 'id-list'.
		If it's found, return 1, otherwise return 0.
		
		More precisely, we scan the text following the ',', skipping over
		"(...)" (which can occur in both an 'attr-list' or an 'id-list').
		We call text outside of "(...)" 'bare' text.  Now, if in the
		bare text, we find:
		
			"::"	-- it's an 'attr-list';  return 1
			"="	-- it's an 'id-list' (since '=' precedes an initialiser);
				                 return 0.
			'\0'	-- found EOL without "::", so an 'id-list';  return 0.
		
		N.B.:
		~~~~~
		--  We must skip over "(...)", since within them we may have
		strings, within which anything may occur, including "::".  E.g.:
			CHARACTER *10, A (LEN ("::"))
		
		-- We must look for '=' (which can only occur in an 'id-list')
		since after this we have an initialiser, within which a "::"
		can occur, e.g.:
			CHARACTER *10, A = B (::2)
		
		or indeed a string can occur, which can contain anything, e.g.:
			CHARACTER *10, A = "::"
		
		So if we see '=', we know it's an 'id-list', so return 0.
		
		-- It's not sufficient to find a single ":".  This can occur in
		both an 'id-list', e.g.:
			CHARACTER *10, A (10:20)
		
		and an 'attr-list', e.g.:
			CHARACTER *10, DIMENSION (10:20) :: A
		
		and so doesn't distinguish the two cases.
		----------------------------------------------------------------*/
		/*}}}*/
		
		while (c = *++ptr) {
			switch (c) {
			case  ':' :
				if (*(ptr+1) == ':')
					return  1;        /* found "::", so its an 'attr-list'*/
				break;
		
			case  '(' :
				get_nonwhite_after_parens (c, ptr);
				ptr--;		/* ...so we test the nonwhite after "(..)" */
				break;
		
			case  '=' :
				return  0;		/* found "=", so it's an 'id-list' */
			}
		}
		return  0;         /* got to EOL without "::", so it's an 'id-list' */
		/*}}}*/
	}
	else if (isalpha (c)) {
		/*{{{  if "word [(..)] letter" return 1 (i.e. in a FUNCTION stmt)*/
		/*--------------------------------------------------------------*/
		/* Found:							*/
		/*	"type [star | (..)] letter"				*/
		/*								*/
		/* If this pattern continues as follows:			*/
		/*								*/
		/*	"word [(..)] letter"					*/
		/*								*/
		/* (where 'letter' in the 1st pattern is the first letter of	*/
		/* 'word' in the 2nd pattern) then return 1, else return 0.	*/
		/* In the former case we're in a FUNCTION stmt.			*/
		/*	Incidentally, in this case 'word' should be 'FUNCTION'	*/
		/* or a 'prefix', viz. 'RECURSIVE', 'PURE' or 'EXTRINSIC(..)',	*/
		/* but we needn't check for this.				*/
		/*--------------------------------------------------------------*/
		
		skip_word (c, ptr);  get_nonwhite_from_here (c, ptr);
		
		if (c == '(')
			get_nonwhite_after_parens (c, ptr);
		
		return  (isalpha (c) ? 1 : 0);		/* 1 for "word [(..)] letter" */
		/*}}}*/
	}
	else
		return  0;	/* the default */
}
/*}}}*/
/*{{{  int  uncond_keyword (next_char)*/
int  uncond_keyword (next_char)
/*---------------------------*/
	int  next_char;		/* char originally at 'end_tk' */
/*----------------------------------------------------------------
	Called by 'lookup_keyword ()' to unconditionally return
the value 1 for 'nwords_to_keyword' (cf. the default action
in 'lookup_keyword()' is to set it to 0).
	So far, this is only invoked for the following keywords:

	ASSIGN  (expecting 'TO');
	ELSEIF  (expecting 'THEN');
	RECURSIVE, PURE, EXTRINSIC (expecting type, 'FUNCTION', 'SUBROUTINE'
	                            or another keyword from this group).
	INDEPENDENT  (expecting either 'NEW' or '\n')
----------------------------------------------------------------*/
{
	return 1;
}
/*}}}*/
/*{{{  int  CASE_keyword (next_char)*/
int  CASE_keyword (next_char)
/*-------------------------*/
	int  next_char;		/* char originally at 'end_tk' */
/*----------------------------------------------------------------
	Called by 'lookup_keyword ()' after keyword "CASE" has
been read.  It returns a value for 'nwords_to_keyword', which
depends on the text following "CASE" as follows:

 text					nwords_to_keyword
 ----					--------------
 "CASE D"  (for "CASE DEFAULT")		1
 anything else  (e.g. "CASE (...)")	0

N.B. The global variable 'end_tk' points to the position in
'textbuf' immediately after the keyword "CASE".  'next_char'
is the character originally stored there, which has been
temporarily replaced by '\0'.  'end_tk' isn't changed, as
the input is just inspected, not read.
----------------------------------------------------------------*/
{
	register char  *ptr;
	register int   c;

	if ( ! (next_char == ' ' || next_char == '\t'))
		return  0;

	ptr = end_tk;
	get_next_nonwhite (c, ptr);
	
	return  (c == 'd' || c == 'D');		/* 1 if char is 'd' or 'D',
						   else 0 */
}
/*}}}*/
/*{{{  int  DO_keyword (next_char)*/
int  DO_keyword (next_char)
/*-----------------------*/
	int  next_char;		/* char originally at 'end_tk' */
/*----------------------------------------------------------------
	Called by 'lookup_keyword ()' after the keyword "DO" has
been read.  It returns a value for 'nwords_to_keyword', which
depends on the text following "DO" as follows:

text					nwords_to_keyword
----					-----------------
 "DO  [label]  [,]  word  ("		1  ('word' should be keyword 'WHILE')
 anything else (e.g. "DO label , id =")	0

N.B.
--  'word' means that only *letters* are allowed (not digits or '_').
--  The global variable 'end_tk' points to the position in
'textbuf' immediately after the keyword "DO".
'next_char' is the character originally stored there, which
has been temporarily replaced by '\0'.  'end_tk' isn't changed,
as the input is just inspected, not read.
----------------------------------------------------------------*/
{
	register char  *ptr;
	register int    c;


	/*{{{  set 'c' and 'ptr' to next non-blank after 'DO'*/
	/*--------------------------------------------------------------*/
	/* set 'c' to the next non-white char after 'DO', and	*/
	/* 'ptr' to point to its posn in 'textbuf'.			*/
	/*--------------------------------------------------------------*/
	ptr = end_tk;		/* N.B. 'next_char' is original value of '*ptr' */
	
	if (next_char == ' ' || next_char == '\t')
		get_next_nonwhite (c, ptr);
	else
		c = next_char;
	/*}}}*/

	/*------------------------------*/
	/* skip over "[label]  [,]"	*/
	/*------------------------------*/
	if (isdigit (c)) {
		do
			c = *++ptr;
		while (isdigit (c));
		get_nonwhite_from_here (c, ptr);
	}
	if (c == ',')
		get_next_nonwhite (c, ptr);
	

	if (isalpha (c)) {
		skip_word (c,ptr);  get_nonwhite_from_here (c, ptr);
		return  (c == '(');		/* 1 if 'c' is '(', else 0 */
	}
	else
		return  0;	/* the default */
}

/*}}}*/
/*{{{  int  FUNCTION_keyword (next_char)   -- calls 'set_local_scope ()'*/
int  FUNCTION_keyword (next_char)
/*-----------------------------*/
	int  next_char;		/* char originally at 'end_tk' */
/*----------------------------------------------------------------
	Called by 'lookup_keyword ()' after keyword "FUNCTION"
has been read.
	It returns the value 2, which is assigned to
'nwords_to_keyword'.  This value is returned in order to allow
for an optional 'RESULT' keyword in the 'function-stmt', thus:

 "... FUNCTION  name  (...)  [RESULT  (...)]"

	If the RESULT clause doesn't appear it doesn't matter,
as (barring a syntax error) there won't be a 2nd 'word' following
'FUNCTION', so any value 0 or >= 2 for 'nwords_to_keyword' will
have the required effect  (i.e. which is simply that the *next*
word *isn't* a keyword).  'nwords_to_keyword' is reset anyway at
the start of the next stmt.

	This function also calls 'set_local_scope ()' to set
(scope_type = SC_LOCAL), so that subsequent ID's have local
scope numbers (i.e. they're *declared* rather than referenced).
----------------------------------------------------------------*/
{
	set_local_scope ();

	return  2;
}

/*}}}*/
/*{{{  int  IMPLICIT_keyword (next_char)  *** IS THIS NEEDED ?? ****/
int  IMPLICIT_keyword (next_char)
/*-----------------------------*/
	int  next_char;		/* char originally at 'end_tk' */
/*----------------------------------------------------------------
	Called by 'lookup_keyword ()' after keyword "IMPLICIT"
has been read.
	In this case, 'nwords_to_keyword' must have the value 1
for the whole statement, which is achieved by setting the
special flag 'impl_stmt'.  This function also returns the
value 1, which is assigned to 'nwords_to_keyword'.
----------------------------------------------------------------*/
{
	impl_stmt = TRUE;
	return  1;
}
/*}}}*/
/*{{{  int  INTERFACE_keyword (next_char)  -- calls 'set_local_scope ()'*/
int  INTERFACE_keyword (next_char)
/*------------------------------*/
	int  next_char;		/* char originally at 'end_tk' */
/*----------------------------------------------------------------
	Called by 'lookup_keyword ()' after keyword "INTERFACE"
has been read.  It returns a value for 'nwords_to_keyword',
which depends on the text following "INTERFACE" as follows:

 text					nwords_to_keyword
 ----					-----------------
 "INTERFACE  word  ("			1  ('word' should be 'OPERATOR'
 					     or 'ASSIGNMENT')
 anything else (e.g. "INTERFACE name")	0

	It also calls 'set_local_scope ()' to set (scope_type =
SC_LOCAL), so that subsequent ID's have local scope numbers
(i.e. they're *declared* rather than referenced).

N.B. The global variable 'end_tk' points to the position in
'textbuf' immediately after the keyword "INTERFACE".
'next_char' is the character originally stored there, which has
been temporarily replaced by '\0'.  'end_tk' isn't changed, as
the input is just inspected, not read.
----------------------------------------------------------------*/
{
	register char  *ptr;
	register int   c;
	char  *next_word;

	set_local_scope ();

	if ( ! (next_char == ' ' || next_char == '\t'))
		return  0;

	ptr = end_tk + 1;
	get_nonwhite_from_here (c, ptr);

	if ( ! isalpha (c))
		return  0;

	skip_word (c,ptr);  get_nonwhite_from_here (c, ptr);

	return  (c == '(');		/* i.e. 1 if 'c' is '(', else 0 */
}
/*}}}*/
/*{{{  int  MODULE_keyword (next_char)*/
int  MODULE_keyword (next_char)
/*---------------------------*/
	int  next_char;		/* char originally at 'end_tk' */
/*----------------------------------------------------------------
	Called by 'lookup_keyword ()' after keyword "MODULE" has
been read.  It returns a value for 'nwords_to_keyword',
which depends on the text following "MODULE" as follows:

 text						nwords_to_keyword
 ----						--------------
 "MODULE  word  id"				1 ('word' should be 'PROCEDURE')
 anything else (e.g. "MODULE module_name")	0

N.B. The global variable 'end_tk' points to the position in
'textbuf' immediately after the keyword "MODULE".
'next_char' is the character originally stored there, which
has been temporarily replaced by '\0'.  'end_tk' isn't changed,
as the input is just inspected, not read.
----------------------------------------------------------------*/
{
	register char  *ptr;
	register int   c;
	char  *next_word;

	if ( ! (next_char == ' ' || next_char == '\t'))
		return  0;

	ptr = end_tk + 1;
	get_nonwhite_from_here (c, ptr);

	if ( ! isalpha (c))
		return  0;

	skip_word (c,ptr);  get_nonwhite_from_here (c, ptr);

	return  (isalpha (c) ? 1 : 0);		/* i.e. 1 if 'id', else 0 */
}
/*}}}*/
/*{{{  int  USE_keyword (next_char)   -- calls 'set_local_scope ()'*/
int  USE_keyword (next_char)
/*------------------------*/
	int  next_char;		/* char originally at 'end_tk' */
/*----------------------------------------------------------------
	Called by 'lookup_keyword ()' after the keyword "USE" has
been read.  It returns a value for 'nwords_to_keyword', which
depends on the text following "USE" as follows:

text					nwords_to_keyword
----					-----------------
 "USE  id  ,  word  :"			2  ('word' should be keyword 'ONLY')
 anything else (e.g. "USE id , id =>")	0

	It also calls 'set_local_scope ()' to set (scope_type =
SC_LOCAL), so that subsequent ID's have local scope numbers
(i.e. they're *declared* rather than referenced).

N.B.
--  'word' means that only *letters* are allowed (not digits or '_').
--  The global variable 'end_tk' points to the position in
'textbuf' immediately after the keyword "USE".
'next_char' is the character originally stored there, which
has been temporarily replaced by '\0'.  'end_tk' isn't changed,
as the input is just inspected, not read.
----------------------------------------------------------------*/
{
	register char  *ptr;
	register int    c;

	set_local_scope ();

	/*{{{  set 'c' and 'ptr' to next non-blank after 'USE'*/
	/*--------------------------------------------------------------*/
	/* set 'c' to the next non-white char after 'USE', and	*/
	/* 'ptr' to point to its posn in 'textbuf'.			*/
	/*--------------------------------------------------------------*/
	ptr = end_tk;		/* N.B. 'next_char' is original value of '*ptr' */
	
	if (next_char == ' ' || next_char == '\t')
		get_next_nonwhite (c, ptr);
	else
		c = next_char;
	/*}}}*/

	if ( !isalpha (c))
		return  0;	/* the default */

	skip_id (c,ptr);  get_nonwhite_from_here (c, ptr);

	if (c != ',')
		return  0;	/* the default */

	get_next_nonwhite (c, ptr);

	if ( !isalpha (c))
		return  0;	/* the default */

	skip_word (c,ptr);  get_nonwhite_from_here (c, ptr);

	return  ((c == ':') ? 2 : 0);
}


/*}}}*/

/*--------------*/
/* for HPF...	*/
/*--------------*/
/*{{{  int  ALIGN_keyword (next_char)      -- may call 'set_local_scope()'*/
int  ALIGN_keyword (next_char)
/*--------------------------*/
	int  next_char;		/* char originally at 'end_tk' */
/*{{{  DISCUSSION*/
/*----------------------------------------------------------------
	Called by 'lookup_keyword ()' after the HPF keyword
"(RE)ALIGN" has been read.  It returns a value for
'nwords_to_keyword', which depends on the text following
"(RE)ALIGN" as follows:

N.B. (nwords_to_keyword == 0) means that no more keywords are expected.


(RE)ALIGN can occur in these contexts:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

(i) in an attr in a combined directive:		nwords_to_keyword
---------------------------------------		-----------------
"(RE)ALIGN  (..)  WITH  [*]  id  [(..)]"	1  (case 1)
"(RE)ALIGN  WITH  [*]  id  [(..)]"		1  (case 2)

(ii) in a (RE)ALIGN directive:			nwords_to_keyword
------------------------------			-----------------
"(RE)ALIGN  id (..)  WITH  [*]  id  [(..)]"	2  (case 3)


Rules for determining 'nwords_to_keyword'
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
text						nwords_to_keyword
----						-----------------
"(RE)ALIGN  ("					1  (case 1 above)
"(RE)ALIGN  id  { ("				2  (case 3 above)
                { anything else"		1  (case 2 above)

If the input doesn't conform to one of the above cases then there's
a syntax error, so it doesn't matter what value is returned for
'nwords_to_keyword'!

	For ALIGN only (not REALIGN), this also calls 'set_local_scope ()'
to set (scope_type = SC_LOCAL), so that the next ID has local
scope number (i.e. it's *declared* rather than referenced).
                                           

N.B. The global variable 'end_tk' points to the position in
'textbuf' immediately after the keyword "(RE)ALIGN".
'next_char' is the character originally stored there, which
has been temporarily replaced by '\0'.  'end_tk' isn't changed,
as the input is just inspected, not read.
----------------------------------------------------------------*/
/*}}}*/
{
	register char  *ptr;
	register int    c;

	if (*(end_tk - 6) != 'E')	/* for ALIGN only, *not* REALIGN */
		set_local_scope ();

	/*{{{  set 'c' and 'ptr' to next non-blank after '(RE)ALIGN'*/
	/*--------------------------------------------------------------*/
	/* set 'c' to the next non-white char after 'ALIGN', and	*/
	/* 'ptr' to point to its posn in 'textbuf'.			*/
	/*--------------------------------------------------------------*/
	ptr = end_tk;		/* N.B. 'next_char' is original value of '*ptr' */
	
	if (next_char == ' ' || next_char == '\t')
		get_next_nonwhite (c, ptr);
	else
		c = next_char;
	/*}}}*/

	if (c == '(')
		/*----------------------------------------------*/
		/* "(RE)ALIGN  ("  -- return 1  (case 1)	*/
		/*----------------------------------------------*/
		return  1;
		
	if (isalpha (c)) {
		skip_id (c, ptr);  get_nonwhite_from_here (c, ptr);
	}		

	/*-------------------------------------------------------*/
	/* "(RE)ALIGN  id  ("              -- return 2  (case 3) */
	/* "(RE)ALIGN  id  anything else"  -- return 1  (case 2) */
	/*-------------------------------------------------------*/
	return  ((c == '(') ? 2 : 1);
}
/*}}}*/
/*{{{  int  DISTRIBUTE_keyword (next_char) -- may call 'set_local_scope ()'*/
int  DISTRIBUTE_keyword (next_char)
/*-------------------------------*/
	int  next_char;		/* char originally at 'end_tk' */
/*{{{  DISCUSSION*/
/*----------------------------------------------------------------
	Called by 'lookup_keyword ()' after the HPF keyword
"(RE)DISTRIBUTE" has been read.  It returns a value for
'nwords_to_keyword', which depends on the text following
"(RE)DISTRIBUTE" as follows:

N.B. (nwords_to_keyword == 0) means that no more keywords are expected.

(RE)DISTRIBUTE can occur in these contexts:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

(i) in an attr in a combined directive:		nwords_to_keyword
---------------------------------------		-----------------
                  {(..)  } { :"			0  (case 1;  id-list follows) 
 "(RE)DISTRIBUTE  {* (..)} { anything else"	1  (case 2;  followed by "ONTO"
                  {*     }			              or  ", attr")

                        {id  } { ," }		   (case 3)
 "(RE)DISTRIBUTE  ONTO  {* id} { :" }		1  (case 4)
                        {*   }			   (  "   )

(ii) in a (RE)DISTRIBUTE directive:		nwords_to_keyword
-----------------------------------		-----------------
                      {(..)  }        {  id"}
 "(RE)DISTRIBUTE  id  {* (..)}  ONTO  {* id"}	2
                      {*     }        {  *" }

                      {(..)  } { \n"}		0 or 2   (We can set it to 2,
 "(RE)DISTRIBUTE  id  {* (..)} { ;" }		although there's only one more
                      {*     }			'word', as it's reset at start
						of next stmt.)
Rules for determining 'nwords_to_keyword'
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

                  {(..)  } { :"			0  (case 1)
 "(RE)DISTRIBUTE  {* (..)} { anything else"	1  (case 2)
                  {*     }

 "(RE)DISTRIBUTE  id1 { letter"			1  (case 3;  id1 == "ONTO")
                      { *  [id2]  <, | :>"	1  (case 4;  id1 == "ONTO")
                      { anything else"		2  (case 5;  any (RE)DISTRIBUTE
						        directive;  id1 == id)

If the input doesn't conform to one of the above cases then there's
a syntax error, so it doesn't matter what value is returned for
'nwords_to_keyword'!

	For DISTRIBUTE only (not REDISTRIBUTE), this also calls
'set_local_scope ()' to set (scope_type = SC_LOCAL), so the next
ID has local scope number (i.e. it's *declared* rather than referenced).

N.B. The global variable 'end_tk' points to the position in
'textbuf' immediately after the keyword "(RE)DISTRIBUTE".
'next_char' is the character originally stored there, which
has been temporarily replaced by '\0'.  'end_tk' isn't changed,
as the input is just inspected, not read.
----------------------------------------------------------------*/
/*}}}*/

{
	register char  *ptr;
	register int    c;

	if (*(end_tk - 11) != 'E')	/* i.e. *not* for REDISTRIBUTE */
		set_local_scope ();

	/*{{{  set 'c' and 'ptr' to next non-blank after '(RE)DISTRIBUTE'*/
	/*--------------------------------------------------------------*/
	/* set 'c' to the next non-white char after 'DISTRIBUTE', and	*/
	/* 'ptr' to point to its posn in 'textbuf'.			*/
	/*--------------------------------------------------------------*/
	ptr = end_tk;		/* N.B. 'next_char' is original value of '*ptr' */
	
	if (next_char == ' ' || next_char == '\t')
		get_next_nonwhite (c, ptr);
	else
		c = next_char;
	/*}}}*/

	if (isalpha (c)) {
		/*{{{  one of cases 3--5  ("(RE)DISTRIBUTE  id  ...")*/
		/*--------------------------------------------------------------*/
		/* One of cases 3--5:						*/
		/*								*/
		/* text						return		*/
		/* ----						------		*/
		/* "(RE)DISTRIBUTE  id1 { letter"		1  (case 3)	*/
		/*                      { *  [id2]  <, | :>"	1  (case 4)	*/
		/*                      { anything else"	2  (case 5)	*/
		/*--------------------------------------------------------------*/
		
		skip_id (c, ptr);  get_nonwhite_from_here (c, ptr);
		
		if (isalpha (c))
			return  1;		/* "DISTRIBUTE  id  letter"  (case 3) */
		
		else if (c == '*') {
		
			get_next_nonwhite (c, ptr);
			if (isalpha (c)) {
				skip_id (c, ptr);  get_nonwhite_from_here (c, ptr);
			}
			if (c == ',' || c == ':')
				return  1;	/* "DISTRIBUTE  id  * [id2] <,|:>"  (case 4) */
		}
		
		return  2;		/* "DISTRIBUTE  id  anything else"  (case 5) */
		/*}}}*/
	}

	/*------------------------------------------------------*/
	/* ...else case 1 or 2:					*/
	/* text					    	return	*/
	/* ----					    	------	*/
	/*                  {(..)  } { :"		   0	*/
	/* "(RE)DISTRIBUTE  {* (..)} { anything else"	   1	*/
	/*                  {*     }				*/
	/*							*/
	/* N.B. If the input doesn't conform to one of the	*/
	/* above cases then there's a syntax error, so it	*/
	/* doesn't matter what the return value is!		*/
	/*------------------------------------------------------*/
	/*----------------------*/
	/* scan "[*] [(..)]"	*/
	/*----------------------*/
	if (c == '*')
		get_next_nonwhite (c, ptr);
	
	if (c == '(')
		get_nonwhite_after_parens (c, ptr);

	return  (c != ':');		/* 0 if char is ':', else 1 */
}
/*}}}*/
/*{{{  int  ONTO_keyword (next_char)       -- calls 'reset_normal_scope ()'*/
int  ONTO_keyword (next_char)
/*-------------------------*/
	int  next_char;		/* char originally at 'end_tk' */
/*----------------------------------------------------------------
	Called by 'lookup_keyword ()' after the HPF keyword
"ONTO" has been read.  It returns a value for 'nwords_to_keyword',
which depends on the text following "ONTO" as follows:

text				nwords_to_keyword
----				-----------------
 "ONTO  [*]  id  ,"		2   (case 1 -- an attr followed by another)
 "ONTO  *  ,"			1   (case 2 --     "      "     "     "   )
 anything else			0   (case 3 -- followed by "::", '\n', or
				                             syntax error)

	This also calls 'reset_normal_scope ()', since the
processors name immediately following TK_ONTO always has normal
scope (e.g. it may be accessed by host or use association).

N.B. The global variable 'end_tk' points to the position in
'textbuf' immediately after the keyword "ONTO".
'next_char' is the character originally stored there, which
has been temporarily replaced by '\0'.  'end_tk' isn't changed,
as the input is just inspected, not read.
----------------------------------------------------------------*/
{
	register char  *ptr;
	register int    c;

	reset_normal_scope ();
	
	/*{{{  set 'c' and 'ptr' to next non-blank after 'ONTO'*/
	/*--------------------------------------------------------------*/
	/* set 'c' to the next non-white char after 'ONTO', and	'ptr'	*/
	/* to point to its posn in 'textbuf'.				*/
	/*--------------------------------------------------------------*/
	ptr = end_tk;		/* N.B. 'next_char' is original value of '*ptr' */
	
	if (next_char == ' ' || next_char == '\t')
		get_next_nonwhite (c, ptr);
	else
		c = next_char;
	/*}}}*/

	if (c == '*') {
		get_next_nonwhite (c, ptr);
		if (c == ',')
			return  1;		/* "ONTO * ,"  (case 2)	*/

		else if (isalpha (c)) {
			skip_id (c, ptr);
			get_nonwhite_from_here (c, ptr);
			if (c == ',')
				return  2;	/* "ONTO * id ," (case 1) */
		}
	}
	else if (isalpha (c)) {
		skip_id (c, ptr);
		get_nonwhite_from_here (c, ptr);
		if (c == ',')
			return  2;		/* "ONTO id ,"  (case 1) */
	}

	return  0;	/* the default */
}
/*}}}*/
/*{{{  int  WITH_keyword (next_char)       -- calls 'reset_normal_scope ()'*/
int  WITH_keyword (next_char)
/*-------------------------*/
	int  next_char;		/* char originally at 'end_tk' */
/*----------------------------------------------------------------
	Called by 'lookup_keyword ()' after the HPF keyword
"WITH" has been read.  It returns a value for 'nwords_to_keyword',
which depends on the text following "WITH" as follows:

text						nwords_to_keyword
----						-----------------
                         { ','			2  (it's an attr that should
 "WITH  [*]  id  [(..)]" {			       be followed by another).
                         { anything else	0  (expected to be followed
						       by "::" or '\n')

If the input doesn't conform to one of these cases then there's
a syntax error, so it doesn't matter what value is returned!

	This also calls 'reset_normal_scope ()', since the
align target immediately following TK_WITH always has normal
scope (e.g. it may be accessed by host or use association).

N.B. The global variable 'end_tk' points to the position in
'textbuf' immediately after the keyword "WITH".
'next_char' is the character originally stored there, which
has been temporarily replaced by '\0'.  'end_tk' isn't changed,
as the input is just inspected, not read.
----------------------------------------------------------------*/
{
	register char  *ptr;
	register int    c;

	reset_normal_scope ();

	/*{{{  set 'c' and 'ptr' to next non-blank after 'WITH'*/
	/*--------------------------------------------------------------*/
	/* set 'c' to the next non-white char after 'WITH', and	*/
	/* 'ptr' to point to its posn in 'textbuf'.			*/
	/*--------------------------------------------------------------*/
	ptr = end_tk;		/* N.B. 'next_char' is original value of '*ptr' */
	
	if (next_char == ' ' || next_char == '\t')
		get_next_nonwhite (c, ptr);
	else
		c = next_char;
	/*}}}*/

	/*------------------------------*/
	/* skip over "[*] id  [(..)]"	*/
	/*------------------------------*/
	if (c == '*')
		get_next_nonwhite (c, ptr);
	
	if (isalpha (c)) {
		skip_id (c, ptr);  get_nonwhite_from_here (c, ptr);
	}		
	
	if (c == '(')
		get_nonwhite_after_parens (c, ptr);

	return  ((c == ',') ? 2 : 0);
}
/*}}}*/
/*}}}*/

/*{{{  <<< statement analysis functions >>>  */
/*{{{  char  *analyse_stmt (startp)  -- called by 'yylex ()' -- DBL CHECKED*/
char  *analyse_stmt (startp)
/*------------------------*/
	char  *startp;		/* pointer to the 1st nonwhite char of stmt */
/*----------------------------------------------------------------
	This is called by 'yylex () at the start of a new statement,
which may follow a stmt separator, ';'.  On entry, 'startp' points
to the start of the statement (i.e. to 'textbuf' or to the first
nonwhite char following a ';').  '*startp' is known to be nonwhite,
i.e. there is significant text in the stmt.  The caller has set
(start_tk == startp) for the benefit of errmsgs generated herein.
	This function does the following:

(i) In free form or after a ';', it checks and installs the statement
	label, if any.  (N.B. A label at the start of a fixed form
	stmt isn't copied into 'textbuf', but is processed when input.)

(ii) Scans the beginning of the stmt to see if it's an assignment
	or named construct, and sets 'nwords_to_keyword' accordingly.

	It returns a pointer to the first nonwhite character of the
statement body (after skipping over the label, if any).  If the
stmt body is empty it returns a pointer to the ';' terminator if
there is one, or NULL otherwise.
----------------------------------------------------------------*/
{
	/*{{{  decls*/
	register char  *ptr;
	register int    c;
	bool  label_given,  scan_var,  simple_name;
	int   label;
	/*}}}*/

	/*{{{  init  nparens, impl_stmt, ptr & c*/
	nparens    = 0;
	impl_stmt  = FALSE;
	
	ptr = startp;		/* known to be nonwhite! */
	c   = *ptr;		/*   "    "     "     "  */
	/*}}}*/

	if (free_source_form || startp > textbuf) {
		/*{{{  if label given, read & check it, and install if legal*/
		stmt_lp     = (Label *) NULL;
		stmt_label  = stmt_newlabel = 0;
		label_given = FALSE;		/* set for errmsg if stmt is empty */
		
		if (isdigit (c)) {
			int  ndigits;
		
			/*------------*/
			/* Read label */
			/*------------*/
			label_given = TRUE;	/* set for errmsg if stmt is empty */
		
			label   = c - '0';
			ndigits = 1;
			for (c = *++ptr;   isdigit (c);   c = *++ptr,  ++ndigits)
				label = 10 * label + (c - '0');
				
			if (free_source_form) {
				/*{{{  check label, and install if valid*/
				/*--------------------------------------------------------------*/
				/* Check that label is valid.  If so, look it up or install it,	*/
				/* and give an errmsg if it has been set already.		*/
				/*								*/
				/* N.B. The lineno reported in errmsgs is that of the *start*	*/
				/* of the label, as that's the current setting of 'start_tk'.	*/
				/*--------------------------------------------------------------*/
				if (ndigits > 5) {
					*ptr = '\0';
					errmsg ("invalid label %q - must not have more than 5 digits", startp);
					*ptr = c;
				}
				
				if ( ! label)
					errmsg ("label must be non-zero");
				
				else if (ndigits <= 5)
					lookup_or_install_label (label, SET_LABEL);   /* sets 'stmt_lp', etc */
				/*}}}*/
			}
			else {
				/*{{{  label is illegal after ';' in fixed form.  Give errmsg*/
				/*--------------------------------------------------------------*/
				/* This label follows a stmt separator (';') in fixed-form	*/
				/* input, which is illegal.  Give an errmsg to that effect,	*/
				/* and don't install the label.					*/
				/*--------------------------------------------------------------*/
				*ptr = '\0';
				errmsg ("label %q cannot follow statement separator (\';\') in fixed source form",
				         startp);
				*ptr = c;
				/*}}}*/
			}
			
			/*------------------------------------------------------
			Now (c == *ptr) is the char immediately after the label.
			Scan to the next non-white char.
				If there isn't any space after the label, give an errmsg
			if this char is a letter.  (If it *isn't* a letter a syntax
			error will be reported anyway, so don't give an errmsg here!)
			--------------------------------------------------------*/
			if (isalpha (c)) {
				*ptr = '\0';
				errmsg ("label %q must be followed by one or more spaces", startp);
				*ptr = c;
			}
			else
				get_nonwhite_from_here (c, ptr);
			
			
			/*--------------------------------------------------------------*/
			/* Now (c == *ptr) is the first nonwhite char of the stmt body.	*/
			/*--------------------------------------------------------------*/
		}
		/*}}}*/
	}
	else {
		label_given = (stmt_label != 0);
		label = stmt_label;
	}

	startp = ptr;	/* to point to start of stmt body, or ';' */
			/* or '\n' -- the return pointer value.   */

	if ( ! c || c == ';') {
		/*{{{  no stmt body.  Give errmsg and return*/
		/*--------------------------------------------------------------*/
		/* No statement body.  Must be an isolated label (which may or	*/
		/* may not be installed) or ';' at the start of the statement.	*/
		/*	In this case 'yylex()' reads the next statement		*/
		/* straight away, so nothing is returned to 'yyparse()' and	*/
		/* 'yyerror()' isn't called.					*/
		/*--------------------------------------------------------------*/
		if (label_given)
			errmsg ("no statement following label \'%d\'", label);
		else
			errmsg ("no statement for \';\' to terminate");
		
		return  (c ? startp : NULL_str);
		/*}}}*/
	}
	
	/*{{{  scan 'name [(..) [(..)]] [% name [(..) [(..)]]]*'*/
	/*--------------------------------------------------------------*/
	/* Scan the most general form of 'variable', i.e. the most	*/
	/* general LHS of an assignment stmt, viz:			*/
	/*								*/
	/*	name [(..) [(..)]] [% name [(..) [(..)]]]*		*/
	/*								*/
	/* N.B. Here we largely ignore syntax errors  (e.g. in the code	*/
	/* within parentheses, and in the form of 'name', which		*/
	/* according to the scan might start with a digit or '_', or	*/
	/* even be totally absent!).   However, if there *is* a syntax	*/
	/* error, then it doesn't matter if we mis-identify the		*/
	/* 'variable' or 'nwords_to_keyword', as a syntax error will be	*/
	/* reported anyway!						*/
	/*--------------------------------------------------------------*/
	scan_var = simple_name = TRUE;
	do {
		/*------------------------------------------------------*/
		/* Scan 'name'.						*/
		/* N.B. no error is detected if 'name' is absent, or	*/
		/* starts with a digit or '_'.  But that doesn't matter	*/
		/* -- as explained above.				*/
		/*------------------------------------------------------*/
		while (isalnum (c) || c == '_')
			c = *++ptr;
	
		get_nonwhite_from_here (c, ptr);
	
		/*----------------------------------------*/
		/* Scan  '(subscr_list) [(substr_range)]' */
		/*----------------------------------------*/
		if (c == '(') {
			simple_name = FALSE;
	
			get_nonwhite_after_parens (c, ptr);
			if (c == '(')
				get_nonwhite_after_parens (c, ptr);
		}
	
		/*----------------------------------------*/
		/* Repeat outer loop if there's a '%'     */
		/*----------------------------------------*/
		if (c == '%') {
			simple_name = FALSE;
			get_next_nonwhite (c, ptr);
		}
		else
			scan_var = FALSE;
	} while (scan_var);
	/*}}}*/

	/*{{{  set 'nwords_to_keyword'*/
	switch (c) {	/* the next sig char after the "var" */
	case  '=' :
		/*------------------*/
		/* Assignment stmt! */
		/*------------------*/
		nwords_to_keyword = 0;
		break;
	
	case  ':' :
		if (simple_name && (*(ptr+1) != ':')) {
			/*----------------------------------------------*/
			/* A labelled construct, viz:			*/
			/*	constr_name : DO ...,			*/
			/*	constr_name : IF ...,			*/
			/*	constr_name : SELECT CASE...		*/
			/* Set 'nwords_to_keyword' to 2, so TK_ID is	*/
			/* returned for the 'constr_name' and the next	*/
			/* word is interpreted as a keyword.		*/
			/*----------------------------------------------*/
			nwords_to_keyword = 2;
			break;
		}
		/*-----------------------------------------*/
		/* ... else, fall through to default case! */
		/*-----------------------------------------*/
	
	default :
		/*-----------------------------------------*/
		/* First word should be a keyword!         */
		/*-----------------------------------------*/
		nwords_to_keyword = 1;
		break;
	}
	/*}}}*/

	return  startp;
}
/*}}}*/

/*{{{  bool  analyse_IF_stmt ()  -- called by 'yyparse ()'*/
bool  analyse_IF_stmt ()
/*--------------------*/
/*----------------------------------------------------------------
	This is called by 'yyparse ()' when the phrase:

	IF (expr)

has been parsed.  On entry, 'end_tk' points to the character after
')' in the above pattern.
	The function scans the succeeding text to determine the
type of 'IF' stmt, viz:

  --  artihmetic-if-stmt (if next word is a number);
  --  if-then-stmt       (if only remaining word is "THEN");
  --  if-stmt            (otherwise);

and sets 'nwords_to_keyword' accordingly.  It also returns TRUE
in the last 2 cases (when 'expr' should be LOGICAL) and FALSE
in the first case (when 'expr' should be INTEGER).
	When the stmt type has been established, this function
also calls 'set_part_and_label ()'.  This must be done before any
output can be generated, and for an 'if-stmt' we want to output
the IF condition (transformed to an 'if-then-stmt') now, before
processing the following action stmt.

N.B.  Incidentally, after outputting the 'if-then-stmt', 'yyparse ()'
zeroes 'stmt_lp', 'stmt_label' and 'stmt_newlabel', so they're not
effective when the action stmt is processed.
----------------------------------------------------------------*/
{
	register char  *ptr;
	register int    c;
	bool   if_then_stmt,  scan_var,  simple_name;
	char  *endp,  *startp;

	ptr = end_tk;
	get_nonwhite_from_here (c, ptr);
	startp = ptr;			/* pointer to 1st sig char */

	if (isdigit (c)) {
		/*{{{  a 'goto_label' -- hence an 'arithmetic-if-stmt'*/
		/*--------------------------------------------------------------*/
		/* The number is a 'goto_label', so the statement is an		*/
		/* 'arithmetic_if_stmt'.  Hence, set 'nwords_to_keyword' = 0	*/
		/* and call 'set_part_and_label ()'.				*/
		/*--------------------------------------------------------------*/
		nwords_to_keyword = 0;
		set_part_and_label (EXEC_PART, TK_ARITH_IF);
		
		return  FALSE;		/* i.e. 'expr' should be INTEGER */
		/*}}}*/
	}
	else if (isalpha (c)) {
		/*{{{  an 'if-then-stmt' or 'if-stmt'*/
		/*-----------------------------------------------------------------
		Scan the most general form of 'variable', i.e. the most general
		RHS of an assignment stmt, viz:
			name [(..) [(..)]] [% name [(..) [(..)]]]*
		
		N.B.:
		--  Here we largely ignore syntax errors (e.g. in the code within
		parentheses, and also if 'name' starts with a digit or '_', or is
		absent).  However, if there *is* a syntax error, then it doesn't
		matter if we mis-identify 'var' or 'nwords_to_keyword', as an error
		will be reported anyway!
		--  Very similar code is contained in 'analyse_stmt ()'.
		----------------------------------------------------------------*/
		scan_var = simple_name = TRUE;
		do {
			/*------------------------------------------------------*/
			/* Scan 'name'.						*/
			/* N.B. In iteration >= 2 of the outer loop, no error	*/
			/* is detected if 'name' is absent, or starts with a	*/
			/* digit or '_'.  But that doesn't matter -- see above.	*/
			/*------------------------------------------------------*/
			while (isalnum (c) || c == '_') {
				if (islower (c)) *ptr = toupper (c);
				c = *++ptr;
			}
			endp = ptr;
		
			get_nonwhite_from_here (c, ptr);
		
			/*----------------------------------------*/
			/* Scan  '(subscr_list) [(substr_range)]' */
			/*----------------------------------------*/
			if (c == '(') {
				simple_name = FALSE;
		
				get_nonwhite_after_parens (c, ptr);
				if (c == '(')
					get_nonwhite_after_parens (c, ptr);
			}
		
			/*----------------------------------------*/
			/* Repeat outer loop if there's a '%'     */
			/*----------------------------------------*/
			if (c == '%') {
				simple_name = FALSE;
				get_next_nonwhite (c, ptr);
			}
			else
				scan_var = FALSE;
		} while (scan_var);
		
		/*--------------------------------------------------------------*/
		/* Set 'nwords_to_keyword' to 0 for an assignment and 1		*/
		/* otherwise, and call 'set_part_and_label ()'.			*/
		/*--------------------------------------------------------------*/
		nwords_to_keyword = (c != '=');		/* 0 if (c == '='), else 1 */
		
		if_then_stmt = FALSE;
		if ( !c  &&  simple_name) {
			delimit_word_in_textbuf (c, endp);
			if_then_stmt = str_match (startp, "THEN");
			restore_textbuf (c, endp);
		}
		
		set_part_and_label (EXEC_PART, (if_then_stmt ? TK_IF_CONSTR : TK_IF));
		
		return  TRUE;		/* i.e. 'expr' should be LOGICAL */
		/*}}}*/
	}
	else {
		/*{{{  syntax error!*/
		/*--------------------------------------------------------------*/
		/* A syntax error!  Hence, the value of 'nwords_to_keyword'	*/
		/* doesn't matter and we needn't call 'set_part_and_label ()'!	*/
		/*--------------------------------------------------------------*/
		nwords_to_keyword = 0;
		
		return  TRUE;		/* ...actually the return value doesn't matter! */
		/*}}}*/
	}
}
/*}}}*/

/*{{{  analyse_WHERE () -- called by 'yyparse ()'*/
analyse_WHERE ()
/*------------*/
/*----------------------------------------------------------------
	This is called by 'yyparse ()' when the phrase:

	WHERE (mask-expr)

has been parsed.  On entry, 'end_tk' points to the character after
')' in the above pattern.
	The function scans the succeeding text to determine the
type of stmt, viz:

  --  where-constr-stmt  (if next nonblank is '\n');
  --  where-stmt         (otherwise).

and then calls 'set_part_and_label ()' with the appropriate stmt
token ('TK_WHERE' or 'TK_WHERE_CONSTR').  This must be done
before any output can be generated, and for a 'where-stmt' we want
to output the 'mask-expr', copied into a temp array, before
processing the following 'assignment-stmt'.
----------------------------------------------------------------*/
{
	register char  *ptr;
	register int    c;

	ptr = end_tk;
	get_nonwhite_from_here (c, ptr);

	set_part_and_label (EXEC_PART,
	                    (( !c || c == ';') ? TK_WHERE_CONSTR : TK_WHERE));
}
/*}}}*/
/*}}}*/

/*{{{  <<< 'text scanning' functions  (used via macro, not directly!) >>>*/
/*{{{  char  *GET_NONWHITE_AFTER_PARENS (ptr)*/
char  *GET_NONWHITE_AFTER_PARENS (ptr)
/*----------------------------------*/
	register char  *ptr;
/*----------------------------------------------------------------
	On entry 'ptr' points to a '(' in 'textbuf'.  This function
returns a pointer to the next non-blank char after the matching ')',
or to '\0' (i.e. end-of-line) if found earlier!

N.B.
---
-- This function is never referenced directly, but only via the macro:

#define  get_nonwhite_after_parens(c,ptr)
		(c = *(ptr = GET_NONWHITE_AFTER_PARENS (ptr)))

which provides a more natural interface.  This function has an
upper-case name to prevent it being used directly by mistake!

-- The task is complicated by the fact that '(' and ')' may occur
within character strings, in which case we should ignore them!
We assume that the initial '(' *isn't* in a string, and that we
must find the syntactically matching closing ')'.
----------------------------------------------------------------*/
{
	register int  c;
	int  nparens,  str_delim;

	nparens = 1;
	while (c = *++ptr) {
		switch (c) {
		case  '(' :
			nparens++;
			break;

		case  ')' :
			if ( ! --nparens) {
				get_next_nonwhite (c, ptr);
				return  ptr;	/* nonwhite after "(..)" */
			}
			break;

		case  '\'' :
		case  '\"' :
			str_delim = c;
			do
				c = *++ptr;
			while (c  &&  (c != str_delim));
			if (!c)			/* (*ptr == '\0'), i.e.   */
				return  ptr;	/* stmt ends prematurely! */
			break;
		}
	}
	return  ptr;       /* (*ptr == '\0'), i.e. stmt ends prematurely! */
}
/*}}}*/
/*}}}*/
/*}}}*/
